//=============================================================================
//  PRGUI.java
//  
//  A visual tool to show the placement and routing of a multiple layered
//  FPGA with input from VPR
//  
//  Created by James Foltz, Summer 2004, at the University of Minnesota
//  for NSF REU program
//
//  jamesfoltz@hotmail.com
//=============================================================================

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.io.*;
import java.text.*;
import java.util.*;
import javax.swing.*;

public class PRGUI {
	
	public static String placementFileLocation, netListFileLocation,
		critPathFileLocation, pName = "", netName = "", critName = "";
	public static int XTot, YTot, ZTot, LoadNum=0, showNetList=0, NetListCon=0,
		NetListShow=0, NumCritPath, view=3, netNum=0, input1, input2, NLF=0,
		critNum, coordPos=1;
	public static double width, height, xLeft, yTop, scaleFactor;
	
	public static void main(String[] args) throws IOException {
		
		JFrame.setDefaultLookAndFeelDecorated(true);
		PRGUIDesktopPane desktop = new PRGUIDesktopPane();
		desktop.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		desktop.setVisible(true);
	}
}

class PRGUIDesktopPane extends JFrame implements ActionListener {
	
	JDesktopPane desktop;
	static PRGUIFrame frame;

    public PRGUIDesktopPane() {
    	
        super("PRGUI:  A Visual Tool for Displaying Placement and Routing");

        //Make the big window be indented 50 pixels from each edge
        //of the screen.
        int inset = 50;
        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        setBounds(inset, inset,
                  screenSize.width  - inset*2,
                  screenSize.height - inset*2);

        //Set up the GUI.
        desktop = new JDesktopPane(); //a specialized layered pane
        setContentPane(desktop);

        //Make dragging a little faster but perhaps uglier.
        desktop.setDragMode(JDesktopPane.OUTLINE_DRAG_MODE);
        setJMenuBar(createMenuBar());
        
        setIconImage(new ImageIcon("icon.gif").getImage());
	}
   
    public JMenuBar createMenuBar() {
    	
    	JMenuBar menuBar = new JMenuBar();
    	
    	JMenu fileMenu = new JMenu("File");
    	fileMenu.setMnemonic(KeyEvent.VK_F);
    	menuBar.add(fileMenu);
		
		JMenu layerMenu = new JMenu("View");
		layerMenu.setMnemonic(KeyEvent.VK_V);
		menuBar.add(layerMenu);
		
		JMenu netListMenu = new JMenu("NetList");
		netListMenu.setMnemonic(KeyEvent.VK_N);
		menuBar.add(netListMenu);
		
		JMenu critPathMenu = new JMenu("Critical Paths");
		critPathMenu.setMnemonic(KeyEvent.VK_C);
		menuBar.add(critPathMenu);
		
		JMenu zoomMenu = new JMenu("Zoom");
		zoomMenu.setMnemonic(KeyEvent.VK_Z);
		menuBar.add(zoomMenu);
		
		JMenu rotMenu = new JMenu("Rotate");
		rotMenu.setMnemonic(KeyEvent.VK_R);
		menuBar.add(rotMenu);
		
		JMenu helpMenu = new JMenu("Help");
		helpMenu.setMnemonic(KeyEvent.VK_H);
		menuBar.add(helpMenu);
		
		JMenuItem miLoadF = new JMenuItem("Load New Placement File...");
		miLoadF.setActionCommand("loadfile");
		miLoadF.addActionListener(this);
		//fileMenu.add(miLoadF);
		
		JMenuItem loadFiles = new JMenuItem("Load Circuit Files...");
		loadFiles.setActionCommand("loadfiles");
		loadFiles.setMnemonic(KeyEvent.VK_L);
        loadFiles.setAccelerator(KeyStroke.getKeyStroke(
        	KeyEvent.VK_L, ActionEvent.CTRL_MASK));
		loadFiles.addActionListener(this);
		fileMenu.add(loadFiles);
		
        JMenuItem miQuit = new JMenuItem("Quit");
        miQuit.setMnemonic(KeyEvent.VK_Q);
        miQuit.setAccelerator(KeyStroke.getKeyStroke(
        	KeyEvent.VK_Q, ActionEvent.CTRL_MASK));
        miQuit.setActionCommand("quit");
        miQuit.addActionListener(this);
        fileMenu.add(miQuit);
        
        JMenuItem go_to = new JMenuItem("Show 2D Layer...");
        go_to.setActionCommand("go_to");
        go_to.addActionListener(this);
        layerMenu.add(go_to);
        
        JMenuItem threeDim = new JMenuItem("Show All Layers in 3D");
        threeDim.setActionCommand("3");
        threeDim.addActionListener(this);
        layerMenu.add(threeDim);
        
        JMenuItem rotCloc = new JMenuItem
        	("Rotate 90 Degrees ClockWise");
        rotCloc.setActionCommand("rcw");
        rotCloc.setMnemonic(KeyEvent.VK_A);
        rotCloc.setAccelerator(KeyStroke.getKeyStroke(
        	KeyEvent.VK_A, ActionEvent.CTRL_MASK));
        rotCloc.addActionListener(this);
        rotMenu.add(rotCloc);
        
        JMenuItem rotCoun = new JMenuItem
        	("Rotate 90 Degrees CounterClockWise");
        rotCoun.setActionCommand("rccw");
        rotCoun.setMnemonic(KeyEvent.VK_S);
        rotCoun.setAccelerator(KeyStroke.getKeyStroke(
        	KeyEvent.VK_S, ActionEvent.CTRL_MASK));
        rotCoun.addActionListener(this);
        rotMenu.add(rotCoun);
        
        JMenuItem netLoad = new JMenuItem("Load NetList File...");
        netLoad.setActionCommand("netload");
        netLoad.addActionListener(this);
        //netListMenu.add(netLoad);
        
        JMenuItem netFilter = new JMenuItem("Filtering And Viewing Options...");
        netFilter.setActionCommand("netFilter");
        netFilter.addActionListener(this);
        netListMenu.add(netFilter);

        JMenuItem netHide = new JMenuItem("Hide NetList");
        netHide.setActionCommand("nethide");
        netHide.addActionListener(this);
        netListMenu.add(netHide);
        
        JMenuItem critLoad = new JMenuItem("Load Critical Paths File...");
        critLoad.setActionCommand("critload");
        critLoad.addActionListener(this);
        //critPathMenu.add(critLoad);
        
        JMenuItem critShow = new JMenuItem("Show Critical Paths...");
        critShow.setActionCommand("critshow");
        critShow.addActionListener(this);
        critPathMenu.add(critShow);
        
        JMenuItem critHide = new JMenuItem("Hide Critical Paths");
        critHide.setActionCommand("crithide");
        critHide.addActionListener(this);
        critPathMenu.add(critHide);
        
        JMenuItem zoomIn = new JMenuItem("Zoom In");
		zoomIn.setActionCommand("zoomin");
		zoomIn.setMnemonic(KeyEvent.VK_Z);
        zoomIn.setAccelerator(KeyStroke.getKeyStroke(
        	KeyEvent.VK_Z, ActionEvent.CTRL_MASK));
		zoomIn.addActionListener(this);
		zoomMenu.add(zoomIn);
        
        JMenuItem zoomOut = new JMenuItem("Zoom Out");
		zoomOut.setActionCommand("zoomout");
		zoomOut.setMnemonic(KeyEvent.VK_X);
        zoomOut.setAccelerator(KeyStroke.getKeyStroke(
        	KeyEvent.VK_X, ActionEvent.CTRL_MASK));
		zoomOut.addActionListener(this);
		zoomMenu.add(zoomOut);
		
		JMenuItem help = new JMenuItem("Help");
		help.setActionCommand("help");
		help.addActionListener(this);
		helpMenu.add(help);
		
		JMenuItem about = new JMenuItem("About");
		about.setActionCommand("about");
		about.addActionListener(this);
		helpMenu.add(about);
		
        return menuBar;
    }
    
    public void actionPerformed(ActionEvent e) {
    	
        if ("quit".equals(e.getActionCommand())) { 
        	System.exit(0);
        }
        else if("loadfiles".equals(e.getActionCommand())) {
        	loadFiles();
        }
        else if("loadfile".equals(e.getActionCommand())) {
        	JFileChooser fc = new JFileChooser();
        	int returnVal = fc.showOpenDialog(PRGUIDesktopPane.this);
        	if (returnVal == JFileChooser.APPROVE_OPTION) {
                File file = fc.getSelectedFile();
                PRGUI.placementFileLocation = file.getPath();
                PRGUI.pName = file.getName();
            }
            PRGUI.showNetList=0;
            PRGUI.netListFileLocation = null;
            try {
        	newPlacement();	
        	PRGUI.LoadNum++;
        	}
        	catch (IOException exception) {
			System.out.println("The File Was Not Found");
			}  
        }
        else if("zoomout".equals(e.getActionCommand())) {
        	PRGUI.width = PRGUI.width/1.5;
			PRGUI.height = PRGUI.height/1.5;
			PRGUI.xLeft = PRGUI.xLeft/1.5;
			PRGUI.yTop = PRGUI.yTop/1.5;
			PRGUI.scaleFactor = PRGUI.scaleFactor/1.5;
			frame.revalidate();
			frame.repaint();
			PRGUIFrame.panel.setPreferredSize(new Dimension(
				((PRGUI.XTot+2)*(int)PRGUI.scaleFactor + (int)
				(PRGUI.scaleFactor*2.5)), (PRGUI.YTot*(int)PRGUI.scaleFactor*
				(PRGUI.ZTot) + (int)(PRGUI.scaleFactor*3.5))));
        	PRGUIFrame.scrollPane.revalidate();
        } 
        else if("zoomin".equals(e.getActionCommand())) {
        	PRGUI.width = PRGUI.width*1.5;
			PRGUI.height = PRGUI.height*1.5;
			PRGUI.xLeft = PRGUI.xLeft*1.5;
			PRGUI.yTop = PRGUI.yTop*1.5;
			PRGUI.scaleFactor = PRGUI.scaleFactor*1.5;
			frame.revalidate();
			frame.repaint();
			PRGUIFrame.panel.setPreferredSize(new Dimension(
				((PRGUI.XTot+2)*(int)PRGUI.scaleFactor + (int)
				(PRGUI.scaleFactor*2.5)), (PRGUI.YTot*(int)PRGUI.scaleFactor*
				(PRGUI.ZTot) + (int)(PRGUI.scaleFactor*3.5))));
        	PRGUIFrame.scrollPane.revalidate();
        }
        else if("go_to".equals(e.getActionCommand())) {
        	String Layer = JOptionPane.showInputDialog
        		("What Layer Do You Want To Go To?\n"
        		+"Choose A Layer From 0 To " + (PRGUI.ZTot-1),
        		"Go To Layer...");
			PRGUIFrame.layerNum = Integer.parseInt(Layer);
			PRGUI.view = 2;
        	frame.revalidate();
			frame.repaint();
        	PRGUIFrame.panel.setPreferredSize(new Dimension(
				((PRGUI.XTot+2)*(int)PRGUI.scaleFactor + (int)
				(PRGUI.scaleFactor*2.5)), (PRGUI.YTot*(int)PRGUI.scaleFactor*
				(PRGUI.ZTot) + (int)(PRGUI.scaleFactor*3.5))));
        	PRGUIFrame.scrollPane.revalidate();	
        }
        else if("3".equals(e.getActionCommand())) {
        	PRGUI.view = 3;
        	frame.revalidate();
			frame.repaint();
        	PRGUIFrame.panel.setPreferredSize(new Dimension(
				((PRGUI.XTot+2)*(int)PRGUI.scaleFactor + (int)
				(PRGUI.scaleFactor*2.5)), (PRGUI.YTot*(int)PRGUI.scaleFactor*
				(PRGUI.ZTot) + (int)(PRGUI.scaleFactor*3.5))));
        	PRGUIFrame.scrollPane.revalidate();
        }
        else if("rccw".equals(e.getActionCommand())) {
        	int rotateTempX, rotateTempY, xHolder, yHolder;
        	for(int i = 0; i < NumberofComponents.recall(); i++) {
        		xHolder = XHolder.GetX(i);
        		yHolder = YJavaHolder.GetY(i);
        		rotateTempX = (xHolder * (int)(Math.cos(Math.PI/2))) - 
        			(yHolder * (int)(Math.sin(Math.PI/2))) + PRGUI.YTot + 1;
        		rotateTempY = (xHolder * (int)(Math.sin(Math.PI/2))) + 
        			(yHolder * (int)(Math.cos(Math.PI/2)));
        		XHolder.AddX(i, rotateTempX);
        		YJavaHolder.AddY(i, rotateTempY);
        		YHolder.AddY(i, Math.abs((PRGUI.XTot + 2) - rotateTempY));
        	}
 
 			PRGUI.coordPos--;       	
        	if(PRGUI.coordPos==0){
        		PRGUI.coordPos=4;
        	}
 			       	
        	int temp;
        	temp = PRGUI.XTot;
        	PRGUI.XTot = PRGUI.YTot;
        	PRGUI.YTot = temp;

        	frame.revalidate();
			frame.repaint();
			PRGUIFrame.panel.setPreferredSize(new Dimension(
				((PRGUI.XTot+2)*(int)PRGUI.scaleFactor + (int)
				(PRGUI.scaleFactor*2.5)), (PRGUI.YTot*(int)PRGUI.scaleFactor*
				(PRGUI.ZTot) + (int)(PRGUI.scaleFactor*3.5))));
        	PRGUIFrame.scrollPane.revalidate();
        }
        else if("rcw".equals(e.getActionCommand())) {
        	int rotateTempX, rotateTempY, xHolder, yHolder;
        	for(int i = 0; i < NumberofComponents.recall(); i++) {
        		xHolder = XHolder.GetX(i);
        		yHolder = YJavaHolder.GetY(i);
        		rotateTempX = (xHolder * (int)(Math.cos(Math.PI/2))) + 
        			(yHolder * (int)(Math.sin(Math.PI/2)));
        		rotateTempY = (yHolder * (int)(Math.cos(Math.PI/2))) - 
        			(xHolder * (int)(Math.sin(Math.PI/2))) + PRGUI.XTot + 1;
        		XHolder.AddX(i, rotateTempX);
        		YJavaHolder.AddY(i, rotateTempY);
        		YHolder.AddY(i, Math.abs((PRGUI.XTot + 2) - rotateTempY));
        	}
        	
        	PRGUI.coordPos++;       	
        	if(PRGUI.coordPos==5){
        		PRGUI.coordPos=1;
        	}
        	
        	int temp;
        	temp = PRGUI.XTot;
        	PRGUI.XTot = PRGUI.YTot;
        	PRGUI.YTot = temp;

        	frame.revalidate();
			frame.repaint();
			PRGUIFrame.panel.setPreferredSize(new Dimension(
				((PRGUI.XTot+2)*(int)PRGUI.scaleFactor + (int)
				(PRGUI.scaleFactor*2.5)), (PRGUI.YTot*(int)PRGUI.scaleFactor*
				(PRGUI.ZTot) + (int)(PRGUI.scaleFactor*3.5))));
        	PRGUIFrame.scrollPane.revalidate();
        }
        else if("netload".equals(e.getActionCommand())) {
        	JFileChooser fc2 = new JFileChooser();
        	int returnVal2 = fc2.showOpenDialog(PRGUIDesktopPane.this);
        	if (returnVal2 == JFileChooser.APPROVE_OPTION) {
                File file2 = fc2.getSelectedFile();
                PRGUI.netListFileLocation = file2.getPath();
                PRGUI.netName = file2.getName();
            }
            try {
        	newNetList();	
        	}
        	catch (IOException exception) {
			System.out.println("The File Was Not Found");
			}
        }
        else if("netFilter".equals(e.getActionCommand())) {
        	PRGUI.showNetList = 0;
        	PRGUI.NLF = 1;
        	netListFrame nlf = new netListFrame();
        	nlf.setVisible(true); //necessary as of 1.3
        	desktop.add(nlf);
        	nlf.toFront();
        	frame.revalidate();
			frame.repaint();
			PRGUIFrame.panel.setPreferredSize(new Dimension(
				((PRGUI.XTot+2)*(int)PRGUI.scaleFactor + (int)
				(PRGUI.scaleFactor*2.5)), (PRGUI.YTot*(int)PRGUI.scaleFactor*
				(PRGUI.ZTot) + (int)(PRGUI.scaleFactor*3.5))));
        	PRGUIFrame.scrollPane.revalidate();	
        }
        else if("nethide".equals(e.getActionCommand())) {
        	PRGUI.showNetList=0;
        	frame.revalidate();
			frame.repaint();
			PRGUIFrame.panel.setPreferredSize(new Dimension(
				((PRGUI.XTot+2)*(int)PRGUI.scaleFactor + (int)
				(PRGUI.scaleFactor*2.5)), (PRGUI.YTot*(int)PRGUI.scaleFactor*
				(PRGUI.ZTot) + (int)(PRGUI.scaleFactor*3.5))));
        	PRGUIFrame.scrollPane.revalidate();	
        }
        else if("critload".equals(e.getActionCommand())) {
        	JFileChooser fc3 = new JFileChooser();
        	int returnVal3 = fc3.showOpenDialog(PRGUIDesktopPane.this);
        	if (returnVal3 == JFileChooser.APPROVE_OPTION) {
                File file3 = fc3.getSelectedFile();
                PRGUI.critPathFileLocation = file3.getPath();
                PRGUI.critName = file3.getName();
            }
            try {
        		newCritPath();	
        		PRGUI.LoadNum++;
        	}
        	catch (IOException exception) {
				System.out.println("The File Was Not Found");
			}
        }
        else if("critshow".equals(e.getActionCommand())) {
        	PRGUI.NumCritPath = 0;
        	critPathFrame cpf = new critPathFrame();
        	cpf.setVisible(true); //necessary as of 1.3
        	desktop.add(cpf);
        	cpf.toFront();
			frame.revalidate();
			frame.repaint();
			PRGUIFrame.panel.setPreferredSize(new Dimension(
				((PRGUI.XTot+2)*(int)PRGUI.scaleFactor + (int)
				(PRGUI.scaleFactor*2.5)), (PRGUI.YTot*(int)PRGUI.scaleFactor*
				(PRGUI.ZTot) + (int)(PRGUI.scaleFactor*3.5))));
        	PRGUIFrame.scrollPane.revalidate();	
        }
        else if("crithide".equals(e.getActionCommand())) {
        	PRGUI.NumCritPath = 0;
        	frame.revalidate();
			frame.repaint();
			PRGUIFrame.panel.setPreferredSize(new Dimension(
				((PRGUI.XTot+2)*(int)PRGUI.scaleFactor + (int)
				(PRGUI.scaleFactor*2.5)), (PRGUI.YTot*(int)PRGUI.scaleFactor*
				(PRGUI.ZTot) + (int)(PRGUI.scaleFactor*3.5))));
        	PRGUIFrame.scrollPane.revalidate();	
        }
        else if("about".equals(e.getActionCommand())) {
        	aboutFrame af = new aboutFrame();
        	af.setVisible(true); //necessary as of 1.3
        	desktop.add(af);
        	af.toFront();
        }
        else if("help".equals(e.getActionCommand())) {
        	try {
        	helpFrame();	
        	}
        	catch (IOException exception) {
			System.out.println("The File Was Not Found");
			}
        }
        else { //quit
            System.exit(0);
        }
    }
    

	public void newNetList() throws IOException {
		
		//Declare All Variables Used in newNetList
		
		String netFileLoc, netLine, temp, BlockNam, BlockSink;
		int numOfCLB = 0, i = 0;
		
		//Input NetList File Information
		
		netFileLoc = PRGUI.netListFileLocation;
		BufferedReader BRNet = new BufferedReader
			(new FileReader(netFileLoc));
		netLine = BRNet.readLine();
		StringTokenizer tokenizer = new StringTokenizer(netLine);
		temp = new String("");
		
		
		//Find the number of CLB Blocks since they are the only ones with
		//that show what they are connect to, so as to find the array size
		
		while(netLine!=null) {
			if(!netLine.trim().equals("")){
				tokenizer = new StringTokenizer(netLine);
				temp = tokenizer.nextToken();
				if(temp.equals(".clb")) {
					numOfCLB++;
				}
			}
			netLine = BRNet.readLine();
		}
		
		PRGUI.netNum = numOfCLB;
		
		//Find the number of connections each CLB Block has
		
		BRNet = new BufferedReader(new FileReader(netFileLoc));
		netLine = BRNet.readLine();
		while(netLine!=null && PRGUI.NetListCon == 0) {
			
			if(!netLine.trim().equals("")) {
				tokenizer = new StringTokenizer(netLine);
				temp = tokenizer.nextToken();
				if(temp.equals(".clb")) {
					netLine = BRNet.readLine();
					tokenizer = new StringTokenizer(netLine);
					temp = tokenizer.nextToken();
					PRGUI.NetListCon = tokenizer.countTokens();
				}
			}
			netLine = BRNet.readLine();
		}
	
		//Set up the Arrays
		
		BNSource.initArray(numOfCLB);
		BNSink.initArray(numOfCLB, PRGUI.NetListCon);
		
		BRNet = new BufferedReader(new FileReader(netFileLoc));
		netLine = BRNet.readLine();
		while(netLine!=null) {
			
			if(!netLine.trim().equals("")) {
				tokenizer = new StringTokenizer(netLine);
				temp = tokenizer.nextToken();
				if(temp.equals(".clb")) {
					BlockNam = tokenizer.nextToken();
					BNSource.AddBNSource(i, BlockNam);
					netLine = BRNet.readLine();
					tokenizer = new StringTokenizer(netLine);
					BlockSink = tokenizer.nextToken();
					for(int j = 0; j < PRGUI.NetListCon; j++) {
						BlockSink = tokenizer.nextToken();
						BNSink.AddBNSink(i, j, BlockSink);
					}
					i++;
				}
			}
			netLine = BRNet.readLine();
		}
		
	}

    public void newPlacement() throws IOException {
    	
    	//Declare All Variables Used in newPlacement
    	
		String BlockNumber, PlaceFileLoc, PlaceLine, temp, BlockName;
		int count, XVal, YVal, YTot, ZVal, total;
		
		//Inputs Placement File Information
		
		PlaceFileLoc = PRGUI.placementFileLocation;
		BufferedReader BRPlace = new BufferedReader
			(new FileReader(PlaceFileLoc));
		
		//Reads in Placement File and Stores data in Arrays
		
		PlaceLine = BRPlace.readLine();
		PlaceLine = BRPlace.readLine();
		StringTokenizer tokenizer = new StringTokenizer(PlaceLine);
		temp = tokenizer.nextToken();
		temp = tokenizer.nextToken();
		XVal = Integer.parseInt(tokenizer.nextToken());
		temp = tokenizer.nextToken();
		YTot = Integer.parseInt(tokenizer.nextToken());
		temp = tokenizer.nextToken();
		ZVal = Integer.parseInt(tokenizer.nextToken());
		total = (XVal + 2) * (YTot + 2) * ZVal;
		PRGUI.XTot = XVal;
		PRGUI.YTot = YTot;
		PRGUI.ZTot = ZVal;
		XHolder.initArray(total);
		YHolder.initArray(total);
		YJavaHolder.initArray(total);
		ZHolder.initArray(total);
		BNHolder.initArray(total);
		BNameHolder.initArray(total);
		PlaceLine = BRPlace.readLine();
		PlaceLine = BRPlace.readLine();
		PlaceLine = BRPlace.readLine();
		PlaceLine = BRPlace.readLine();
		count = 0;
		while(PlaceLine!=null && !PlaceLine.trim().equals("")) {
			tokenizer = new StringTokenizer(PlaceLine);
			BlockName = tokenizer.nextToken();
			XVal = Integer.parseInt(tokenizer.nextToken());
			YVal = Integer.parseInt(tokenizer.nextToken());
			ZVal = Integer.parseInt(tokenizer.nextToken());
			temp = tokenizer.nextToken();
			BlockNumber = tokenizer.nextToken();
			XHolder.AddX(count, XVal);
			YJavaHolder.AddY(count, YVal);
			YHolder.AddY(count, Math.abs((YTot + 2) - YVal));
			ZHolder.AddZ(count, ZVal);
			BNHolder.AddBN(count, BlockNumber);
			BNameHolder.AddBName(count, BlockName);
			count++;
			PlaceLine = BRPlace.readLine();
		}
		
		//Stores the Actual Number of Components
		
		NumberofComponents.store(count);
		
		//Initialize the Frame with initFrame Method
		
		initFrame();
		
        
    }
    
    public void initFrame() {
    	
    	//Set Default Size
		
		PRGUI.width = 40;
		PRGUI.height = 40;
		PRGUI.xLeft = 40;
		PRGUI.yTop = 80;
		PRGUI.scaleFactor = 80;
    	
	   	//Initializes frame
    	if(PRGUI.LoadNum==0) {    	
    		frame = new PRGUIFrame();
        	frame.setVisible(true); //necessary as of 1.3
        	desktop.add(frame);
        	try {
            	frame.setSelected(true);
        	} 
        	catch (java.beans.PropertyVetoException e) {
        	}
        }
        if(PRGUI.LoadNum>=1) {
        	PRGUIFrame.layerNum = 0;
        	PRGUI.width = 40;
			PRGUI.height = 40;
			PRGUI.xLeft = 40;
			PRGUI.yTop = 80;
			PRGUI.scaleFactor = 80;
			try {
				frame.setClosed(true);
			}
			catch (java.beans.PropertyVetoException e) {
			}
			frame = new PRGUIFrame();
        	frame.setVisible(true); //necessary as of 1.3
        	desktop.add(frame);
        	try {
            	frame.setSelected(true);
        	} 
        	catch (java.beans.PropertyVetoException e) {
        	}
        	PRGUI.netListFileLocation = "";
        	PRGUI.critPathFileLocation = "";
        	PRGUI.view = 3;
        	PRGUI.showNetList = 0;
        	PRGUI.netName = "";
        	PRGUI.critName = "";
        	PRGUI.netNum = 0;
        }
                
    }
    
	public void newCritPath() throws IOException {
    	
    	//Declare All Variables Used in newCritPath
    	
    	String cpFileLoc, cpTemp, cpLine, pound = ("#");
    	int count=0, cpLength, countTemp;
    	
    	//Input CritPath File Information
    	
    	cpFileLoc = PRGUI.critPathFileLocation;
    	BufferedReader CP = new BufferedReader
			(new FileReader(cpFileLoc));
    	cpLine = CP.readLine();
    	
    	StringTokenizer tokenizer = new StringTokenizer(cpLine);
     	cpLength = Integer.parseInt(tokenizer.nextToken());
     	PRGUI.critNum = cpLength;
     	cpLine = CP.readLine();
     	
     	while(cpLine!=null && !cpLine.trim().equals("")) {
     		tokenizer = new StringTokenizer(cpLine);
     		cpTemp = tokenizer.nextToken();
     		countTemp = Integer.parseInt(tokenizer.nextToken());
     		if(countTemp > count) {
     			count = countTemp;
     		}
     		cpLine = CP.readLine();
     	}

     	count++;
     	count++;
     	CPArray.initArray(cpLength, count);
     	
     	CP = new BufferedReader(new FileReader(cpFileLoc));
     	cpLine = CP.readLine();
     	cpLine = CP.readLine();
     	tokenizer = new StringTokenizer(cpLine);
     	for(int i = 0; i < cpLength; i++) {
     		for(int j = 0; j < count; j++) {
     				if(tokenizer.countTokens()!=0) {
     					cpTemp = tokenizer.nextToken();
     					CPArray.AddCPArray(i, j, cpTemp);
     				}
     		}
     		cpLine = CP.readLine();
     		if(cpLine!=null) {
     			tokenizer = new StringTokenizer(cpLine);
     		}
     	}
     	CP = new BufferedReader(new FileReader(cpFileLoc));
     	cpLine = CP.readLine();
     	cpLine = CP.readLine();
     	tokenizer = new StringTokenizer(cpLine);
     	cpTemp = tokenizer.nextToken();
     	cpTemp = tokenizer.nextToken();
     	for(int i = 0; i < cpLength; i++) {
     		for(int j = 2; j < count; j++){
     			if(tokenizer.countTokens()!=0) {
     				cpTemp = tokenizer.nextToken();
     				cpTemp = pound + cpTemp;
     				CPArray.AddCPArray(i, j, cpTemp);
     			}
     		}
     	cpLine = CP.readLine();
     	if(cpLine!=null) {
     			tokenizer = new StringTokenizer(cpLine);
     			cpTemp = tokenizer.nextToken();
     			cpTemp = tokenizer.nextToken();
     		}
     	}
	
    }
  

	public void helpFrame() throws IOException {
		
		final JInternalFrame hf = new JInternalFrame();
		hf.setTitle("Help");
		hf.setSize(720, 540);
		hf.setLocation(100, 100);
		hf.setFrameIcon(new ImageIcon("icon.gif"));
		
		BufferedReader helpfile = new BufferedReader
			(new FileReader("help.txt"));
		
		String strTemp;
		String strOut = "";
		String newline = "\n";
		
        while ((strTemp = helpfile.readLine()) != null) {
         	strOut += (strTemp + newline);
       	}
       	
       	JTextArea helpLabel = new JTextArea(strOut);
       	JScrollPane helpScrollPane = new JScrollPane();
        helpScrollPane.getViewport().add(helpLabel);
       	hf.getContentPane().add(helpScrollPane);
       	
       	JButton exit = new JButton("OK");
		hf.getContentPane().add(exit, BorderLayout.SOUTH);
		ActionListener listener = new ActionListener() {
           	public void actionPerformed(ActionEvent evt) {
				try {
       	     		hf.setClosed(true);
        		} 
       			catch (java.beans.PropertyVetoException e) {
       			}
       	    }
       	};
		exit.addActionListener(listener);
		
		hf.setVisible(true); //necessary as of 1.3
        desktop.add(hf);
        hf.toFront();
       	
       	//objTextPane.setText(strOut);
		
		
		
		
	}  
	
	public void loadFiles() {
		
		final JInternalFrame lf = new JInternalFrame();
		lf.setTitle("Load Circuit Files...");
		lf.setSize(300, 275);
		lf.setLocation(100, 100);
		lf.setFrameIcon(new ImageIcon("icon.gif"));
		
		final JPanel buttonPanel = new JPanel();
		JButton ld_placement = new JButton("Load Placement File");
		final JLabel pf = new JLabel(" .p File Loaded:  " + PRGUI.pName);
		JButton ld_netlist = new JButton("Load NetList File");
		final JLabel nf = new JLabel(" .net File Loaded:  " + PRGUI.netName);
		JButton ld_netcrit = new JButton("Load Net Criticalities File");
		final JLabel nc = new JLabel(" .nc File Loaded:  "  );
		JButton ld_critpath = new JButton("Load Critical Paths File");
		final JLabel cp = new JLabel(" .kcp File Loaded:  " + PRGUI.critName);
		
		
		buttonPanel.setLayout(new BoxLayout(buttonPanel, BoxLayout.Y_AXIS));
		buttonPanel.add(Box.createRigidArea(new Dimension(0, 10)));
		ld_placement.setAlignmentX(Component.LEFT_ALIGNMENT);
		ActionListener placeFile = new ActionListener() {
           	public void actionPerformed(ActionEvent evt) {
       	     	JFileChooser fc = new JFileChooser();
     		   	int returnVal = fc.showOpenDialog(PRGUIDesktopPane.this);
        		if (returnVal == JFileChooser.APPROVE_OPTION) {
           			File file = fc.getSelectedFile();
           			PRGUI.placementFileLocation = file.getPath();
               		PRGUI.pName = file.getName();
               		PRGUI.showNetList=0;
           			PRGUI.netListFileLocation = null;
           			try {
       					newPlacement();	
       					PRGUI.LoadNum++;
       				}
       				catch (IOException exception) {
						System.out.println("The File Was Not Found");
					}
					lf.toFront();
               		pf.setText(" .p File Loaded:  " + PRGUI.pName);
               		nf.setText(" .net File Loaded:  " + PRGUI.netName);
               		cp.setText(" .kcp File Loaded:  " + PRGUI.critName);
           		}
           		
       	    }
       	};
       	ld_placement.addActionListener(placeFile);
		buttonPanel.add(ld_placement);
		buttonPanel.add(pf);
		buttonPanel.add(Box.createRigidArea(new Dimension(0, 10)));
		ld_netlist.setAlignmentX(Component.LEFT_ALIGNMENT);
		ActionListener netList = new ActionListener() {
           	public void actionPerformed(ActionEvent evt) {
       	     	JFileChooser fc2 = new JFileChooser();
        		int returnVal2 = fc2.showOpenDialog(PRGUIDesktopPane.this);
        		if (returnVal2 == JFileChooser.APPROVE_OPTION) {
       	        	File file2 = fc2.getSelectedFile();
                	PRGUI.netListFileLocation = file2.getPath();
                	PRGUI.netName = file2.getName();
            	}
            	try {
        			newNetList();	
        		}
        		catch (IOException exception) {
					System.out.println("The File Was Not Found");
				}
				nf.setText(" .net File Loaded:  " + PRGUI.netName);
       	    }
       	};
       	ld_netlist.addActionListener(netList);
       	buttonPanel.add(ld_netlist);
       	buttonPanel.add(nf);
       	buttonPanel.add(Box.createRigidArea(new Dimension(0, 10)));
       	ld_netcrit.setAlignmentX(Component.LEFT_ALIGNMENT);
       	ActionListener netCrit = new ActionListener() {
           	public void actionPerformed(ActionEvent evt) {
       	     	nc.setText(" .nc File Loaded:  " + PRGUI.netName);
       	    }
       	};
       	ld_netcrit.addActionListener(netCrit);
       	buttonPanel.add(ld_netcrit);
       	buttonPanel.add(nc);
       	buttonPanel.add(Box.createRigidArea(new Dimension(0, 10)));
       	ld_critpath.setAlignmentX(Component.LEFT_ALIGNMENT);
       	ActionListener critPath = new ActionListener() {
           	public void actionPerformed(ActionEvent evt) {
           		JFileChooser fc3 = new JFileChooser();
        		int returnVal3 = fc3.showOpenDialog(PRGUIDesktopPane.this);
        		if (returnVal3 == JFileChooser.APPROVE_OPTION) {
         	        File file3 = fc3.getSelectedFile();
          	        PRGUI.critPathFileLocation = file3.getPath();
            	    PRGUI.critName = file3.getName();
           		}
            	try {
        			newCritPath();	
        			PRGUI.LoadNum++;
        		}
        		catch (IOException exception) {
					System.out.println("The File Was Not Found");
				}
       	     	cp.setText(" .kcp File Loaded:  " + PRGUI.critName);
       	    }
       	};
       	ld_critpath.addActionListener(critPath);
       	buttonPanel.add(ld_critpath);
       	buttonPanel.add(cp);
       	
		lf.getContentPane().add(buttonPanel);
		
		JButton exit = new JButton("OK");
		lf.getContentPane().add(exit, BorderLayout.SOUTH);
		ActionListener listener = new ActionListener() {
           	public void actionPerformed(ActionEvent evt) {
				try {
       	     		lf.setClosed(true);
        		} 
       			catch (java.beans.PropertyVetoException e) {
       			}
       	    }
       	};
		exit.addActionListener(listener);
		
		lf.setVisible(true); //necessary as of 1.3
        desktop.add(lf);
        lf.toFront();


	}
}


class critPathFrame extends JInternalFrame {
	
	JPanel buttonPanel;
	ButtonGroup group;
	static JTextField input1, input2;
	
	public critPathFrame() {
		super("Choose to Display Critical Paths...", true, true, true, true);
		setSize(375, 200);
		setLocation(100, 100);
		setFrameIcon(new ImageIcon("icon.gif"));
		
		if(PRGUI.critName.equals("")){
			JOptionPane.showMessageDialog(this, 
			"You Need To Load A Critical Path File First!!");
		}
		else {
			Container contentPane = getContentPane();
			buttonPanel = new JPanel();
    		group = new ButtonGroup();
    		input1 = new JTextField("Insert Critical Path Number (Between 1 "
    			+ "And " + (PRGUI.critNum) + ")");
			input2 = new JTextField("Insert Critical Path Number (Between 1 "
    			+ "And " + (PRGUI.critNum) + ")");
		
    		addRadioButton("Show The ___ Most Critical Path", 1);
    		addRadioButton("Show The First ___ Most Critical Paths", 2);
    		addRadioButton("Hide Critical Paths", 0);
    	
    		contentPane.add(buttonPanel, BorderLayout.NORTH);
    	
    		JButton exit = new JButton("OK");
    	
    		ActionListener listener = new ActionListener() {
       	      	public void actionPerformed(ActionEvent evt) {
					try {
       	     			setClosed(true);
        			} 
        			catch (java.beans.PropertyVetoException e) {
        			}
         	    }
         	};
		
			exit.addActionListener(listener);
    	
    		contentPane.add(exit, BorderLayout.SOUTH);
    	}
	}
	
	public void addRadioButton(String name, final int paths) {
		
		JRadioButton button;
		if(paths==0) {
			button = new JRadioButton(name, true);
		}
		else{
			button = new JRadioButton(name);
		}
		group.add(button);
        buttonPanel.setLayout(new GridLayout(0, 1));
        buttonPanel.add(button);
        
        if(paths==1) {
     		buttonPanel.add(input1);
        }
        if(paths==2) {
     		buttonPanel.add(input2);
        }

        ActionListener listener = new ActionListener() {
             public void actionPerformed(ActionEvent evt) {
                PRGUI.NumCritPath = paths;
             }
          };

       button.addActionListener(listener);
    }
		
}

class aboutFrame extends JInternalFrame {
	
	public aboutFrame() {
		
		super("About PRGUI");
		setSize(250, 375);
		setLocation(100, 100);
		setFrameIcon(new ImageIcon("icon.gif"));
		
		Container contentPane = getContentPane();
		contentPane.setLayout(new BoxLayout(contentPane, BoxLayout.PAGE_AXIS));
		contentPane.add(Box.createRigidArea(new Dimension(0,10)));
		
		JLabel PRGUI = new JLabel("PRGUI");
		PRGUI.setAlignmentX(Component.CENTER_ALIGNMENT);
		contentPane.add(PRGUI);
		JLabel one = new JLabel("A visual tool to show the placement");
		one.setAlignmentX(Component.CENTER_ALIGNMENT);
		contentPane.add(one);
		JLabel two = new JLabel("and routing of a multiple layered");
		two.setAlignmentX(Component.CENTER_ALIGNMENT);
		contentPane.add(two);
		JLabel three = new JLabel("FPGA with input from VPR");
		three.setAlignmentX(Component.CENTER_ALIGNMENT);
		contentPane.add(three);
		contentPane.add(Box.createRigidArea(new Dimension(0,10)));
		JLabel four = new JLabel("Created by James Foltz");
		four.setAlignmentX(Component.CENTER_ALIGNMENT);
		contentPane.add(four);
		JLabel five = new JLabel("Summer 2004");
		five.setAlignmentX(Component.CENTER_ALIGNMENT);
		contentPane.add(five);
		JLabel six = new JLabel("University of Minnesota");
		six.setAlignmentX(Component.CENTER_ALIGNMENT);
		contentPane.add(six);
		JLabel seven = new JLabel("NSF REU Program");
		seven.setAlignmentX(Component.CENTER_ALIGNMENT);
		contentPane.add(seven);
		contentPane.add(Box.createRigidArea(new Dimension(0,10)));
		JLabel eight = new JLabel("jamesfoltz@hotmail.com");
		eight.setAlignmentX(Component.CENTER_ALIGNMENT);
		contentPane.add(eight);
		contentPane.add(Box.createRigidArea(new Dimension(0,10)));
		JLabel nine = new JLabel("A special thanks to");
		nine.setAlignmentX(Component.CENTER_ALIGNMENT);
		contentPane.add(nine);
		JLabel ten = new JLabel("Dr. Kia Bazargan,");
		ten.setAlignmentX(Component.CENTER_ALIGNMENT);
		contentPane.add(ten);
		JLabel eleven = new JLabel("Cristinel Ababei,");
		eleven.setAlignmentX(Component.CENTER_ALIGNMENT);
		contentPane.add(eleven);
		JLabel twelve = new JLabel("and Hushrav Mogal,");
		twelve.setAlignmentX(Component.CENTER_ALIGNMENT);
		contentPane.add(twelve);
		JLabel thirteen = new JLabel("for without their patient");
		thirteen.setAlignmentX(Component.CENTER_ALIGNMENT);
		contentPane.add(thirteen);
		JLabel fourteen = new JLabel("teachings and guidance,");
		fourteen.setAlignmentX(Component.CENTER_ALIGNMENT);
		contentPane.add(fourteen);
		JLabel fifteen = new JLabel("none of this would be possible");
		fifteen.setAlignmentX(Component.CENTER_ALIGNMENT);
		contentPane.add(fifteen);
		contentPane.add(Box.createRigidArea(new Dimension(0,10)));
		
	
		JButton exit = new JButton("OK");
    	
    	ActionListener listener = new ActionListener() {
            public void actionPerformed(ActionEvent evt) {
				try {
            		setClosed(true);
        		} 
        		catch (java.beans.PropertyVetoException e) {
        		}
             }
        };

		exit.addActionListener(listener);
		exit.setAlignmentX(Component.CENTER_ALIGNMENT);
		exit.setAlignmentY(Component.BOTTOM_ALIGNMENT);
    	
    	contentPane.add(exit);
	}
}

class netListFrame extends JInternalFrame {
	
	JPanel buttonPanel;
	ButtonGroup group;
	static JTextField input1, input2, input3, input4, input5;
	
	public netListFrame() {
		
		super("NetList Filtering And Viewing Options...", true, true, 
			true, true);
		setSize(375, 350);
		setLocation(100, 100);
		setFrameIcon(new ImageIcon("icon.gif"));

		if(PRGUI.netName.equals("")){
			JOptionPane.showMessageDialog(this, 
			"You Need To Load A NetList File First!!");
		}
		else {	
			Container contentPane = getContentPane();
			buttonPanel = new JPanel();
			input1 = new JTextField("Insert First Layer Number (Between 0 And " 
				+ (PRGUI.ZTot-1) + ")");
			input2 = new JTextField("Insert Second Layer Number (Between 0 And " 
				+ (PRGUI.ZTot-1) + ")");
			input3 = new JTextField("Insert First Layer Number (Between 0 And " 
				+ (PRGUI.ZTot-1) + ")");
			input4 = new JTextField("Insert Second Layer Number (Between 0 And " 
				+ (PRGUI.ZTot-1) + ")");
			input5 = new JTextField("Insert Number Of Layers To Span (Between"
				+ " 0 And " + (PRGUI.ZTot-1) + ")");
    		group = new ButtonGroup();
		
    		addRadioButton("Show NetList For Only Local Traffic", 1, 0);
   	 		addRadioButton("Show NetList For Only Outgoing Traffic", 1, 1);
    		addRadioButton("Show NetList For All Traffic", 1, 2);
    		addRadioButton("Show All Nets That Connect Layer#__ And Layer#__",
    			1, 3);
    		addRadioButton(
    			"Show All Nets That Are Between Layer#__ And Layer#__", 1, 4);
    		addRadioButton(
    			"Show All Nets That Span __ Layers", 1, 5);
    		addRadioButton("Hide NetList", 0, 0);
    	
    		contentPane.add(buttonPanel, BorderLayout.NORTH);
    	
    		JButton exit = new JButton("OK");
    	
    		ActionListener listener = new ActionListener() {
            	 public void actionPerformed(ActionEvent evt) {
					try {
            			setClosed(true);
            			PRGUI.NLF=0;
        			} 
        			catch (java.beans.PropertyVetoException e) {
        			}
             	}
          	};

			exit.addActionListener(listener);
    	
    		contentPane.add(exit, BorderLayout.SOUTH);
    	}
	}
	
	public void addRadioButton(String name, final int onOrOff, final int show) {
		
		JRadioButton button;
		if(onOrOff==0 && show==0) {
			button = new JRadioButton(name, true);
		}
		else{
			button = new JRadioButton(name);
		}
		group.add(button);
		
    	buttonPanel.setLayout(new GridLayout(0, 1));
        buttonPanel.add(button);
        
        if(show==3) {
     		buttonPanel.add(input1);
        	buttonPanel.add(input2);
        }
        if(show==4) {
     		buttonPanel.add(input3);
        	buttonPanel.add(input4);
        }
        if(show==5) {
        	buttonPanel.add(input5);
        }


        ActionListener listener = new ActionListener() {
             public void actionPerformed(ActionEvent evt) {
                PRGUI.NetListShow = show;
                PRGUI.showNetList = onOrOff;
             }
          };

       button.addActionListener(listener);
    }
		
}

class PRGUIFrame extends JInternalFrame {
	
    static final int xOffset = 30, yOffset = 30;
    public static PRGUIPanel panel;
    public static JScrollPane scrollPane;
    public static int layerNum;

    public PRGUIFrame() {
        super("PRGUI Window", 
              true, //resizable
              true, //closable
              true, //maximizable
              true);//iconifiable
        setSize(300, 300);
        setLocation(xOffset*layerNum, yOffset*layerNum);
        setFrameIcon(new ImageIcon("icon.gif"));
    	
        panel = new PRGUIPanel();
        panel.setBackground(Color.DARK_GRAY);
        panel.setPreferredSize(new Dimension(((PRGUI.XTot+2)*
        	(int)PRGUI.scaleFactor + (int)(PRGUI.scaleFactor*2.5)), 
        	(PRGUI.YTot*(int)PRGUI.scaleFactor*(PRGUI.ZTot) + 
        	(int)(PRGUI.scaleFactor*3.5))));
        
        scrollPane = new JScrollPane();
        scrollPane.getViewport().add(panel);
        
        Container contentPane = getContentPane();
		contentPane.add(scrollPane);
	}
}

class PRGUIPanel extends JPanel {
	
	public void paintComponent(Graphics g) {
		
		super.paintComponent(g);
		Graphics2D g2 = (Graphics2D)g;



	//This Section Draws EVERYTHING If 3D Mode is Selected!!
	if(PRGUI.view==3) {

		
		g2.setPaint(Color.WHITE);
		g2.drawString("Placement File:  " + PRGUI.pName, 0, 10);
		g2.drawString("NetList File:  " + PRGUI.netName, 0, 20);
		g2.drawString("Critical Paths File:  " + PRGUI.critName, 0, 30);
		g2.drawString("Total #Cells = " + NumberofComponents.recall(), 0, 50);
		if(PRGUI.netNum!=0) {
			g2.drawString("Total #Nets = " + PRGUI.netNum, 0, 60);
		}
		else {
			g2.drawString("Total #Nets = ", 0, 60);
		}
		g2.drawString("Total #Layers = " + PRGUI.ZTot, 0, 70);
		g2.drawString("All Layers Displayed", 0, 90);
		
		
		
		for(int i = 0; i < (PRGUI.XTot + 2); i++) {
			for(int j = 0; j < (PRGUI.YTot + 2); j++) {
				 for(int k = 0; k < (PRGUI.ZTot); k++) {
				Polygon poly = new Polygon();
				poly.addPoint((int)(PRGUI.xLeft + PRGUI.width + i*PRGUI.width +
					j*PRGUI.width), (int)(PRGUI.yTop + PRGUI.YTot * 
					PRGUI.scaleFactor/2 + PRGUI.width/2 - i*PRGUI.width/2 + 
					j*PRGUI.width/2 + k*PRGUI.width*(PRGUI.YTot+4)));
				poly.addPoint((int)(PRGUI.xLeft + PRGUI.width + i*PRGUI.width + 
					j*PRGUI.width + PRGUI.width/2), (int)(PRGUI.yTop + 
					PRGUI.YTot*PRGUI.scaleFactor/2 + PRGUI.width/4 + 
					PRGUI.width/2 - i*PRGUI.width/2 + j*PRGUI.width/2 + 
					k*PRGUI.width*(PRGUI.YTot+4)));
				poly.addPoint((int)(PRGUI.xLeft + PRGUI.width + i*PRGUI.width + 
					j*PRGUI.width + PRGUI.width), (int)(PRGUI.yTop + 
					PRGUI.YTot*PRGUI.scaleFactor/2 + PRGUI.width/2 - 
					i*PRGUI.width/2 + j*PRGUI.width/2 + 
					k*PRGUI.width*(PRGUI.YTot+4)));
				poly.addPoint((int)(PRGUI.xLeft + PRGUI.width + i*PRGUI.width + 
					j*PRGUI.width + PRGUI.width/2), (int)(PRGUI.yTop + 
					PRGUI.YTot*PRGUI.scaleFactor/2 - PRGUI.width/4 + 
					PRGUI.width/2 - i*PRGUI.width/2 + j*PRGUI.width/2 + 
					k*PRGUI.width*(PRGUI.YTot+4)));
				g2.setPaint(Color.LIGHT_GRAY);
				g2.fill(poly);
			}
			}	
		}


		//These Sections Referring to if(PRGUI.coordPos==) display the 
		//Coordinates on 3 corners of the 0 layer of the displayed
		//circuit		
		if(PRGUI.coordPos==1) {

			int tmp1=0, tmp2=PRGUI.YTot+2, tmp3=0;
			g2.setPaint(Color.WHITE);
			g2.drawString("(0,0)", (int)(PRGUI.xLeft + PRGUI.width + 
				tmp1*PRGUI.width + tmp2*PRGUI.width), (int)(PRGUI.yTop + 
				PRGUI.YTot * PRGUI.scaleFactor/2 + PRGUI.width/2 - 
				tmp1*PRGUI.width/2 + tmp2*PRGUI.width/2 + 
				tmp3*PRGUI.width*(PRGUI.YTot+4)));
					
			tmp1=PRGUI.YTot+2;
			tmp2=0;
			tmp3=0;
			g2.setPaint(Color.WHITE);
			g2.drawString("(" + (PRGUI.YTot+1) + "," + (PRGUI.YTot+1) + ")", 
				(int)(PRGUI.xLeft + PRGUI.width + tmp1*PRGUI.width +
				tmp2*PRGUI.width), (int)(PRGUI.yTop + PRGUI.YTot * 
				PRGUI.scaleFactor/2 + PRGUI.width/2 - tmp1*PRGUI.width/2 + 
				tmp2*PRGUI.width/2 + tmp3*PRGUI.width*(PRGUI.YTot+4)));
					
			tmp1=0;
			tmp2=(-1);
			tmp3=0;
			g2.setPaint(Color.WHITE);
			g2.drawString("(0," + (PRGUI.YTot+1) + ")", 
				(int)(PRGUI.xLeft + PRGUI.width + tmp1*PRGUI.width +
				tmp2*PRGUI.width), (int)(PRGUI.yTop + PRGUI.YTot * 
				PRGUI.scaleFactor/2 + PRGUI.width/2 - tmp1*PRGUI.width/2 + 
				tmp2*PRGUI.width/2 + tmp3*PRGUI.width*(PRGUI.YTot+4)));
		}
		
		if(PRGUI.coordPos==2) {

			int tmp1=0, tmp2=PRGUI.YTot+2, tmp3=0;
			g2.setPaint(Color.WHITE);
			g2.drawString("(" + (PRGUI.YTot+1) + ",0)", 
				(int)(PRGUI.xLeft + PRGUI.width + 
				tmp1*PRGUI.width + tmp2*PRGUI.width), (int)(PRGUI.yTop + 
				PRGUI.YTot * PRGUI.scaleFactor/2 + PRGUI.width/2 - 
				tmp1*PRGUI.width/2 + tmp2*PRGUI.width/2 + 
				tmp3*PRGUI.width*(PRGUI.YTot+4)));
					
			tmp1=PRGUI.YTot+2;
			tmp2=0;
			tmp3=0;
			g2.setPaint(Color.WHITE);
			g2.drawString("(0," + (PRGUI.YTot+1) + ")", 
				(int)(PRGUI.xLeft + PRGUI.width + tmp1*PRGUI.width +
				tmp2*PRGUI.width), (int)(PRGUI.yTop + PRGUI.YTot * 
				PRGUI.scaleFactor/2 + PRGUI.width/2 - tmp1*PRGUI.width/2 + 
				tmp2*PRGUI.width/2 + tmp3*PRGUI.width*(PRGUI.YTot+4)));
					
			tmp1=0;
			tmp2=(-1);
			tmp3=0;
			g2.setPaint(Color.WHITE);
			g2.drawString("(0,0)", 
				(int)(PRGUI.xLeft + PRGUI.width + tmp1*PRGUI.width +
				tmp2*PRGUI.width), (int)(PRGUI.yTop + PRGUI.YTot * 
				PRGUI.scaleFactor/2 + PRGUI.width/2 - tmp1*PRGUI.width/2 + 
				tmp2*PRGUI.width/2 + tmp3*PRGUI.width*(PRGUI.YTot+4)));
		}
		
		if(PRGUI.coordPos==3) {

			int tmp1=0, tmp2=PRGUI.YTot+2, tmp3=0;
			g2.setPaint(Color.WHITE);
			g2.drawString("(" + (PRGUI.YTot+1) + "," + (PRGUI.YTot+1) + ")",
				(int)(PRGUI.xLeft + PRGUI.width + 
				tmp1*PRGUI.width + tmp2*PRGUI.width), (int)(PRGUI.yTop + 
				PRGUI.YTot * PRGUI.scaleFactor/2 + PRGUI.width/2 - 
				tmp1*PRGUI.width/2 + tmp2*PRGUI.width/2 + 
				tmp3*PRGUI.width*(PRGUI.YTot+4)));
					
			tmp1=PRGUI.YTot+2;
			tmp2=0;
			tmp3=0;
			g2.setPaint(Color.WHITE);
			g2.drawString("(0,0)", 
				(int)(PRGUI.xLeft + PRGUI.width + tmp1*PRGUI.width +
				tmp2*PRGUI.width), (int)(PRGUI.yTop + PRGUI.YTot * 
				PRGUI.scaleFactor/2 + PRGUI.width/2 - tmp1*PRGUI.width/2 + 
				tmp2*PRGUI.width/2 + tmp3*PRGUI.width*(PRGUI.YTot+4)));
					
			tmp1=0;
			tmp2=(-1);
			tmp3=0;
			g2.setPaint(Color.WHITE);
			g2.drawString("(" + (PRGUI.YTot+1) + ",0)", 
				(int)(PRGUI.xLeft + PRGUI.width + tmp1*PRGUI.width +
				tmp2*PRGUI.width), (int)(PRGUI.yTop + PRGUI.YTot * 
				PRGUI.scaleFactor/2 + PRGUI.width/2 - tmp1*PRGUI.width/2 + 
				tmp2*PRGUI.width/2 + tmp3*PRGUI.width*(PRGUI.YTot+4)));
		}
		
		if(PRGUI.coordPos==4) {

			int tmp1=0, tmp2=PRGUI.YTot+2, tmp3=0;
			g2.setPaint(Color.WHITE);
			g2.drawString("(0," + (PRGUI.YTot+1) + ")", 
				(int)(PRGUI.xLeft + PRGUI.width + 
				tmp1*PRGUI.width + tmp2*PRGUI.width), (int)(PRGUI.yTop + 
				PRGUI.YTot * PRGUI.scaleFactor/2 + PRGUI.width/2 - 
				tmp1*PRGUI.width/2 + tmp2*PRGUI.width/2 + 
				tmp3*PRGUI.width*(PRGUI.YTot+4)));
					
			tmp1=PRGUI.YTot+2;
			tmp2=0;
			tmp3=0;
			g2.setPaint(Color.WHITE);
			g2.drawString("(" + (PRGUI.YTot+1) + ",0)", 
				(int)(PRGUI.xLeft + PRGUI.width + tmp1*PRGUI.width +
				tmp2*PRGUI.width), (int)(PRGUI.yTop + PRGUI.YTot * 
				PRGUI.scaleFactor/2 + PRGUI.width/2 - tmp1*PRGUI.width/2 + 
				tmp2*PRGUI.width/2 + tmp3*PRGUI.width*(PRGUI.YTot+4)));
					
			tmp1=0;
			tmp2=(-1);
			tmp3=0;
			g2.setPaint(Color.WHITE);
			g2.drawString("(" + (PRGUI.YTot+1) + "," + (PRGUI.YTot+1) + ")",
				(int)(PRGUI.xLeft + PRGUI.width + tmp1*PRGUI.width +
				tmp2*PRGUI.width), (int)(PRGUI.yTop + PRGUI.YTot * 
				PRGUI.scaleFactor/2 + PRGUI.width/2 - tmp1*PRGUI.width/2 + 
				tmp2*PRGUI.width/2 + tmp3*PRGUI.width*(PRGUI.YTot+4)));
		}
		
		
		
		
		
		
		for(int i = 0; i < (PRGUI.ZTot); i++) {
			g2.setPaint(Color.WHITE);
			g2.drawString("Layer #" + i, (int)(PRGUI.xLeft + PRGUI.width + 
				(PRGUI.XTot+2)*PRGUI.width + (PRGUI.YTot+2)*PRGUI.width + 
				PRGUI.width), (int)(PRGUI.yTop + PRGUI.YTot*PRGUI.scaleFactor/2 
				+ PRGUI.width/2 - (PRGUI.XTot+2)*PRGUI.width/2 + 
				(PRGUI.YTot+2)*PRGUI.width/2 + i*PRGUI.width*(PRGUI.YTot+4)));
		}
		

		for(int i = 0; i < NumberofComponents.recall(); i++) {
			double xVal = XHolder.GetX(i);
			double yVal = YHolder.GetY(i);
			double zVal = ZHolder.GetZ(i);
			String value = BNHolder.GetBN(i);
				Polygon poly = new Polygon();
				poly.addPoint((int)(PRGUI.xLeft + xVal*PRGUI.width + 
					yVal*PRGUI.width), (int)(PRGUI.yTop + 
					PRGUI.YTot*PRGUI.scaleFactor/2 - xVal*PRGUI.width/2 + 
					yVal*PRGUI.width/2 + zVal*PRGUI.width*(PRGUI.YTot+4)));
				poly.addPoint((int)(PRGUI.xLeft + xVal*PRGUI.width + 
					yVal*PRGUI.width + PRGUI.width/2), (int)(PRGUI.yTop + 
					PRGUI.YTot*PRGUI.scaleFactor/2 + PRGUI.width/4 - 
					xVal*PRGUI.width/2 + yVal*PRGUI.width/2 + 
					zVal*PRGUI.width*(PRGUI.YTot+4)));
				poly.addPoint((int)(PRGUI.xLeft + xVal*PRGUI.width + 
					yVal*PRGUI.width + PRGUI.width), (int)(PRGUI.yTop + 
					PRGUI.YTot*PRGUI.scaleFactor/2 - xVal*PRGUI.width/2 + 
					yVal*PRGUI.width/2 + zVal*PRGUI.width*(PRGUI.YTot+4)));
				poly.addPoint((int)(PRGUI.xLeft + xVal*PRGUI.width + 
					yVal*PRGUI.width + PRGUI.width/2), (int)(PRGUI.yTop + 
					PRGUI.YTot*PRGUI.scaleFactor/2 - PRGUI.width/4 - 
					xVal*PRGUI.width/2 + yVal*PRGUI.width/2 + 
					zVal*PRGUI.width*(PRGUI.YTot+4)));
				g2.setPaint(Color.WHITE);
				g2.fill(poly);
				if(PRGUI.scaleFactor >= 150) {
					g2.setPaint(Color.BLACK);
					g2.drawString(value, (int)(PRGUI.xLeft + xVal*PRGUI.width + 
						yVal*PRGUI.width + 20), (int)(PRGUI.yTop + 
						PRGUI.YTot*PRGUI.scaleFactor/2 - xVal*PRGUI.width/2 + 
						yVal*PRGUI.width/2 + zVal*PRGUI.width*(PRGUI.YTot+4)));
				}
		}
		
		int i, ZStart, ZEnd;
		double startX, startY, XHalf, YHalf, endX, endY;
		String temp;
		String temp2 = new String("");
		
		//This Section Displays The NetList
		if(PRGUI.showNetList==1) {
			
			for(int j = 0; j < BNSource.BNSource.length; j++) {
				temp = BNSource.GetBNSource(j);
				for(i = 0; i < NumberofComponents.recall() && 
					temp.compareTo(temp2)!=0; i++) {
						temp2 = BNameHolder.GetBName(i);
				}
				
				if(i == 0){
					i = NumberofComponents.recall();
				}

				ZStart = ZHolder.GetZ(i - 1);
				startX = PRGUI.xLeft + (XHolder.GetX(i - 1))*PRGUI.width + 
					(YHolder.GetY(i - 1))*PRGUI.width + PRGUI.width/2;
				startY = PRGUI.yTop + PRGUI.YTot*PRGUI.scaleFactor/2 - 
					(XHolder.GetX(i - 1))*PRGUI.width/2 + 
					(YHolder.GetY(i - 1))*PRGUI.width/2 + 
					(ZHolder.GetZ(i - 1))*PRGUI.width*(PRGUI.YTot+4);
				
				

				for(int k = 0; k < PRGUI.NetListCon; k++) {
					temp = BNSink.GetBNSink(j, k);
					for(i = 0; i < NumberofComponents.recall() &&
						temp.compareTo(temp2)!=0; i++) {
							temp2 = BNameHolder.GetBName(i);
						}

					if(temp.compareTo("open")!=0) {
						
					if(i == 0){
						i = NumberofComponents.recall();
					}

					ZEnd = ZHolder.GetZ(i - 1);
					endX = PRGUI.xLeft + (XHolder.GetX(i - 1))*PRGUI.width + 
						(YHolder.GetY(i - 1))*PRGUI.width + PRGUI.width/2;
					endY = PRGUI.yTop + PRGUI.YTot*PRGUI.scaleFactor/2 - 
						(XHolder.GetX(i - 1))*PRGUI.width/2 + 
						(YHolder.GetY(i - 1))*PRGUI.width/2 + 
						(ZHolder.GetZ(i - 1))*PRGUI.width*(PRGUI.YTot+4);
					
					//This Section Displays NetList If Show Local Traffic
					//Or Show All Traffic Is Selected
					if(ZStart == ZEnd 
						&& (PRGUI.NetListShow == 0 || PRGUI.NetListShow == 2)) {
						Line2D line = new 
							Line2D.Double(startX, startY, endX, endY);
						g2.setPaint(Color.CYAN);
						g2.draw(line);
					}//End Show Local Traffic
					
					//This Section Displays NetList If Show OutGoing Traffic
					//Or Show All Traffic Is Selected
					else if(ZStart != ZEnd
						&& (PRGUI.NetListShow == 1 || PRGUI.NetListShow == 2)) {
						Line2D line2 = new 
							Line2D.Double(startX, startY, endX, endY);
						g2.setPaint(Color.YELLOW);
						g2.draw(line2);
					}//End Show Outgoing Traffic
					
					//This Section Displays NetList If Show Connections That
					//Connect 2 Specific Layers is Selected
					else if(PRGUI.NetListShow == 3 && PRGUI.NLF == 0) {
						PRGUI.input1 = Integer.parseInt(
							netListFrame.input1.getText().trim());
						PRGUI.input2 = Integer.parseInt(
							netListFrame.input2.getText().trim());
						int val1 = PRGUI.input1;
						int val2 = PRGUI.input2;
						if(val1==ZStart && val2==ZEnd) {
							Line2D line2 = new 
							Line2D.Double(startX, startY, endX, endY);
							g2.setPaint(Color.YELLOW);
							g2.draw(line2);
						}
						else if(val2==ZStart && val1==ZEnd) {
							Line2D line2 = new 
							Line2D.Double(startX, startY, endX, endY);
							g2.setPaint(Color.GREEN);
							g2.draw(line2);
						}
					}//End Show Connections between 2 Specific Layers
					
					//This Section Displays NetList If Show Connections
					//Between 2 Layers Is Selected
					else if(PRGUI.NetListShow == 4 && PRGUI.NLF == 0) {
						PRGUI.input1 = Integer.parseInt(
							netListFrame.input3.getText().trim());
						PRGUI.input2 = Integer.parseInt(
							netListFrame.input4.getText().trim());
						int val1 = PRGUI.input1;
						int val2 = PRGUI.input2;
						if(val1 > val2) {
							if(ZStart < val1 && ZStart > val2 && 
								ZEnd < val1 && ZEnd > val2) {
									Line2D line2 = new 
									Line2D.Double(startX, startY, endX, endY);
									g2.setPaint(Color.CYAN);
									g2.draw(line2);
							}
						}
						else if(val2 > val1) {
							if(ZStart < val2 && ZStart > val1 && 
								ZEnd < val2 && ZEnd > val1) {
									Line2D line2 = new 
									Line2D.Double(startX, startY, endX, endY);
									g2.setPaint(Color.CYAN);
									g2.draw(line2);
							}
						}
					}//End Show Connections Between 2 Layers
					
					//This Section Displays NetList If Show Connections
					//That Span X Layers is Selected
					else if(PRGUI.NetListShow == 5 && PRGUI.NLF == 0) {
						PRGUI.input1 = Integer.parseInt(
							netListFrame.input5.getText().trim());
						int val1 = PRGUI.input1;
						if((ZStart - ZEnd) == val1) {
							Line2D line2 = new 
								Line2D.Double(startX, startY, endX, endY);
							g2.setPaint(Color.CYAN);
							g2.draw(line2);
						}
						else if((ZEnd - ZStart) == val1) {
							Line2D line2 = new 
								Line2D.Double(startX, startY, endX, endY);
							g2.setPaint(Color.CYAN);
							g2.draw(line2);
						}
					}//End Show Connections That Span X Layers
					
					}
					
				}
			}
		}//End Of NetList Display
		
		//This Section Displays Critical Paths If Show The Kth Most
		//Critical Path is Selected
		String CPTemp, CPTemp2 = "", CPTemp3, CPTemp4 = "";
		int count;
		double cpStartX, cpStartY, cpStartZ, cpEndX, cpEndY, cpEndZ;
		if(PRGUI.NumCritPath==1) {
			PRGUI.input1 = Integer.parseInt(
							critPathFrame.input1.getText().trim());
			if(PRGUI.input1==1) {
				int numofCP = Integer.parseInt(CPArray.GetCPArray(0, 1));
				for(int j = 2; j < (numofCP+1); j++) {
					CPTemp = CPArray.GetCPArray(0, j);
					for(count = 0; count < NumberofComponents.recall() &&
						CPTemp.compareTo(CPTemp2)!=0; count++) {
							CPTemp2 = BNHolder.GetBN(count);
					}
					if(count == 0) {
						count = NumberofComponents.recall();
					}
					cpStartX = PRGUI.xLeft + (XHolder.GetX(count - 1))*
						PRGUI.width + (YHolder.GetY(count - 1))*PRGUI.width + 
						PRGUI.width/2;
					cpStartZ = ZHolder.GetZ(count - 1);
					cpStartY = PRGUI.yTop + PRGUI.YTot*PRGUI.scaleFactor/2 - 
						(XHolder.GetX(count - 1))*PRGUI.width/2 + 
						(YHolder.GetY(count - 1))*PRGUI.width/2 + 
						(ZHolder.GetZ(count - 1))*PRGUI.width*(PRGUI.YTot+4);
						
					Polygon critPoly = new Polygon();
					critPoly.addPoint((int)(PRGUI.xLeft + 
						(XHolder.GetX(count-1))*PRGUI.width + 
						(YHolder.GetY(count-1))*PRGUI.width), (int)(PRGUI.yTop + 
						PRGUI.YTot*PRGUI.scaleFactor/2 - 
						(XHolder.GetX(count-1))*PRGUI.width/2 + 
						(YHolder.GetY(count-1))*PRGUI.width/2 + 
						(ZHolder.GetZ(count-1))*PRGUI.width*(PRGUI.YTot+4)));
					critPoly.addPoint((int)(PRGUI.xLeft + 
						(XHolder.GetX(count-1))*PRGUI.width + 
						(YHolder.GetY(count-1))*PRGUI.width + PRGUI.width/2), 
						(int)(PRGUI.yTop + 	PRGUI.YTot*PRGUI.scaleFactor/2 + 
						PRGUI.width/4 - (XHolder.GetX(count-1))*PRGUI.width/2 + 
						(YHolder.GetY(count-1))*PRGUI.width/2 + 
						(ZHolder.GetZ(count-1))*PRGUI.width*(PRGUI.YTot+4)));
					critPoly.addPoint((int)(PRGUI.xLeft + 
						(XHolder.GetX(count-1))*PRGUI.width + 
						(YHolder.GetY(count-1))*PRGUI.width + PRGUI.width), 
						(int)(PRGUI.yTop + PRGUI.YTot*PRGUI.scaleFactor/2 - 
						(XHolder.GetX(count-1))*PRGUI.width/2 + 
						(YHolder.GetY(count-1))*PRGUI.width/2 + 
						(ZHolder.GetZ(count-1))*PRGUI.width*(PRGUI.YTot+4)));
					critPoly.addPoint((int)(PRGUI.xLeft + 
						(XHolder.GetX(count-1))*PRGUI.width + 
						(YHolder.GetY(count-1))*PRGUI.width + PRGUI.width/2), 
						(int)(PRGUI.yTop + PRGUI.YTot*PRGUI.scaleFactor/2 - 
						PRGUI.width/4 - (XHolder.GetX(count-1))*PRGUI.width/2 + 
						(YHolder.GetY(count-1))*PRGUI.width/2 + 
						(ZHolder.GetZ(count-1))*PRGUI.width*(PRGUI.YTot+4)));
					g2.setPaint(Color.RED);
					g2.setStroke(new BasicStroke(3));
					g2.draw(critPoly);	
						
					if(j!=(numofCP+1)) {
						CPTemp3 = CPArray.GetCPArray(0, j + 1);
						for(count = 0; count < NumberofComponents.recall() &&
							CPTemp3.compareTo(CPTemp4)!=0; count++) {
								CPTemp4 = BNHolder.GetBN(count);
							}
						cpEndX = PRGUI.xLeft + (XHolder.GetX(count - 1))*
							PRGUI.width + (YHolder.GetY(count - 1))*PRGUI.width
							+ PRGUI.width/2;
						cpEndY = PRGUI.yTop + PRGUI.YTot*PRGUI.scaleFactor/2 - 
							(XHolder.GetX(count - 1))*PRGUI.width/2 + 
							(YHolder.GetY(count - 1))*PRGUI.width/2 + 
							(ZHolder.GetZ(count - 1))*PRGUI.width*
							(PRGUI.YTot+4);
							
						Line2D line = new 
							Line2D.Double(cpStartX, cpStartY, cpEndX, cpEndY);
						g2.setPaint(Color.RED);
						g2.setStroke(new BasicStroke(3));
						g2.draw(line);
						
						Polygon critPoly2 = new Polygon();
						critPoly2.addPoint((int)(PRGUI.xLeft + 
								(XHolder.GetX(count-1))*PRGUI.width + 
							(YHolder.GetY(count-1))*PRGUI.width), 
							(int)(PRGUI.yTop + PRGUI.YTot*PRGUI.scaleFactor/2 - 
							(XHolder.GetX(count-1))*PRGUI.width/2 + 
							(YHolder.GetY(count-1))*PRGUI.width/2 + 
							(ZHolder.GetZ(count-1))*
							PRGUI.width*(PRGUI.YTot+4)));
						critPoly2.addPoint((int)(PRGUI.xLeft + 
							(XHolder.GetX(count-1))*PRGUI.width + 
							(YHolder.GetY(count-1))*PRGUI.width + 
							PRGUI.width/2), (int)(PRGUI.yTop + 	
							PRGUI.YTot*PRGUI.scaleFactor/2 + PRGUI.width/4 - 
							(XHolder.GetX(count-1))*PRGUI.width/2 + 
							(YHolder.GetY(count-1))*PRGUI.width/2 + 
							(ZHolder.GetZ(count-1))*PRGUI.width*
							(PRGUI.YTot+4)));
						critPoly2.addPoint((int)(PRGUI.xLeft + 
							(XHolder.GetX(count-1))*PRGUI.width + 
							(YHolder.GetY(count-1))*PRGUI.width + PRGUI.width), 
							(int)(PRGUI.yTop + PRGUI.YTot*PRGUI.scaleFactor/2 - 
							(XHolder.GetX(count-1))*PRGUI.width/2 + 
							(YHolder.GetY(count-1))*PRGUI.width/2 + 
							(ZHolder.GetZ(count-1))*PRGUI.width*
							(PRGUI.YTot+4)));
						critPoly2.addPoint((int)(PRGUI.xLeft + 
							(XHolder.GetX(count-1))*PRGUI.width + 
							(YHolder.GetY(count-1))*PRGUI.width + 
							PRGUI.width/2), (int)(PRGUI.yTop + 
							PRGUI.YTot*PRGUI.scaleFactor/2 - PRGUI.width/4 - 
							(XHolder.GetX(count-1))*PRGUI.width/2 + 
							(YHolder.GetY(count-1))*PRGUI.width/2 + 
							(ZHolder.GetZ(count-1))*PRGUI.width*
							(PRGUI.YTot+4)));
						g2.setPaint(Color.RED);
						g2.setStroke(new BasicStroke(3));
						g2.draw(critPoly2);
					}
					
					
				}
			}//End if 1-most critical path is shown
			else {
				int numofCP = Integer.parseInt(
					CPArray.GetCPArray((PRGUI.input1-1), 1));
				for(int j = 2; j < (numofCP+1); j++) {
					CPTemp = CPArray.GetCPArray((PRGUI.input1-1), j);
					for(count = 0; count < NumberofComponents.recall() &&
						CPTemp.compareTo(CPTemp2)!=0; count++) {
							CPTemp2 = BNHolder.GetBN(count);
					}
					if(count == 0) {
						count = NumberofComponents.recall();
					}
					cpStartX = PRGUI.xLeft + (XHolder.GetX(count - 1))*
						PRGUI.width + (YHolder.GetY(count - 1))*PRGUI.width + 
						PRGUI.width/2;
					cpStartZ = ZHolder.GetZ(count - 1);
					cpStartY = PRGUI.yTop + PRGUI.YTot*PRGUI.scaleFactor/2 - 
						(XHolder.GetX(count - 1))*PRGUI.width/2 + 
						(YHolder.GetY(count - 1))*PRGUI.width/2 + 
						(ZHolder.GetZ(count - 1))*PRGUI.width*(PRGUI.YTot+4);
						
					Polygon critPoly = new Polygon();
					critPoly.addPoint((int)(PRGUI.xLeft + 
						(XHolder.GetX(count-1))*PRGUI.width + 
						(YHolder.GetY(count-1))*PRGUI.width), (int)(PRGUI.yTop + 
						PRGUI.YTot*PRGUI.scaleFactor/2 - 
						(XHolder.GetX(count-1))*PRGUI.width/2 + 
						(YHolder.GetY(count-1))*PRGUI.width/2 + 
						(ZHolder.GetZ(count-1))*PRGUI.width*(PRGUI.YTot+4)));
					critPoly.addPoint((int)(PRGUI.xLeft + 
						(XHolder.GetX(count-1))*PRGUI.width + 
						(YHolder.GetY(count-1))*PRGUI.width + PRGUI.width/2), 
						(int)(PRGUI.yTop + 	PRGUI.YTot*PRGUI.scaleFactor/2 + 
						PRGUI.width/4 - (XHolder.GetX(count-1))*PRGUI.width/2 + 
						(YHolder.GetY(count-1))*PRGUI.width/2 + 
						(ZHolder.GetZ(count-1))*PRGUI.width*(PRGUI.YTot+4)));
					critPoly.addPoint((int)(PRGUI.xLeft + 
						(XHolder.GetX(count-1))*PRGUI.width + 
						(YHolder.GetY(count-1))*PRGUI.width + PRGUI.width), 
						(int)(PRGUI.yTop + PRGUI.YTot*PRGUI.scaleFactor/2 - 
						(XHolder.GetX(count-1))*PRGUI.width/2 + 
						(YHolder.GetY(count-1))*PRGUI.width/2 + 
						(ZHolder.GetZ(count-1))*PRGUI.width*(PRGUI.YTot+4)));
					critPoly.addPoint((int)(PRGUI.xLeft + 
						(XHolder.GetX(count-1))*PRGUI.width + 
						(YHolder.GetY(count-1))*PRGUI.width + PRGUI.width/2), 
						(int)(PRGUI.yTop + PRGUI.YTot*PRGUI.scaleFactor/2 - 
						PRGUI.width/4 - (XHolder.GetX(count-1))*PRGUI.width/2 + 
						(YHolder.GetY(count-1))*PRGUI.width/2 + 
						(ZHolder.GetZ(count-1))*PRGUI.width*(PRGUI.YTot+4)));
					g2.setPaint(Color.RED);
					g2.setStroke(new BasicStroke(1));
					g2.draw(critPoly);
						
					if(j!=(numofCP+1)) {
						CPTemp3 = CPArray.GetCPArray((PRGUI.input1-1), j + 1);
						for(count = 0; count < NumberofComponents.recall() &&
							CPTemp3.compareTo(CPTemp4)!=0; count++) {
								CPTemp4 = BNHolder.GetBN(count);
							}
						cpEndX = PRGUI.xLeft + (XHolder.GetX(count - 1))*
							PRGUI.width + (YHolder.GetY(count - 1))*PRGUI.width
							+ PRGUI.width/2;
						cpEndY = PRGUI.yTop + PRGUI.YTot*PRGUI.scaleFactor/2 - 
							(XHolder.GetX(count - 1))*PRGUI.width/2 + 
							(YHolder.GetY(count - 1))*PRGUI.width/2 + 
							(ZHolder.GetZ(count - 1))*PRGUI.width*
							(PRGUI.YTot+4);
							
						Line2D line = new 
							Line2D.Double(cpStartX, cpStartY, cpEndX, cpEndY);
						g2.setPaint(Color.RED);
						g2.setStroke(new BasicStroke(1));
						g2.draw(line);
						
						Polygon critPoly2 = new Polygon();
						critPoly2.addPoint((int)(PRGUI.xLeft + 
								(XHolder.GetX(count-1))*PRGUI.width + 
							(YHolder.GetY(count-1))*PRGUI.width), 
							(int)(PRGUI.yTop + PRGUI.YTot*PRGUI.scaleFactor/2 - 
							(XHolder.GetX(count-1))*PRGUI.width/2 + 
							(YHolder.GetY(count-1))*PRGUI.width/2 + 
							(ZHolder.GetZ(count-1))*
							PRGUI.width*(PRGUI.YTot+4)));
						critPoly2.addPoint((int)(PRGUI.xLeft + 
							(XHolder.GetX(count-1))*PRGUI.width + 
							(YHolder.GetY(count-1))*PRGUI.width + 
							PRGUI.width/2), (int)(PRGUI.yTop + 	
							PRGUI.YTot*PRGUI.scaleFactor/2 + PRGUI.width/4 - 
							(XHolder.GetX(count-1))*PRGUI.width/2 + 
							(YHolder.GetY(count-1))*PRGUI.width/2 + 
							(ZHolder.GetZ(count-1))*PRGUI.width*
							(PRGUI.YTot+4)));
						critPoly2.addPoint((int)(PRGUI.xLeft + 
							(XHolder.GetX(count-1))*PRGUI.width + 
							(YHolder.GetY(count-1))*PRGUI.width + PRGUI.width), 
							(int)(PRGUI.yTop + PRGUI.YTot*PRGUI.scaleFactor/2 - 
							(XHolder.GetX(count-1))*PRGUI.width/2 + 
							(YHolder.GetY(count-1))*PRGUI.width/2 + 
							(ZHolder.GetZ(count-1))*PRGUI.width*
							(PRGUI.YTot+4)));
						critPoly2.addPoint((int)(PRGUI.xLeft + 
							(XHolder.GetX(count-1))*PRGUI.width + 
							(YHolder.GetY(count-1))*PRGUI.width + 
							PRGUI.width/2), (int)(PRGUI.yTop + 
							PRGUI.YTot*PRGUI.scaleFactor/2 - PRGUI.width/4 - 
							(XHolder.GetX(count-1))*PRGUI.width/2 + 
							(YHolder.GetY(count-1))*PRGUI.width/2 + 
							(ZHolder.GetZ(count-1))*PRGUI.width*
							(PRGUI.YTot+4)));
						g2.setPaint(Color.RED);
						g2.setStroke(new BasicStroke(1));
						g2.draw(critPoly2);
					}
				}
			}//End if 1-most critical path is shown
		}//End Show Kth-Most Critical Path
		
		//This Section Shows Critical Paths if show first K-Most Paths
		//Is Selected
		if(PRGUI.NumCritPath==2) {
			PRGUI.input1 = Integer.parseInt(
							critPathFrame.input2.getText().trim());
			for(int u = 0; u < PRGUI.input1; u++) {
				int numofCP = Integer.parseInt(
					CPArray.GetCPArray((u), 1));
				for(int j = 2; j < (numofCP+1); j++) {
					CPTemp = CPArray.GetCPArray((u), j);
					for(count = 0; count < NumberofComponents.recall() &&
						CPTemp.compareTo(CPTemp2)!=0; count++) {
							CPTemp2 = BNHolder.GetBN(count);
					}
					if(count == 0) {
						count = NumberofComponents.recall();
					}
					cpStartX = PRGUI.xLeft + (XHolder.GetX(count - 1))*
						PRGUI.width + (YHolder.GetY(count - 1))*PRGUI.width + 
						PRGUI.width/2;
					cpStartZ = ZHolder.GetZ(count - 1);
					cpStartY = PRGUI.yTop + PRGUI.YTot*PRGUI.scaleFactor/2 - 
						(XHolder.GetX(count - 1))*PRGUI.width/2 + 
						(YHolder.GetY(count - 1))*PRGUI.width/2 + 
						(ZHolder.GetZ(count - 1))*PRGUI.width*(PRGUI.YTot+4);
						
					Polygon critPoly = new Polygon();
					critPoly.addPoint((int)(PRGUI.xLeft + 
						(XHolder.GetX(count-1))*PRGUI.width + 
						(YHolder.GetY(count-1))*PRGUI.width), (int)(PRGUI.yTop + 
						PRGUI.YTot*PRGUI.scaleFactor/2 - 
						(XHolder.GetX(count-1))*PRGUI.width/2 + 
						(YHolder.GetY(count-1))*PRGUI.width/2 + 
						(ZHolder.GetZ(count-1))*PRGUI.width*(PRGUI.YTot+4)));
					critPoly.addPoint((int)(PRGUI.xLeft + 
						(XHolder.GetX(count-1))*PRGUI.width + 
						(YHolder.GetY(count-1))*PRGUI.width + PRGUI.width/2), 
						(int)(PRGUI.yTop + 	PRGUI.YTot*PRGUI.scaleFactor/2 + 
						PRGUI.width/4 - (XHolder.GetX(count-1))*PRGUI.width/2 + 
						(YHolder.GetY(count-1))*PRGUI.width/2 + 
						(ZHolder.GetZ(count-1))*PRGUI.width*(PRGUI.YTot+4)));
					critPoly.addPoint((int)(PRGUI.xLeft + 
						(XHolder.GetX(count-1))*PRGUI.width + 
						(YHolder.GetY(count-1))*PRGUI.width + PRGUI.width), 
						(int)(PRGUI.yTop + PRGUI.YTot*PRGUI.scaleFactor/2 - 
						(XHolder.GetX(count-1))*PRGUI.width/2 + 
						(YHolder.GetY(count-1))*PRGUI.width/2 + 
						(ZHolder.GetZ(count-1))*PRGUI.width*(PRGUI.YTot+4)));
					critPoly.addPoint((int)(PRGUI.xLeft + 
						(XHolder.GetX(count-1))*PRGUI.width + 
						(YHolder.GetY(count-1))*PRGUI.width + PRGUI.width/2), 
						(int)(PRGUI.yTop + PRGUI.YTot*PRGUI.scaleFactor/2 - 
						PRGUI.width/4 - (XHolder.GetX(count-1))*PRGUI.width/2 + 
						(YHolder.GetY(count-1))*PRGUI.width/2 + 
						(ZHolder.GetZ(count-1))*PRGUI.width*(PRGUI.YTot+4)));
					g2.setPaint(Color.RED);
					g2.setStroke(new BasicStroke(1));
					g2.draw(critPoly);
						
					if(j!=(numofCP+1)) {
						CPTemp3 = CPArray.GetCPArray((u), j + 1);
						for(count = 0; count < NumberofComponents.recall() &&
							CPTemp3.compareTo(CPTemp4)!=0; count++) {
								CPTemp4 = BNHolder.GetBN(count);
							}
						cpEndX = PRGUI.xLeft + (XHolder.GetX(count - 1))*
							PRGUI.width + (YHolder.GetY(count - 1))*PRGUI.width
							+ PRGUI.width/2;
						cpEndY = PRGUI.yTop + PRGUI.YTot*PRGUI.scaleFactor/2 - 
							(XHolder.GetX(count - 1))*PRGUI.width/2 + 
							(YHolder.GetY(count - 1))*PRGUI.width/2 + 
							(ZHolder.GetZ(count - 1))*PRGUI.width*
							(PRGUI.YTot+4);
							
						Line2D line = new 
							Line2D.Double(cpStartX, cpStartY, cpEndX, cpEndY);
						g2.setPaint(Color.RED);
						g2.setStroke(new BasicStroke(1));
						g2.draw(line);
						
						Polygon critPoly2 = new Polygon();
						critPoly2.addPoint((int)(PRGUI.xLeft + 
								(XHolder.GetX(count-1))*PRGUI.width + 
							(YHolder.GetY(count-1))*PRGUI.width), 
							(int)(PRGUI.yTop + PRGUI.YTot*PRGUI.scaleFactor/2 - 
							(XHolder.GetX(count-1))*PRGUI.width/2 + 
							(YHolder.GetY(count-1))*PRGUI.width/2 + 
							(ZHolder.GetZ(count-1))*
							PRGUI.width*(PRGUI.YTot+4)));
						critPoly2.addPoint((int)(PRGUI.xLeft + 
							(XHolder.GetX(count-1))*PRGUI.width + 
							(YHolder.GetY(count-1))*PRGUI.width + 
							PRGUI.width/2), (int)(PRGUI.yTop + 	
							PRGUI.YTot*PRGUI.scaleFactor/2 + PRGUI.width/4 - 
							(XHolder.GetX(count-1))*PRGUI.width/2 + 
							(YHolder.GetY(count-1))*PRGUI.width/2 + 
							(ZHolder.GetZ(count-1))*PRGUI.width*
							(PRGUI.YTot+4)));
						critPoly2.addPoint((int)(PRGUI.xLeft + 
							(XHolder.GetX(count-1))*PRGUI.width + 
							(YHolder.GetY(count-1))*PRGUI.width + PRGUI.width), 
							(int)(PRGUI.yTop + PRGUI.YTot*PRGUI.scaleFactor/2 - 
							(XHolder.GetX(count-1))*PRGUI.width/2 + 
							(YHolder.GetY(count-1))*PRGUI.width/2 + 
							(ZHolder.GetZ(count-1))*PRGUI.width*
							(PRGUI.YTot+4)));
						critPoly2.addPoint((int)(PRGUI.xLeft + 
							(XHolder.GetX(count-1))*PRGUI.width + 
							(YHolder.GetY(count-1))*PRGUI.width + 
							PRGUI.width/2), (int)(PRGUI.yTop + 
							PRGUI.YTot*PRGUI.scaleFactor/2 - PRGUI.width/4 - 
							(XHolder.GetX(count-1))*PRGUI.width/2 + 
							(YHolder.GetY(count-1))*PRGUI.width/2 + 
							(ZHolder.GetZ(count-1))*PRGUI.width*
							(PRGUI.YTot+4)));
						g2.setPaint(Color.RED);
						g2.setStroke(new BasicStroke(1));
						g2.draw(critPoly2);
					}
				}
			}
		}//End Show First K-Most Critical Paths

	}//End 3D View Section


	//This section displays EVERYTHING if 2D mode is selected!!
	if(PRGUI.view==2) {
		
		g2.setPaint(Color.WHITE);
		g2.drawString("Placement File:  " + PRGUI.pName,
			(int)((PRGUI.XTot + 2)*PRGUI.scaleFactor + PRGUI.xLeft), 10);
		g2.drawString("NetList File:  " + PRGUI.netName,
			(int)((PRGUI.XTot + 2)*PRGUI.scaleFactor + PRGUI.xLeft), 20);
		g2.drawString("Critical Paths File:  " + PRGUI.critName,
			(int)((PRGUI.XTot + 2)*PRGUI.scaleFactor + PRGUI.xLeft), 30);
		g2.drawString("Layer #" + PRGUIFrame.layerNum + " Is Displayed",
			(int)((PRGUI.XTot + 2)*PRGUI.scaleFactor + PRGUI.xLeft), 90);
		g2.drawString("Total #Cells = " + NumberofComponents.recall(),
			(int)((PRGUI.XTot + 2)*PRGUI.scaleFactor + PRGUI.xLeft), 50);	
		if(PRGUI.netNum!=0) {
			g2.drawString("Total #Nets = " + PRGUI.netNum,
				(int)((PRGUI.XTot + 2)*PRGUI.scaleFactor + PRGUI.xLeft), 60);
		}
		else {
			g2.drawString("Total #Nets = ",
				(int)((PRGUI.XTot + 2)*PRGUI.scaleFactor + PRGUI.xLeft), 60);
		}
		g2.drawString("Total #Layers = " + PRGUI.ZTot,
			(int)((PRGUI.XTot + 2)*PRGUI.scaleFactor + PRGUI.xLeft), 70);	

		for(int i = 0; i < (PRGUI.XTot + 2); i++) {
			for(int j = 0; j < (PRGUI.YTot + 2); j++) {
				Rectangle2D rect = new Rectangle2D.Double((i*PRGUI.scaleFactor 
					+ PRGUI.xLeft), (j*PRGUI.scaleFactor + PRGUI.yTop),
					PRGUI.width, PRGUI.height);
					g2.setPaint(Color.LIGHT_GRAY);
					g2.fill(rect);
			}
		}
		
		//These sections beginning with if(PRGUI.coordPos== are drawing
		//the coordinates on the three corners visible in 2D mode
		if(PRGUI.coordPos==1) {

			int tmp1=0, tmp2=PRGUI.YTot+3;
			g2.setPaint(Color.WHITE);
			g2.drawString("(0,0)", 
				(int)(tmp1*PRGUI.scaleFactor + PRGUI.xLeft +
				PRGUI.scaleFactor/4), (int)(tmp2*PRGUI.scaleFactor +
				PRGUI.scaleFactor/4));
				
			tmp1=0;
			tmp2=1;
			g2.setPaint(Color.WHITE);
			g2.drawString("(0," + (PRGUI.YTot+1) + ")", 
				(int)(tmp1*PRGUI.scaleFactor + PRGUI.xLeft +
				PRGUI.scaleFactor/4), (int)(tmp2*PRGUI.scaleFactor-3));
					
			tmp1=PRGUI.YTot+1;
			tmp2=PRGUI.YTot+3;
			g2.setPaint(Color.WHITE);
			g2.drawString("(" + (PRGUI.YTot+1) + ",0)", 
				(int)(tmp1*PRGUI.scaleFactor + PRGUI.xLeft +
				PRGUI.scaleFactor/4), (int)(tmp2*PRGUI.scaleFactor +
				PRGUI.scaleFactor/4));
		}
		
		if(PRGUI.coordPos==2) {

			int tmp1=0, tmp2=PRGUI.YTot+3;
			g2.setPaint(Color.WHITE);
			g2.drawString("(" + (PRGUI.YTot+1) + ",0)", 
				(int)(tmp1*PRGUI.scaleFactor + PRGUI.xLeft +
				PRGUI.scaleFactor/4), (int)(tmp2*PRGUI.scaleFactor +
				PRGUI.scaleFactor/4));
				
			tmp1=0;
			tmp2=1;
			g2.setPaint(Color.WHITE);
			g2.drawString("(0,0)", 
				(int)(tmp1*PRGUI.scaleFactor + PRGUI.xLeft +
				PRGUI.scaleFactor/4), (int)(tmp2*PRGUI.scaleFactor-3));
					
			tmp1=PRGUI.YTot+1;
			tmp2=PRGUI.YTot+3;
			g2.setPaint(Color.WHITE);
			g2.drawString("(" + (PRGUI.YTot+1) + "," + (PRGUI.YTot+1) + ")", 
				(int)(tmp1*PRGUI.scaleFactor + PRGUI.xLeft +
				PRGUI.scaleFactor/4), (int)(tmp2*PRGUI.scaleFactor +
				PRGUI.scaleFactor/4));
		}
		
		if(PRGUI.coordPos==3) {

			int tmp1=0, tmp2=PRGUI.YTot+3;
			g2.setPaint(Color.WHITE);
			g2.drawString("(" + (PRGUI.YTot+1) + "," + (PRGUI.YTot+1) + ")",
				(int)(tmp1*PRGUI.scaleFactor + PRGUI.xLeft +
				PRGUI.scaleFactor/4), (int)(tmp2*PRGUI.scaleFactor +
				PRGUI.scaleFactor/4));
				
			tmp1=0;
			tmp2=1;
			g2.setPaint(Color.WHITE);
			g2.drawString("(" + (PRGUI.YTot+1) + ",0)", 
				(int)(tmp1*PRGUI.scaleFactor + PRGUI.xLeft +
				PRGUI.scaleFactor/4), (int)(tmp2*PRGUI.scaleFactor-3));
					
			tmp1=PRGUI.YTot+1;
			tmp2=PRGUI.YTot+3;
			g2.setPaint(Color.WHITE);
			g2.drawString("(0," + (PRGUI.YTot+1) + ")", 
				(int)(tmp1*PRGUI.scaleFactor + PRGUI.xLeft +
				PRGUI.scaleFactor/4), (int)(tmp2*PRGUI.scaleFactor +
				PRGUI.scaleFactor/4));
		}
		
		if(PRGUI.coordPos==4) {

			int tmp1=0, tmp2=PRGUI.YTot+3;
			g2.setPaint(Color.WHITE);
			g2.drawString("(0," + (PRGUI.YTot+1) + ")", 
				(int)(tmp1*PRGUI.scaleFactor + PRGUI.xLeft +
				PRGUI.scaleFactor/4), (int)(tmp2*PRGUI.scaleFactor +
				PRGUI.scaleFactor/4));
				
			tmp1=0;
			tmp2=1;
			g2.setPaint(Color.WHITE);
			g2.drawString("(" + (PRGUI.YTot+1) + "," + (PRGUI.YTot+1) + ")",
				(int)(tmp1*PRGUI.scaleFactor + PRGUI.xLeft +
				PRGUI.scaleFactor/4), (int)(tmp2*PRGUI.scaleFactor-3));
					
			tmp1=PRGUI.YTot+1;
			tmp2=PRGUI.YTot+3;
			g2.setPaint(Color.WHITE);
			g2.drawString("(0,0)", 
				(int)(tmp1*PRGUI.scaleFactor + PRGUI.xLeft +
				PRGUI.scaleFactor/4), (int)(tmp2*PRGUI.scaleFactor +
				PRGUI.scaleFactor/4));
		}
		
		for(int i = 0; i < NumberofComponents.recall(); i++) {
			double xVal = XHolder.GetX(i);
			double yVal = YHolder.GetY(i);
			double zVal = ZHolder.GetZ(i);
			String value = BNHolder.GetBN(i);
			if(zVal == PRGUIFrame.layerNum) {
				Rectangle2D rect = new Rectangle2D.Double((xVal*
					PRGUI.scaleFactor + PRGUI.xLeft), (yVal*PRGUI.scaleFactor), 
					PRGUI.width, PRGUI.height);
				g2.setPaint(Color.WHITE);
				g2.fill(rect);
				if(PRGUI.scaleFactor >= 80) {
					g2.setPaint(Color.BLACK);
					g2.drawString(value, (int)(xVal*PRGUI.scaleFactor + 5 + 
						PRGUI.xLeft), (int)(yVal*PRGUI.scaleFactor + 20));
				}
			}	
		}
		
		
		
		int i, ZStart, ZEnd;
		double startX, startY, XHalf, YHalf, endX, endY;
		String temp;
		String temp2 = new String("");
		if(PRGUI.showNetList==1) {
			
			for(int j = 0; j < BNSource.BNSource.length; j++) {
				temp = BNSource.GetBNSource(j);
				for(i = 0; i < NumberofComponents.recall() && 
					temp.compareTo(temp2)!=0; i++) {
						temp2 = BNameHolder.GetBName(i);
				}
				
				if(i == 0){
					i = NumberofComponents.recall();
				}

				ZStart = ZHolder.GetZ(i - 1);
				startX = (XHolder.GetX(i - 1))*PRGUI.scaleFactor + PRGUI.xLeft +
					PRGUI.scaleFactor/4;
				startY = (YHolder.GetY(i - 1))*PRGUI.scaleFactor +
					PRGUI.scaleFactor/4;
				for(int k = 0; k < PRGUI.NetListCon; k++) {
					temp = BNSink.GetBNSink(j, k);
					for(i = 0; i < NumberofComponents.recall() &&
						temp.compareTo(temp2)!=0; i++) {
							temp2 = BNameHolder.GetBName(i);
						}

					if(temp.compareTo("open")!=0) {
						
					if(i == 0){
						i = NumberofComponents.recall();
					}

					ZEnd = ZHolder.GetZ(i - 1);
					endX = (XHolder.GetX(i - 1))*PRGUI.scaleFactor + 
						PRGUI.xLeft + PRGUI.scaleFactor/4;
					endY = (YHolder.GetY(i - 1))*PRGUI.scaleFactor +
						PRGUI.scaleFactor/4;
					XHalf = ((endX+startX)/2);
					YHalf = ((endY+startY)/2);
					
					//This Section For Displaying Just Local Traffic
					if(ZStart == ZEnd && ZStart == PRGUIFrame.layerNum
						&& (PRGUI.NetListShow == 0 || PRGUI.NetListShow == 2)) {
						Line2D line = new 
							Line2D.Double(startX, startY, endX, endY);
						g2.setPaint(Color.CYAN);
						g2.draw(line);
					}//End Local Traffic
					
					//This Section For Displaying Outgoing Traffic
					else if(ZStart < ZEnd && ZStart == PRGUIFrame.layerNum
						&& (PRGUI.NetListShow == 1 || PRGUI.NetListShow == 2)) {
						Line2D line2 = new 
							Line2D.Double(startX, startY, XHalf, YHalf);
						g2.setPaint(Color.YELLOW);
						g2.draw(line2);
					}
					else if(ZStart < ZEnd && ZEnd == PRGUIFrame.layerNum
						&& (PRGUI.NetListShow == 1 || PRGUI.NetListShow == 2)) {
						Line2D line = new 
							Line2D.Double(XHalf, YHalf, endX, endY);
						g2.setPaint(Color.GREEN);
						g2.draw(line);
					}
					else if(ZStart > ZEnd && ZStart == PRGUIFrame.layerNum
						&& (PRGUI.NetListShow == 1 || PRGUI.NetListShow == 2)) {
						Line2D line2 = new 
							Line2D.Double(startX, startY, XHalf, YHalf);
						g2.setPaint(Color.GREEN);
						g2.draw(line2);
					}
					else if(ZStart > ZEnd && ZEnd == PRGUIFrame.layerNum
						&& (PRGUI.NetListShow == 1 || PRGUI.NetListShow == 2)) {
						Line2D line2 = new 
							Line2D.Double(XHalf, YHalf, endX, endY);
						g2.setPaint(Color.YELLOW);
						g2.draw(line2);
					}//End Outgoing Traffic
					
					//This Section is For Connections That Connect 2 Specific
					//Layers That The User Specifies
					else if(PRGUI.NetListShow == 3 && PRGUI.NLF == 0) {
						PRGUI.input1 = Integer.parseInt(
							netListFrame.input1.getText().trim());
						PRGUI.input2 = Integer.parseInt(
							netListFrame.input2.getText().trim());
						int val1 = PRGUI.input1;
						int val2 = PRGUI.input2;
						
						if(val1==ZStart && val2==ZEnd) {
							if(ZStart == PRGUIFrame.layerNum) {
								Line2D line2 = new 
									Line2D.Double(startX, startY, 
									XHalf, YHalf);
								g2.setPaint(Color.YELLOW);
								g2.draw(line2);
							}
							if(ZEnd == PRGUIFrame.layerNum) {
								Line2D line = new 
								Line2D.Double(XHalf, YHalf, endX, endY);
								g2.setPaint(Color.YELLOW);
								g2.draw(line);
							}	
						}
						else if(val2==ZStart && val1==ZEnd) {
							if(ZStart == PRGUIFrame.layerNum) {
								Line2D line2 = new 
									Line2D.Double(startX, startY, XHalf, YHalf);
								g2.setPaint(Color.GREEN);
								g2.draw(line2);
							}
							if(ZEnd == PRGUIFrame.layerNum) {
								Line2D line = new 
								Line2D.Double(XHalf, YHalf, endX, endY);
								g2.setPaint(Color.GREEN);
								g2.draw(line);
							}
							
						}
					}//End 2 Specific Layer Connections
					
					//This Section is For When Users want to Display
					//All Connections Between Two Layers They Specify
					else if(PRGUI.NetListShow == 4 && PRGUI.NLF == 0) {
						PRGUI.input1 = Integer.parseInt(
							netListFrame.input3.getText().trim());
						PRGUI.input2 = Integer.parseInt(
							netListFrame.input4.getText().trim());
						int val1 = PRGUI.input1;
						int val2 = PRGUI.input2;
						
						if(val1 > val2) {
							if(ZStart < val1 && ZStart > val2 && 
								ZEnd < val1 && ZEnd > val2) {
									if(ZStart == PRGUIFrame.layerNum) {
										Line2D line2 = new 
											Line2D.Double(startX, startY, 
											XHalf, YHalf);
										g2.setPaint(Color.CYAN);
										g2.draw(line2);
									}
									if(ZEnd == PRGUIFrame.layerNum) {
										Line2D line = new 
										Line2D.Double(XHalf, YHalf, endX, endY);
										g2.setPaint(Color.CYAN);
										g2.draw(line);
									}	
							}
						}
						else if(val2 > val1) {
							if(ZStart < val2 && ZStart > val1 && 
								ZEnd < val2 && ZEnd > val1) {
								if(ZStart == PRGUIFrame.layerNum) {
									Line2D line2 = new 
										Line2D.Double(startX, startY, 
										XHalf, YHalf);
									g2.setPaint(Color.CYAN);
									g2.draw(line2);
								}
								if(ZEnd == PRGUIFrame.layerNum) {
									Line2D line = new 
										Line2D.Double(XHalf, YHalf, endX, endY);
									g2.setPaint(Color.CYAN);
									g2.draw(line);
								}	
							}
						}
					}//End All Connections Between 2 Layers
					
					//This Section Displays NetList If Show Connections
					//That Span X Layers is Selected
					else if(PRGUI.NetListShow == 5 && PRGUI.NLF == 0) {
						PRGUI.input1 = Integer.parseInt(
							netListFrame.input5.getText().trim());
						int val1 = PRGUI.input1;
						if((ZStart - ZEnd) == val1) {
							if(ZStart == PRGUIFrame.layerNum) {
									Line2D line2 = new 
										Line2D.Double(startX, startY, 
										XHalf, YHalf);
									g2.setPaint(Color.CYAN);
									g2.draw(line2);
								}
								if(ZEnd == PRGUIFrame.layerNum) {
									Line2D line = new 
										Line2D.Double(XHalf, YHalf, endX, endY);
									g2.setPaint(Color.CYAN);
									g2.draw(line);
								}
						}
						else if((ZEnd - ZStart) == val1) {
							if(ZStart == PRGUIFrame.layerNum) {
									Line2D line2 = new 
										Line2D.Double(startX, startY, 
										XHalf, YHalf);
									g2.setPaint(Color.CYAN);
									g2.draw(line2);
								}
								if(ZEnd == PRGUIFrame.layerNum) {
									Line2D line = new 
										Line2D.Double(XHalf, YHalf, endX, endY);
									g2.setPaint(Color.CYAN);
									g2.draw(line);
								}
						}
					}//End Show Connections That Span X Layers
					
					
					}
					
				}
			}
		}
		

		//This Section Displays Critical Paths If Show The Kth Most
		//Critical Path is Selected
		String CPTemp, CPTemp2 = "", CPTemp3, CPTemp4 = "";
		int count;
		double cpStartX, cpStartY, cpStartZ, cpEndX, cpEndY, cpEndZ,
			cpRectX, cpRectY, cpRect2X, cpRect2Y;
		if(PRGUI.NumCritPath==1) {
			PRGUI.input1 = Integer.parseInt(
							critPathFrame.input1.getText().trim());
			if(PRGUI.input1==1) {
				int numofCP = Integer.parseInt(CPArray.GetCPArray(0, 1));
				for(int j = 2; j < (numofCP+1); j++) {
					CPTemp = CPArray.GetCPArray(0, j);
					for(count = 0; count < NumberofComponents.recall() &&
						CPTemp.compareTo(CPTemp2)!=0; count++) {
							CPTemp2 = BNHolder.GetBN(count);
					}
					if(count == 0) {
						count = NumberofComponents.recall();
					}
					cpStartX = (XHolder.GetX(count - 1))*PRGUI.scaleFactor + 
						PRGUI.xLeft + PRGUI.scaleFactor/4;
					cpStartY = startY = (YHolder.GetY(count - 1))*PRGUI.scaleFactor 
						+ PRGUI.scaleFactor/4;
					cpStartZ = ZHolder.GetZ(count - 1);
					cpRectX = XHolder.GetX(count - 1) * PRGUI.scaleFactor + 
						PRGUI.xLeft;
					cpRectY = YHolder.GetY(count - 1) * PRGUI.scaleFactor;
					if(cpStartZ==PRGUIFrame.layerNum) {
						Rectangle2D cpRect = new Rectangle2D.Double(cpRectX,
							cpRectY, PRGUI.width, PRGUI.height);
							g2.setPaint(Color.RED);
							g2.setStroke(new BasicStroke(3));
							g2.draw(cpRect);
					}
					if(cpStartZ>PRGUIFrame.layerNum) {
						Rectangle2D cpRect = new Rectangle2D.Double(cpRectX,
							cpRectY, PRGUI.width, PRGUI.height);
							g2.setPaint(Color.GREEN);
							g2.setStroke(new BasicStroke(3));
							g2.fill(cpRect);
					}
					if(cpStartZ<PRGUIFrame.layerNum) {
						Rectangle2D cpRect = new Rectangle2D.Double(cpRectX,
							cpRectY, PRGUI.width, PRGUI.height);
							g2.setPaint(Color.YELLOW);
							g2.setStroke(new BasicStroke(3));
							g2.fill(cpRect);
					}
					if(j!=(numofCP+1)) {
						CPTemp3 = CPArray.GetCPArray(0, j + 1);
						for(count = 0; count < NumberofComponents.recall() &&
							CPTemp3.compareTo(CPTemp4)!=0; count++) {
								CPTemp4 = BNHolder.GetBN(count);
							}
						cpEndX = (XHolder.GetX(count - 1))*PRGUI.scaleFactor + 
							PRGUI.xLeft + PRGUI.scaleFactor/4;
						cpEndY = startY = (YHolder.GetY(count - 1)) * 
							PRGUI.scaleFactor + PRGUI.scaleFactor/4;
						cpEndZ = ZHolder.GetZ(count - 1);
						cpRect2X = XHolder.GetX(count - 1) * PRGUI.scaleFactor + 
							PRGUI.xLeft;
						cpRect2Y = YHolder.GetY(count - 1) * PRGUI.scaleFactor;
							
						Line2D line = new 
							Line2D.Double(cpStartX, cpStartY, cpEndX, cpEndY);
						g2.setPaint(Color.RED);
						g2.setStroke(new BasicStroke(3));
						g2.draw(line);
						if(j==numofCP) {
							if(cpStartZ==PRGUIFrame.layerNum) {
								Rectangle2D cpRect2 = new Rectangle2D.Double(
									cpRect2X, cpRect2Y, PRGUI.width, 
									PRGUI.height);
								g2.draw(cpRect2);
							}
							if(cpStartZ>PRGUIFrame.layerNum) {
								Rectangle2D cpRect2 = new Rectangle2D.Double(
									cpRect2X, cpRect2Y, PRGUI.width, 
									PRGUI.height);
								g2.setPaint(Color.GREEN);
								g2.fill(cpRect2);
							}
							if(cpStartZ<PRGUIFrame.layerNum) {
								Rectangle2D cpRect2 = new Rectangle2D.Double(
									cpRect2X, cpRect2Y, PRGUI.width, 
									PRGUI.height);
								g2.setPaint(Color.YELLOW);
								g2.fill(cpRect2);
							}
						}
					}

				}
			}//End if 1-most critical path is shown
			else {
				int numofCP = Integer.parseInt(
					CPArray.GetCPArray((PRGUI.input1-1), 1));
				for(int j = 2; j < (numofCP+1); j++) {
					CPTemp = CPArray.GetCPArray((PRGUI.input1-1), j);
					for(count = 0; count < NumberofComponents.recall() &&
						CPTemp.compareTo(CPTemp2)!=0; count++) {
							CPTemp2 = BNHolder.GetBN(count);
					}
					if(count == 0) {
						count = NumberofComponents.recall();
					}
					cpStartX = (XHolder.GetX(count - 1))*PRGUI.scaleFactor + 
						PRGUI.xLeft + PRGUI.scaleFactor/4;
					cpStartY = startY = (YHolder.GetY(count - 1))*PRGUI.scaleFactor 
						+ PRGUI.scaleFactor/4;
					cpStartZ = ZHolder.GetZ(count - 1);
					cpRectX = XHolder.GetX(count - 1) * PRGUI.scaleFactor + 
						PRGUI.xLeft;
					cpRectY = YHolder.GetY(count - 1) * PRGUI.scaleFactor;
					if(cpStartZ==PRGUIFrame.layerNum) {
						Rectangle2D cpRect = new Rectangle2D.Double(cpRectX,
							cpRectY, PRGUI.width, PRGUI.height);
							g2.setPaint(Color.RED);
							g2.setStroke(new BasicStroke(1));
							g2.draw(cpRect);
					}
					if(cpStartZ>PRGUIFrame.layerNum) {
						Rectangle2D cpRect = new Rectangle2D.Double(cpRectX,
							cpRectY, PRGUI.width, PRGUI.height);
							g2.setPaint(Color.GREEN);
							g2.setStroke(new BasicStroke(1));
							g2.fill(cpRect);
					}
					if(cpStartZ<PRGUIFrame.layerNum) {
						Rectangle2D cpRect = new Rectangle2D.Double(cpRectX,
							cpRectY, PRGUI.width, PRGUI.height);
							g2.setPaint(Color.YELLOW);
							g2.setStroke(new BasicStroke(1));
							g2.fill(cpRect);
					}
						
					if(j!=(numofCP+1)) {
						CPTemp3 = CPArray.GetCPArray((PRGUI.input1-1), j + 1);
						for(count = 0; count < NumberofComponents.recall() &&
							CPTemp3.compareTo(CPTemp4)!=0; count++) {
								CPTemp4 = BNHolder.GetBN(count);
							}
						cpEndX = (XHolder.GetX(count - 1))*PRGUI.scaleFactor + 
							PRGUI.xLeft + PRGUI.scaleFactor/4;
						cpEndY = startY = (YHolder.GetY(count - 1)) * 
							PRGUI.scaleFactor + PRGUI.scaleFactor/4;
						cpEndZ = ZHolder.GetZ(count - 1);
						cpRect2X = XHolder.GetX(count - 1) * PRGUI.scaleFactor + 
							PRGUI.xLeft;
						cpRect2Y = YHolder.GetY(count - 1) * PRGUI.scaleFactor;
							
						Line2D line = new 
							Line2D.Double(cpStartX, cpStartY, cpEndX, cpEndY);
						g2.setPaint(Color.RED);
						g2.setStroke(new BasicStroke(1));
						g2.draw(line);
						
						if(j==numofCP) {
							if(cpStartZ==PRGUIFrame.layerNum) {
								Rectangle2D cpRect2 = new Rectangle2D.Double(
									cpRect2X, cpRect2Y, PRGUI.width, 
									PRGUI.height);
								g2.draw(cpRect2);
							}
							if(cpStartZ>PRGUIFrame.layerNum) {
								Rectangle2D cpRect2 = new Rectangle2D.Double(
									cpRect2X, cpRect2Y, PRGUI.width, 
									PRGUI.height);
								g2.setPaint(Color.GREEN);
								g2.fill(cpRect2);
							}
							if(cpStartZ<PRGUIFrame.layerNum) {
								Rectangle2D cpRect2 = new Rectangle2D.Double(
									cpRect2X, cpRect2Y, PRGUI.width, 
									PRGUI.height);
								g2.setPaint(Color.YELLOW);
								g2.fill(cpRect2);
							}
						}
					}
				}
			}//End if 1-most critical path is shown
		}//End Show Kth-Most Critical Path
		
		//This Section Shows Critical Paths if show first K-Most Paths
		//Is Selected
		if(PRGUI.NumCritPath==2) {
			PRGUI.input1 = Integer.parseInt(
							critPathFrame.input2.getText().trim());
			for(int u = 0; u < PRGUI.input1; u++) {
				int numofCP = Integer.parseInt(
					CPArray.GetCPArray((u), 1));
				for(int j = 2; j < (numofCP+1); j++) {
					CPTemp = CPArray.GetCPArray((u), j);
					for(count = 0; count < NumberofComponents.recall() &&
						CPTemp.compareTo(CPTemp2)!=0; count++) {
							CPTemp2 = BNHolder.GetBN(count);
					}
					if(count == 0) {
						count = NumberofComponents.recall();
					}
					cpStartX = (XHolder.GetX(count - 1))*PRGUI.scaleFactor + 
						PRGUI.xLeft + PRGUI.scaleFactor/4;
					cpStartY = startY = (YHolder.GetY(count - 1))*PRGUI.scaleFactor 
						+ PRGUI.scaleFactor/4;
					cpStartZ = ZHolder.GetZ(count - 1);
					cpRectX = XHolder.GetX(count - 1) * PRGUI.scaleFactor + 
						PRGUI.xLeft;
					cpRectY = YHolder.GetY(count - 1) * PRGUI.scaleFactor;
					if(cpStartZ==PRGUIFrame.layerNum) {
						Rectangle2D cpRect = new Rectangle2D.Double(cpRectX,
							cpRectY, PRGUI.width, PRGUI.height);
							g2.setPaint(Color.RED);
							g2.setStroke(new BasicStroke(1));
							g2.draw(cpRect);
					}
					if(cpStartZ>PRGUIFrame.layerNum) {
						Rectangle2D cpRect = new Rectangle2D.Double(cpRectX,
							cpRectY, PRGUI.width, PRGUI.height);
							g2.setPaint(Color.GREEN);
							g2.setStroke(new BasicStroke(1));
							g2.fill(cpRect);
					}
					if(cpStartZ<PRGUIFrame.layerNum) {
						Rectangle2D cpRect = new Rectangle2D.Double(cpRectX,
							cpRectY, PRGUI.width, PRGUI.height);
							g2.setPaint(Color.YELLOW);
							g2.setStroke(new BasicStroke(1));
							g2.fill(cpRect);
					}
						
					if(j!=(numofCP+1)) {
						CPTemp3 = CPArray.GetCPArray((u), j + 1);
						for(count = 0; count < NumberofComponents.recall() &&
							CPTemp3.compareTo(CPTemp4)!=0; count++) {
								CPTemp4 = BNHolder.GetBN(count);
							}
						cpEndX = (XHolder.GetX(count - 1))*PRGUI.scaleFactor + 
							PRGUI.xLeft + PRGUI.scaleFactor/4;
						cpEndY = startY = (YHolder.GetY(count - 1)) * 
							PRGUI.scaleFactor + PRGUI.scaleFactor/4;
						cpEndZ = ZHolder.GetZ(count - 1);
						cpRect2X = XHolder.GetX(count - 1) * PRGUI.scaleFactor + 
							PRGUI.xLeft;
						cpRect2Y = YHolder.GetY(count - 1) * PRGUI.scaleFactor;
							
						Line2D line = new 
							Line2D.Double(cpStartX, cpStartY, cpEndX, cpEndY);
						g2.setPaint(Color.RED);
						g2.setStroke(new BasicStroke(1));
						g2.draw(line);
						
						if(j==numofCP) {
							if(cpStartZ==PRGUIFrame.layerNum) {
								Rectangle2D cpRect2 = new Rectangle2D.Double(
									cpRect2X, cpRect2Y, PRGUI.width, 
									PRGUI.height);
								g2.draw(cpRect2);
							}
							if(cpStartZ>PRGUIFrame.layerNum) {
								Rectangle2D cpRect2 = new Rectangle2D.Double(
									cpRect2X, cpRect2Y, PRGUI.width, 
									PRGUI.height);
								g2.setPaint(Color.GREEN);
								g2.fill(cpRect2);
							}
							if(cpStartZ<PRGUIFrame.layerNum) {
								Rectangle2D cpRect2 = new Rectangle2D.Double(
									cpRect2X, cpRect2Y, PRGUI.width, 
									PRGUI.height);
								g2.setPaint(Color.YELLOW);
								g2.fill(cpRect2);
							}
						}
					}
				}
			}
		}//End Show First K-Most Critical Paths

	}//End 2D View Mode	
		
	}
}

//The Holder classes Hold Placement Information in Arrays

class XHolder {
	
	static int XArray[];
	
	public XHolder() {
	}
		
	public static void AddX(int i, int value) {
		
		XArray[i] = value;
	}
	
	public static void initArray(int i) {
		
		XArray = new int[i];
	}
	
	public static int GetX(int i) {
		
		return XArray[i];
	}
}

class YHolder {
	
	static int YArray[];
	
	public YHolder() {
	}
	
	public static void AddY(int i, int value) {
		
		YArray[i] = value;
	}
	
	public static void initArray(int i) {
		
		YArray = new int[i];
	}
	
	public static int GetY(int i) {
		
		return YArray[i];
	}
}

class YJavaHolder {
	
	static int YArray[];
	
	public YJavaHolder() {
	}
	
	public static void AddY(int i, int value) {
		
		YArray[i] = value;
	}
	
	public static void initArray(int i) {
		
		YArray = new int[i];
	}
	
	public static int GetY(int i) {
		
		return YArray[i];
	}
}

class ZHolder {
	
	static int ZArray[];
	
	public ZHolder() {
	}
	
	public static void AddZ(int i, int value) {
		
		ZArray[i] = value;
	}
	
	public static void initArray(int i) {
		
		ZArray = new int[i];
	}
	
	public static int GetZ(int i) {
		
		return ZArray[i];
	}
}

class BNHolder {
	
	static String BNArray[];
	
	public BNHolder() {
	}
	
	public static void AddBN(int i, String value) {
		
		BNArray[i] = value;
	}
	
	public static void initArray(int i) {
		
		BNArray = new String[i];
	}
	
	public static String GetBN(int i) {
		
		return BNArray[i];
	}
}

class BNameHolder {
	
	static String BNameArray[];
	
	public BNameHolder() {
	}
	
	public static void AddBName(int i, String value) {
		
		BNameArray[i] = value;
	}
	
	public static void initArray(int i) {
		
		BNameArray = new String[i];
	}
	
	public static String GetBName(int i) {
		
		return BNameArray[i];
	}
}

class BNSource {
	
	static String BNSource[];
	
	public BNSource() {
	}
	
	public static void AddBNSource(int i, String value) {
		
		BNSource[i] = value;
	}
	
	public static void initArray(int i) {
		
		BNSource = new String[i];
	}
	
	public static String GetBNSource(int i) {
		
		return BNSource[i];
	}
}

class BNSink {
	
	static String BNSink[][];
	
	public BNSink() {
	}
	
	public static void AddBNSink(int i,int j, String value) {
		
		BNSink[i][j] = value;
	}
	
	public static void initArray(int i, int j) {
		
		BNSink = new String[i][j];
	}
	
	public static String GetBNSink(int i, int j) {
		
		return BNSink[i][j];
	}
}

//NumberofComponents Stores the Number of Components

class NumberofComponents {
	
	private static int count;
	
	public NumberofComponents() {
	}
	
	public static void store(int i) {
		
		count = i;
	}
	
	public static int recall() {
		
		return count;
	}
}

//CPArray Class Holds The Array to Store Critical Path Information

class CPArray {
	
	static String CPA[][];
	
	public CPArray() {
	}
	
	public static void AddCPArray(int i,int j, String value) {
		
		CPA[i][j] = value;
	}
	
	public static void initArray(int i, int j) {
		
		CPA = new String[i][j];
	}
	
	public static String GetCPArray(int i, int j) {
		
		return CPA[i][j];
	}
}