#!/usr/bin/perl
# cristinel.ababei
# This script goes thru directories of type results/ai_layj
# and extracts data from out files of TPR. Then, it dumps
# out these info as 2D tables.

# ===========================================================================

$path = "/home/cristinel/projects/tpr4/results/";

@dirs = ("a1_lay", "a2_lay");
@circuits = ("dsip", "tseng", "diffeq", "bigkey", "s298", "frisc", "elliptic", "s38417", "s38584.1", "clma");
@place_metrics = ("wl", "delay", "run_time", "nets_span_more_layers");
@route_metrics = ("wl", "delay", "run_time", "HCW", "VCW", "avg_net_length", "route_area");

# ===========================================================================


sub createFileName
{
    my($arch_no, $layer_no, $cir_no, $run_no) = @_;
    my($fileName);
    $fileName = "$path";
    $fileName =  $fileName . @dirs[$arch_no]. $layer_no . "/" . @circuits[$cir_no] . "." . ($arch_no+1) . "_" . $run_no . ".out";
    #print("\n $fileName");
}# eor


sub extractFromFile
{
# Lines with data in out files generated by TPR look like this:
# Place-level estimations:
#        Delay     	WL	Run_time	Frac_nets_span_more_layers.
# Final: 5.69578e-08	8781	9.4		0.0436761
# Route-level actual computations:
#        Delay     	WL	Run_time	HCW	VCW	Avg_net_length	Optimistic_route_area.
# Final: 7.24162e-08	23994	19.68		19	5	21.8525		2.75224e+06

    my($fileName, $arch_no, $layer_no, $cir_no) = @_;
    my($i, @v, $line, @fileContent);

    open(INFILE, $fileName) || die "Error opening $fileName!\n";

    $processed_place_info = 0;
    @fileContent = <INFILE>;
    for ($i=0; $i<scalar(@fileContent); $i++)
    {
	$line = @fileContent[$i];
	chop($line);
	@v = split(/\s+/, $line);
	if ( (@v[0] =~ /Final:/)&&($processed_place_info =~ 0) )
	{
	    $place_delay[$cir_no][$layer_no] = @v[1];
	    $place_wl[$cir_no][$layer_no] = @v[2];
	    $place_run_time[$cir_no][$layer_no] = @v[3];
	    $place_nets_span[$cir_no][$layer_no] = @v[4];
	    $processed_place_info = 1;
	}
	elsif ( (@v[0] =~ /Final:/)&&($processed_place_info =~ 1) ) 
	{
	    $route_delay[$cir_no][$layer_no] = @v[1];
	    $route_wl[$cir_no][$layer_no] = @v[2];
	    $route_run_time[$cir_no][$layer_no] = @v[3];
	    $route_HCW[$cir_no][$layer_no] = @v[4];
	    $route_VCW[$cir_no][$layer_no] = @v[5];
	    $route_avg_net_length[$cir_no][$layer_no] = @v[6];
	    $route_route_area[$cir_no][$layer_no] = @v[7];
	    $processed_place_info = 0;
	    return; # Done processing current file.
	}
    }
# Here I arrive only if the out file was created with premature 
# exit and "Final:" lines were not dumped into .out file.
return;
}# eor
                       
                            
sub readFiles
{
# Read and process out files created for arch_no. 
    my($arch_no) = @_;
    my($i, $j, $currFileName, $run_no);

    $run_no = 1; # For now I work with only one run.
    for ($j=1; $j<11; $j++) # Simulations are for upto 10 layers.
    {
	for ($i=0; $i<scalar(@circuits); $i++)
	{
	    $currFileName = &createFileName( $arch_no, $j, $i, $run_no);
	    &extractFromFile( $currFileName, $arch_no, $j, $i);
	}
    }                               
}# eor


sub dumpValues
{
# Prints arrays with collected data for arch_no.
# Has to be called after a call of &readFiles.
    my($arch_no) = @_;
    my($i, $j);

    if ($arch_no =~ 0)
    {
	print "\n Results for ARCHITECTURE Sing-Seg\n";
    }
    elsif ($arch_no =~ 1)
    {
	print "\n Results for ARCHITECTURE Multi-Seg\n";
    }
    else 
    {
	exit;
    }


    print "\nPLACE results";
    print "\nPlace Delay";
    for ($i=0; $i<scalar(@circuits); $i++)
    {
	print("\n");
	for ($j=1; $j<11; $j++) # Simulations are for upto 10 layers.
	{
	    print("$place_delay[$i][$j] ");
	}
    }          
    print "\n Place WL";
    for ($i=0; $i<scalar(@circuits); $i++)
    {
	print("\n");
	for ($j=1; $j<11; $j++) # Simulations are for upto 10 layers.
	{
	    print("$place_wl[$i][$j] ");
	}
    }
    print "\nPlace run_time";
    for ($i=0; $i<scalar(@circuits); $i++)
    {
	print("\n");
	for ($j=1; $j<11; $j++) # Simulations are for upto 10 layers.
	{
	    print("$place_run_time[$i][$j] ");
	}
    }
    print "\nPlace proc of nets_span moree layers";
    for ($i=0; $i<scalar(@circuits); $i++)
    {
	print("\n");
	for ($j=1; $j<11; $j++) # Simulations are for upto 10 layers.
	{
	    print("$place_nets_span[$i][$j] ");
	}
    }


    print "\nROUTE results";
    print "\nRoute Delay";
    for ($i=0; $i<scalar(@circuits); $i++)
    {
	print("\n");
	for ($j=1; $j<11; $j++) # Simulations are for upto 10 layers.
	{
	    print("$route_delay[$i][$j] ");
	}
    }          
    print "\nRoute WL";
    for ($i=0; $i<scalar(@circuits); $i++)
    {
	print("\n");
	for ($j=1; $j<11; $j++) # Simulations are for upto 10 layers.
	{
	    print("$route_wl[$i][$j] ");
	}
    }
    print "\nRoute run_time";
    for ($i=0; $i<scalar(@circuits); $i++)
    {
	print("\n");
	for ($j=1; $j<11; $j++) # Simulations are for upto 10 layers.
	{
	    print("$route_run_time[$i][$j] ");
	}
    }
    print "\nRoute HCW";
    for ($i=0; $i<scalar(@circuits); $i++)
    {
	print("\n");
	for ($j=1; $j<11; $j++) # Simulations are for upto 10 layers.
	{
	    print("$route_HCW[$i][$j] ");
	}
    }
    print "\nRoute VCW";
    for ($i=0; $i<scalar(@circuits); $i++)
    {
	print("\n");
	for ($j=1; $j<11; $j++) # Simulations are for upto 10 layers.
	{
	    print("$route_VCW[$i][$j] ");
	}
    }
    print "\nRoute avg_net_length";
    for ($i=0; $i<scalar(@circuits); $i++)
    {
	print("\n");
	for ($j=1; $j<11; $j++) # Simulations are for upto 10 layers.
	{
	    print("$route_avg_net_length[$i][$j] ");
	}
    }
    print "\nRoute route_area";
    for ($i=0; $i<scalar(@circuits); $i++)
    {
	print("\n");
	for ($j=1; $j<11; $j++) # Simulations are for upto 10 layers.
	{
	    print("$route_route_area[$i][$j] ");
	}
    }
}# eor

# ===========================================================================


# And now, do the actual work.

# For Sing-Seg architecture.
&readFiles( 0 );
&dumpValues( 0 );
# For Multi-Seg architecture.
# &readFiles( 1 );
# &dumpValues( 1 );
