// ======================================================================= //
// Cristinel Ababei 10/3/2003
// ======================================================================= //

#include "util.h"
#include "fpga3d.h"
#include "tt_hgraph.h"


using namespace std;

// ======================================================================= //
//                            s_fpga3d
// ======================================================================= //
s_fpga3d::s_fpga3d()
{



  // --------------------------------------------------------------------- //
  //                              SECTION I
  // Stuff that should normally be here only
  // --------------------------------------------------------------------- //

  num_nets=0; num_blocks=0;
  vertical_delay_same=1; vertical_uwl=1;//default values!
  total_num_subblocks=0;
  num_p_inputs=0; num_p_outputs=0; num_clbs=0; num_globals=0;



  // --------------------------------------------------------------------- //
  //                              SECTION II
  // Stuff that should have been in s_parser::s_parser()
  // --------------------------------------------------------------------- //

  chunk_bytes_avail = 0;
  chunk_next_avail_mem = NULL;

  char *local_names[NUMINP] = {"io_rat", 
   "chan_width_x", "chan_width_y",
   "chan_width_io", "outpin", "inpin", 
   "subblocks_per_clb", "subblock_lut_size", 
   "chan_width_z", 
   "Fc_output", "Fc_input", "Fc_pad", "Fc_type", 
   "switch_block_type", 
   "segment", "via",
   "switch", "R_minW_nmos", "R_minW_pmos",
   "C_ipin_cblock", "T_ipin_cblock", "T_sblk_opin_to_sblk_ipin", 
   "T_clb_ipin_to_sblk_ipin", "T_sblk_opin_to_clb_opin", "T_ipad", "T_opad",
   "T_subblock"};
  for(int i=0;i<NUMINP;i++)
    {
      names[i] =(char *)my_malloc((strlen(local_names[i])+1) * sizeof(char));
      strcpy(names[i], local_names[i]);
      //printf(" %d -- %s %s\n",(i+1),local_names[i],names[i]);
    }

  // --------------------------------------------------------------------- //
  //                              SECTION III
  // --------------------------------------------------------------------- //
  num_layers=2; //default value
  for(int j=0;j<20;j++) //safe initialization
    {
      balance[j]=0;
      pi_balance[j]=0;
      po_balance[j]=0;
      clb_balance[j]=0;
    }
  // --------------------------------------------------------------------- //
  //                              SECTION V
  // --------------------------------------------------------------------- //
  
  // Used to keep my own list of free linked integers, for speed reasons.
  free_edge_list_head = NULL;

  // --------------------------------------------------------------------- //
  //                              SECTION VI
  // --------------------------------------------------------------------- //
  rr_node_to_rt_node = NULL;  
  rt_node_free_list = NULL;
  rt_edge_free_list = NULL;

  rr_node_route_inf = NULL;   
  route_bb = NULL;
  heap_free_head = NULL; 
  trace_free_head = NULL;
  rr_modified_head = NULL;
  linked_f_pointer_free_head = NULL;

  tedge_ch_list_head = NULL;
  tedge_ch_bytes_avail = 0;
  tedge_ch_next_avail = NULL;

 
}//eof
// ======================================================================= //


// ======================================================================= //
tt_hgraph * s_fpga3d::create_timing_tt_hgraph_01( t_subblock_data subblock_data )
{
  // Returns tt_hgraph which mirrors all subblocks and PI/PO pads
  // inside fpga3d's block[]. This hgraph is used for implementation of STA.
  // "subblock_data" is declared as variable inside main(). 
  //
  // "_01" First type of timing graph. All subblocks and IO OUT PADS
  // are made vertices.  Nets include also connections which are 
  // completely inside a block (block inner).
  // Currently not used.  Instead use "_02" which is similar to the
  // timing graph used inside vpr, where vertices are only clb_pins 
  // which are used + all pi po pads.
  // 

  // Get some global stuff in here, locally.  Magariile astea contin
  // info despre all subblocks:
  t_subblock **subblock_inf;
  int *num_subblocks_per_block;
  int subblock_lut_size;
  subblock_inf = subblock_data.subblock_inf;
  num_subblocks_per_block = subblock_data.num_subblocks_per_block;
  subblock_lut_size = subblock_data.subblock_lut_size;

  int i=0,j=0;
  tt_hgraph *g = new tt_hgraph(); //allocate object timing hgraph


  // === First, allocate all vertices in verts[]
  for (i=0;i<total_num_subblocks;i++) {
    // - 1st create vertex:
    Vertex t_vertex;
    // - add this vertex to verts[] of tt_hgraph. It also 
    // increments NumVerts of "g"
    g->addVertex( t_vertex, i );
  }//for


  // === Second, add all subblocks of blocks. 
  // PI/PO pads are also among blocks. This is done by visiting all
  // nets.
  int source_block_id=0,sink_block_id=0;
  int source_block_id_top_level=0;
  int sink_block_id_top_level=0;
  int blk_pin_of_source_block=0;
  int blk_pin_of_sink_block=0;
  for (i=0;i<num_nets;i++) 
   {
    if (is_global[i] == false) { 
    // Global nets are not created in my timing graph.
    // That is why I will have a num_nets-1 nets in my timing graph!
    //..............................................................
      //printf(" %d %d ",i,net[i].blocks[0]);
      for (j=0;j<net[i].num_pins;j++)
	{
	  //printf(" %d",net[i].blocks[j]);
	  // register block-level source id or sink id
	  if (j == 0) // --- source_block_id
	    {
	      source_block_id = net[i].blocks[j];
	      if ( block[source_block_id].type == INPAD )
		{
		  //printf(" %d",source_block_id);
		  source_block_id_top_level =
		    block[source_block_id].id_top_level;//pi
		  // will take place only once:
		  g->verts[source_block_id_top_level].setType( INPAD );
		  g->verts[source_block_id_top_level].blockID=source_block_id;
		}//if
	      if ( block[source_block_id].type == CLB )
		{
		  //printf(" %d",source_block_id);
		  // source_block_id_top_level is given by the
		  // id_top_level of the subblock, which drives
		  // this net. so, get first blk_pin, which
		  // should be a "is_opin".
		  blk_pin_of_source_block = net[i].blk_pin[j];
		  // now, the question is which subblock inside
		  // block source drives this pin. that particular
		  // subblock establishes source_block_id_top_level.
		  source_block_id_top_level = 
		    get_id_top_level_of_subblock_driving_opin (
			 blk_pin_of_source_block, source_block_id, 
			 &subblock_data );//il facem pointer
		  // will take place only once:
		  g->verts[source_block_id_top_level].setType( CLB );
		  g->verts[source_block_id_top_level].blockID=source_block_id;
		}//if

	    }//if (j == 0)
	  else        // --- sink_block_id
	    {
	      sink_block_id = net[i].blocks[j];
	      if ( block[sink_block_id].type == OUTPAD )// 
		{
		  //printf(" %d",sink_block_id);
		  sink_block_id_top_level =
		    block[sink_block_id].id_top_level;//po
		  // will take place only once:
		  g->verts[sink_block_id_top_level].setType( OUTPAD );
		  g->verts[sink_block_id_top_level].blockID=sink_block_id;
		  // Add fanin and fanout elements accordingly:
		  g->verts[source_block_id_top_level].addElementInFanout( 
					         sink_block_id_top_level );
		  g->verts[sink_block_id_top_level].addElementInFanin( 
				            source_block_id_top_level );
		}//if
	      if ( block[sink_block_id].type == CLB )
		{
		  // sink_block_id_top_level is given by the
		  // id_top_level of the subblock, which is driven
		  // by this net. so, get first blk_pin, which
		  // should be a "is_ipin".
		  blk_pin_of_sink_block = net[i].blk_pin[j];
		  // now, the question is which subblock(s) inside
		  // block source is(are) driven by this pin. That particular
		  // subblock establishes sink_block_id_top_level.
		  // The problem here is that more subblocks can 
		  // be driven. So, here I go and check all
		  // inputs of each subblock
   // ---------------------------------------------
   // sink_block_id_top_level = 
   // get_id_top_level_of_subblock_driven_by_ipin (
   // blk_pin_of_sink_block, sink_block_id, 
   // &subblock_data );

   int isub=0, ipin=0;
   int is_blk_pin_input_to_subblock =0;
   int loc_num_subblocks = 
     num_subblocks_per_block[ sink_block_id ];
   for (isub=0;isub<loc_num_subblocks;isub++) 
     {
       is_blk_pin_input_to_subblock =0;//reset
       for (ipin=0;ipin<subblock_lut_size;ipin++) {// Input pins
	 if ( subblock_inf[sink_block_id][isub].inputs[ipin] == 
	      blk_pin_of_sink_block ) {//found subblock driven by ipin
	   is_blk_pin_input_to_subblock =1;
	   sink_block_id_top_level = 
	     subblock_inf[sink_block_id][isub].id_top_level;
	 }
       }
       if ( is_blk_pin_input_to_subblock ) {
	 // above was found subblock driven by ipin
	 // so account for it in tt_hgraph.
	 // Add fanin and fanout elements accordingly:
	 g->verts[source_block_id_top_level].addElementInFanout( 
					sink_block_id_top_level );
	 g->verts[sink_block_id_top_level].addElementInFanin( 
				   source_block_id_top_level );
       } 
     }//for
   // In momentul asta ar fi trebuit sa fi added Elements
   // macar o data in if-ul de mai sus.
   // ---------------------------------------------
		}//if

	    }//else

	}//for inner
      //..............................................................
    }//if
   }//for outer

  // In momentul asta inca sint nets care nu sint complete: alea
  // interne fiecarui block!
  int num_blk_inner_connections=0;
  for (i=0;i<num_blocks;i++) 
    {
      if ( block[i].type == CLB )//go thru all clbs only
	{
	  num_blk_inner_connections = num_blk_inner_connections +
	    finish_building_fanin_fanouts_of_verts_dueto_this_block_01 (
	    i, &subblock_data, g );
	}
    }
  printf("\n Processed %d blk_inner connections.",num_blk_inner_connections);

  // Build all nets, though I am not sure if needed.
  g->BuildNets();

  printf("\nDone create_timing_tt_hgraph.");
  return g; 
}//eof


// ======================================================================= //


int s_fpga3d::get_id_top_level_of_subblock_driving_opin (
	       int blk_opin, int block_id,
	       t_subblock_data *subblock_data_ptr )
{
  // given:  blk_opin, block_id
  // find:  id_top_level of subblock driving this opin
  int to_be_returned=-1;

  // Get some global stuff in here, locally.  Magariile astea contin
  //  info despre all subblocks:
  t_subblock *subblock_inf;
  int num_subblocks_per_block;
  subblock_inf = subblock_data_ptr->subblock_inf[ block_id ];
  num_subblocks_per_block = subblock_data_ptr->num_subblocks_per_block[ block_id ];

  int isub=0;
  for (isub=0;isub<num_subblocks_per_block;isub++) 
    {
      if ( subblock_inf[ isub ].output == blk_opin )
	to_be_returned = subblock_inf[ isub ].id_top_level;

    }//for

  if ( to_be_returned ==-1 )
    { 
      printf ("Abort:  output pin %d on block #%d (%s) is not driven "
	      "by any subblock.\n", blk_opin, block_id, block[block_id].name);
      printf (" -- get_id_top_level_of_subblock_driving_opin() ");
      exit(1);
    }

  return to_be_returned;
}//eof
// ======================================================================= //
int s_fpga3d::get_id_top_level_of_subblock_driven_by_ipin (
	       int blk_ipin, int block_id,
	       t_subblock_data *subblock_data_ptr )
{
  // given:  blk_ipin, block_id
  // find:  id_top_level of subblock(S) ??? driven by this ipin

}//eof
// ======================================================================= //
int s_fpga3d::finish_building_fanin_fanouts_of_verts_dueto_this_block_01 (
	       int block_id,
	       t_subblock_data *subblock_data_ptr,
               tt_hgraph *g)
{
  // After going through all nets and building tt_hgraph there
  // are stil interconections, which are not captured by fanin and fanout
  // of verts. These interconnections are those internal/inside
  // a block and which connect out of a subblock to the input(s) of
  // other subblock of same block.
  // Here we go and search for such cases and add them into the top-level
  // timing graph.  However, we *mark* these nets as being blk_inner
  // to know that they have delay very small.

  // ---------------------------------------------------------------
  // Get some global stuff in here, locally.  Magariile astea contin
  // info despre all subblocks:
  int subblock_lut_size;
  int max_subblocks_per_block;
  int num_subblocks;
  t_subblock *subblock_inf;
  subblock_lut_size = subblock_data_ptr->subblock_lut_size;
  max_subblocks_per_block = subblock_data_ptr->max_subblocks_per_block;
  num_subblocks = subblock_data_ptr->num_subblocks_per_block[ block_id ];
  subblock_inf = subblock_data_ptr->subblock_inf[ block_id ];
  // ---------------------------------------------------------------
  int counter=0;
  int isub=0,ipin=0,subblk_id_driving=0;
  int source_subblk_id_top_level=0,sink_subblk_id_top_level=0;
  int loc_inputs_of_pin=0;
  for (isub=0;isub<num_subblocks;isub++)//SUBBBLOCK(s)
    {
      // See if this subblock has used inputs
      // and if so, then take measures
      for (ipin=0;ipin<subblock_lut_size;ipin++)//IPINS
	{
	  if (subblock_inf[isub].inputs[ipin] != OPEN) 
	    {
	      // Means that this subblock has input used
	      // Let's see who drives it.  If driven by another 
	      // subblock: that's it - I found a blk-inner
	      // connection and I have to add it to the timing 
	      // graph
	      loc_inputs_of_pin=subblock_inf[isub].inputs[ipin];
	      if ( loc_inputs_of_pin >= pins_per_clb )
		{
		  // Means that this input does not come from a 
		  // clb pin but rather from the out of other 
		  // subbblock, whose index inside this block, as
		  // subblock is:
		  subblk_id_driving = 
		    loc_inputs_of_pin - pins_per_clb;
		  source_subblk_id_top_level =
		    subblock_inf[subblk_id_driving].id_top_level;
		  sink_subblk_id_top_level =
		    subblock_inf[isub].id_top_level;
		  // Reflect this new blk_inner connection just
		  // fount into timing graph "g".  Check also for 
		  // the possibility of having a subblock 
		  // completely conencted inside the host block.

		  g->verts[source_subblk_id_top_level].setType( CLB );
		  g->verts[source_subblk_id_top_level].blockID=block_id;
		  g->verts[sink_subblk_id_top_level].setType( CLB );
		  g->verts[sink_subblk_id_top_level].blockID=block_id;

		  g->verts[source_subblk_id_top_level].addElementInFanout( 
			sink_subblk_id_top_level );
		  g->verts[sink_subblk_id_top_level].addElementInFanin( 
			source_subblk_id_top_level );
		  
		  counter++;
		}//if
	    }//if != OPEN
	}//for (ipin=0;
    }//for (isub=0;
  // ---------------------------------------------------------------

  return counter;
}//eof
// ======================================================================= //






// ======================================================================= //
tt_hgraph *s_fpga3d::create_timing_tt_hgraph_02(
	   t_subblock_data subblock_data,t_timing_inf timing_inf)
{
  //
  // "_02" Second type of timing graph.
  //
  // Returns tt_hgraph which has as vertices only clb_pins 
  // which are used (i.e., are net terminals) + all IO OUT PADs.
  // In this way the timing graph is simpler and has two kinds of 
  // nets: (i) those which are nets[] inside fpga3d class with delay
  // given by the routing, and (ii) nets between ipins and opins 
  // of same block with constant/fixed delay determined thru how
  // many LUT signals go.
  // 



  // Get some global stuff in here, locally.  Magariile astea contin
  // info despre all subblocks:
  t_subblock **subblock_inf;

  int max_subblocks_per_block;
  int subblock_lut_size;
  subblock_inf = subblock_data.subblock_inf;
  max_subblocks_per_block = subblock_data.max_subblocks_per_block;
  subblock_lut_size = subblock_data.subblock_lut_size;

  int i=0,j=0,total_num_ffs=0,contor=0;
  tt_hgraph *g = new tt_hgraph(); //allocate object timing hgraph

  // ---------------------------------------------------------------
  // === First, allocate all vertices in verts[] for all IN OUT
  // PADS and all used pins of all clb-blocks.
  int pin=0,counting_vertices=0;
  int max_pin=0;
  for (i=0;i<num_blocks;i++) 
    {
      // Skip blocks which are PI and drive global nets, such
      // as pclk, etc. are skiped below when vertices are counted

      // Allocate memory for "*id_tt_hgraph_02" of each block.
      if (block[i].type == INPAD || block[i].type == OUTPAD)
	max_pin = 1;
      else
	max_pin = pins_per_clb;
      // Obs:  below I use this porcarie my_alloc because "new" 
      // screws up all the rest and leads to segm-fault.  Lesson: do
      // *not* mix malloc with new!
      block[i].id_tt_hgraph_02 = (int *) my_malloc(max_pin * sizeof(int));
      for (pin=0;pin<max_pin;pin++)//PINS
	{
	  // If a pin is used create Vertex for it.
	  // Here some "opin" may be the output of a subblock 
	  // connected or "sent" thru a FF!
	  if( ( block[i].nets[pin] != OPEN )&&( pin != pins_per_clb-1 ) &&
	      ( is_global[ block[i].nets[pin] ] == false ) )
	    // Clock as global net is normally an INPAD.  I do not create
	    // a vertex for it because I also do not create the global net; 
	    // it would mess with my timing analyzer.
	    // Here: I will not create any
	    // vertex in my timing-graph for the last ipin of a block,
	    // which may be connected to a ".global" clock net in
	    // sequential circuits.
	    {
	      block[i].id_tt_hgraph_02[pin] = counting_vertices;
	      Vertex t_vertex;
	      // - add this vertex to verts[] of tt_hgraph type _02. 
	      // It also increments NumVerts of "g"
	      g->addVertex( t_vertex, counting_vertices );
	      counting_vertices++;//increment for next one if any
	    }//if
	}//for
	
    }//for
  printf("\nTotal number of verts in type _02 timing hgraph = %d,",
	 counting_vertices);


  // ---------------------------------------------------------------
  // === Second, populate timing graph.
  // PI/PO pads are also vertices. 
  int source_block_id=0,sink_block_id=0;
  int source_block_id_top_level=0;
  int sink_block_id_top_level=0;
  int blk_pin_of_source_block=0;
  int blk_pin_of_sink_block=0;

  for (i=0;i<num_nets;i++) {
   if (is_global[i] == false) { 
   // Global nets are not created in my timing graph.
   // That is why I will have a num_nets-1 nets in my timing graph!
   //..............................................................
    //printf(" %d %d ",i,net[i].blocks[0]);
    for (j=0;j<net[i].num_pins;j++) {//go thru all terminals
      if (j == 0) { // --- source_block_id
	source_block_id = net[i].blocks[j];
	if ( block[source_block_id].type == INPAD )
	  {
	    source_block_id_top_level =
	      block[source_block_id].id_tt_hgraph_02[0];//pi
	    //printf(" %d",source_block_id_top_level);
	    // will take place only once:
	    g->verts[source_block_id_top_level].setType( INPAD );
	    g->verts[source_block_id_top_level].blockID=source_block_id;
	  }//if
	
	if ( block[source_block_id].type == CLB )
	  {
	    // source_block_id_top_level is given by the
	    // id_tt_hgraph_02[] of the opin of the block
	    // which drives this net. so, get first blk_pin, which
	    // should be a "is_opin".
	    blk_pin_of_source_block = net[i].blk_pin[j];
	    source_block_id_top_level = //pi
	      block[source_block_id].id_tt_hgraph_02[blk_pin_of_source_block];
	    //printf(" %d",source_block_id_top_level);
	    // will take place only once:
	    g->verts[source_block_id_top_level].setType( CLB );
	    g->verts[source_block_id_top_level].blockID=source_block_id;
	  }//if
	// Record in vertex of tim_hg the driven-net index/id of 
	// fpga3d object .
	g->verts[source_block_id_top_level].netID_fpga3d = i;
	//printf(" %d %d |",source_block_id_top_level,i);

      }//if (j == 0)
      else {      // --- sink_block_id
	sink_block_id = net[i].blocks[j];
	if ( block[sink_block_id].type == OUTPAD )
	  {
	    sink_block_id_top_level =
	      block[sink_block_id].id_tt_hgraph_02[0];//po
	    //printf(" %d",sink_block_id_top_level);
	    // will take place only once:
	    g->verts[sink_block_id_top_level].setType( OUTPAD );
	    g->verts[sink_block_id_top_level].blockID=sink_block_id;
	    // Add fanin and fanout elements accordingly:
	    g->verts[source_block_id_top_level].addElementInFanout( 
			sink_block_id_top_level );
	    g->verts[sink_block_id_top_level].addElementInFanin( 
			source_block_id_top_level );
		}//if
	
	if ( block[sink_block_id].type == CLB )
	  {
	    // sink_block_id_top_level is given by the
	    // id_tt_hgraph_02[] of the ipin driven
	    // by this net. so, get first blk_pin, which
	    // should be a "is_ipin".
	    blk_pin_of_sink_block = net[i].blk_pin[j];
	    sink_block_id_top_level = //po
	      block[sink_block_id].id_tt_hgraph_02[blk_pin_of_sink_block];
	    g->verts[sink_block_id_top_level].setType( CLB );
	    g->verts[sink_block_id_top_level].blockID=sink_block_id;
	    // Add fanin and fanout elements accordingly:
	    g->verts[source_block_id_top_level].addElementInFanout( 
			sink_block_id_top_level );
	    g->verts[sink_block_id_top_level].addElementInFanin( 
			source_block_id_top_level );
	    
	  }//if
      }//else
    }//for inner
   //..............................................................
   }//if
  }//for outer

  // ---------------------------------------------------------------
  // In momentul asta inca sint nets (i.e., fanins and fanouts) care nu 
  // sint complete: alea interne fiecarui block!  Orice opin trebuie 
  // sa aiba fanin numai ipin care sint in "con" iar fiecare ipin (used,
  // adica care este facut vertex) tre sa aiba ca fanout opin corespunzator.
  // Aceste nets au delay fix/constant dat de max nr de LUT in serie de la
  // fanin ipins la opin respectiv.
  int opin=0;  
  int isub1=0;
  // Counts no of series LUT between opin and its fanin ipin(s):
  int series_fixed_delay_up_to_now=0;
  int subblk_id_driving_opin=0;
  int num_subblocks;
  int sink_opin_id_top_level=0;

  for (i=0;i<num_blocks;i++) {
    if ( block[i].type == CLB ) {//go thru all clbs only

      // Check all opins of this block. First pins [0..max_] are ipins,
      // then the opins, and lastly the clock.
      for (opin=(pins_per_clb - 2);//because (pins_per_clb - 1) is CLOCK
	   opin>=(pins_per_clb - (1 + max_subblocks_per_block));opin--) {
	//printf(" %d",block[i].nets[opin]);
	if ( block[i].nets[opin] != OPEN ) {
	  // Means that this opin is a net terminal.  Let's see
	  // now which ipins are its fanins.  Every fanin ipin
	  // forms the driving source of a net with fixed delay
	  // determined by the no of series LUT between that 
	  // ipin and this opin.
	  // First find the subblock whose output drives this 
	  // block opin:
	  num_subblocks = 
	    subblock_data.num_subblocks_per_block[ i ];
	  for (isub1=0;isub1<num_subblocks;isub1++)//SUBBBLOCK(s)
	    {
	      if (subblock_inf[i][isub1].output == opin) {
		subblk_id_driving_opin = isub1;
		// ..................................................
		// I have to look if this opin
		// of current block is driven from inside the block
		// by the out of a "subblock", which is latched (i.e.,
		// sent thru a FF); if so, then I will set the flag
		// "FF" of the vertex to 1, to recognize that this
		// vertex is driven thru a FF internal to this block.
		// Such a vertex will translate into two "virtual" 
		// nodes (PO + PI) during timing analysis, and paths
		// should stop at such a vertex and new paths should
		// start from here (these virtual vertices 
		// form same vertex which is in some sense split) 
		// So, see if this subblock_output is sent to
		// current block opin thru a FF.
		if (subblock_inf[i][isub1].clock != OPEN) {
		  // In this case flag the vertex in my timing-graph
		  // accordingly.
		  g->verts[ block[i].id_tt_hgraph_02[opin] ].FF = 1;
		  total_num_ffs = total_num_ffs + 1;
		}// yes, it's sent thru a FF
		// ..................................................
	      }//if
	    }//for
	  // At this moment subblk_id_driving_opin should have
	  // corect subblock id.
	  series_fixed_delay_up_to_now = 1;
	  // Used for all recursions of same block
	  sink_opin_id_top_level = 
	    block[i].id_tt_hgraph_02[opin];
	  
	  // Call recursive function which takes care of the rest of
	  // this block.  it will add an hedge for all connections
	  // ipin-opin of this block, in the timing hgraph. 
	  rec_finish_building_fanin_fanouts_of_verts_dueto_this_block_02 (
	      i, &subblock_data, g, 
	      subblk_id_driving_opin, series_fixed_delay_up_to_now,
	      sink_opin_id_top_level, timing_inf);
	  
	}//if
      }//for
    }//if
  }//for
  printf(" out of which %d are driven thru FFs inside CLBs.",
	 total_num_ffs);

  // ---------------------------------------------------------------

  // Build all nets.
  g->BuildNets();
  //g->Print_tt_hgraph( "dbg_tim_graph_02.echo" );
  // ---------------------------------------------------------------
  printf("\nDone create_timing_tt_hgraph_02.");

  return g; 
}//eof
// ======================================================================= //

// ======================================================================= //
void s_fpga3d::rec_finish_building_fanin_fanouts_of_verts_dueto_this_block_02 (
	       int block_id,t_subblock_data *subblock_data_ptr,tt_hgraph *g,
               int subblk_id,int fixed_delay_up_to_now,
	       int sink_opin_id_top_level,t_timing_inf timing_inf )
{
  // After going through all nets and building tt_hgraph there
  // are left nets with constant delay = nets from vertices representing 
  // ipins of blocks going to opin of same block.  Opin is also vertex.
  // Delay of these nets is dictated by the number of LUT in series.
  // This routine is used for detecting all fanins in the cone
  // with tip (i.e., virf) "subblk_id", which is at its turn on a path 
  // from "opin" of block_id backwards to ipins of block_id.

  // ---------------------------------------------------------------
  // Get some global stuff in here, locally.  Magariile astea contin
  // info despre all subblocks:
  int subblock_lut_size;
  int max_subblocks_per_block;
  int num_subblocks;
  t_subblock *subblock_inf;
  subblock_lut_size = subblock_data_ptr->subblock_lut_size;
  max_subblocks_per_block = subblock_data_ptr->max_subblocks_per_block;
  num_subblocks = subblock_data_ptr->num_subblocks_per_block[ block_id ];
  subblock_inf = subblock_data_ptr->subblock_inf[ block_id ];

  // ---------------------------------------------------------------
  int subblk_ipin=0;
  int subblk_id_driving_subblock=0;
  int source_ipin_id_top_level=0;
  int loc_inputs_of_pin=0;
  t_T_subblock T_subblock;// vezi s_fpga3d::print_arch() in read_arch.cpp


  for (subblk_ipin=0;subblk_ipin<subblock_lut_size;subblk_ipin++) {
    if (subblock_inf[subblk_id].inputs[subblk_ipin] != OPEN) {
      // Means this subblk_ipin is used.  Let's see if it's
      // driven by another subblock.
      loc_inputs_of_pin = 
	subblock_inf[subblk_id].inputs[subblk_ipin];
      //printf(" %d",loc_inputs_of_pin);
      // ...........................................................
      if ( loc_inputs_of_pin >= pins_per_clb ) { // --- other subblock

	// Means that this subblock input does not come from a 
	// clb ipin but rather from the out of other 
	// subbblock, whose index inside this block, as
	// subblock is:
	subblk_id_driving_subblock = 
	  loc_inputs_of_pin - pins_per_clb;
	//printf(" %d",subblk_id_driving_subblock);
	// Now, call recursively same function for the new 
	// found subblock in this cone
	rec_finish_building_fanin_fanouts_of_verts_dueto_this_block_02 (
	    block_id, subblock_data_ptr, g, 
	    subblk_id_driving_subblock, (fixed_delay_up_to_now+1),
            sink_opin_id_top_level, timing_inf );

      }
      // ...........................................................
      else { // --- driven by ipin of block

	// Means we have to add fanin - fanout
	source_ipin_id_top_level =
	  block[block_id].id_tt_hgraph_02[loc_inputs_of_pin];

	// Add fanin and fanout elements accordingly:
	//printf(" %d %d ",source_ipin_id_top_level,sink_opin_id_top_level);
	g->verts[source_ipin_id_top_level].addElementInFanout( 
		sink_opin_id_top_level );
	g->verts[sink_opin_id_top_level].addElementInFanin( 
		source_ipin_id_top_level );
	// Also set fixed delay of the hedge between ipin and
	// opin of this block.  This is like a wire with constant
	// delay given by the no of series LUT.  This fixed delay
	// is stored/kept untouched in ".miu" of fanout element.

	// Delay from subblock input to subblock output is
	// stored in: T_subblock.T_comb, 
	// which I have to multiply with num of subblocks in series.
	// For now I consider working with a clb with a single 
	// subblock.  For more complex architectures I will
	// have to come and sophisticate acest delay.
	// So, extract the timing information from loaction "0":
	T_subblock = timing_inf.T_subblock[0];
	// Set the delay.
	g->verts[source_ipin_id_top_level].
	   fOUTverts[sink_opin_id_top_level].miu = 
	      (double)( fixed_delay_up_to_now * T_subblock.T_comb );

	// Set the type of delay of net driven by this vertex:
	g->verts[source_ipin_id_top_level].delay_type = FIXED_DELAY;
	//printf(" %d",fixed_delay_up_to_now);

      }//if else
      // ...........................................................
      // Recursion ends when all inputs of subblk_id are coming
      // from ipins of block or are nor used (i.e., are OPEN).
    }//if
  }//for
  // ---------------------------------------------------------------

  
}//eof
// ======================================================================= //



// ======================================================================= //
  // allocates and stores delta arrays generated for largest circuit "pdc"
  // with vpr.  the above functions to generate these tables locally
  // have some bugs and do not work properly at this moment...
  // i will work on them after iccad deadline.


  // delta_clb_to_clb for "clma.net"
  double delta_clb_to_clb_hard[92][92] = {
  0.00e+00,  2.86e-09,  3.43e-09,  3.43e-09,  3.43e-09,  3.43e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  5.32e-09,  5.32e-09,  5.32e-09,  5.32e-09,  5.32e-09,  5.32e-09,  6.26e-09,  6.26e-09,  6.26e-09,  6.26e-09,  6.26e-09,  6.26e-09,  7.21e-09,  7.21e-09,  7.21e-09,  7.21e-09,  7.21e-09,  7.21e-09,  8.15e-09,  8.15e-09,  8.15e-09,  8.15e-09,  8.15e-09,  8.15e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.64e-08,  1.64e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.63e-08,  1.70e-08,  1.69e-08,
  2.19e-09,  2.86e-09,  3.43e-09,  3.43e-09,  3.43e-09,  3.43e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  5.32e-09,  5.32e-09,  5.32e-09,  5.32e-09,  5.32e-09,  5.32e-09,  6.26e-09,  6.26e-09,  6.26e-09,  6.26e-09,  6.26e-09,  6.26e-09,  7.21e-09,  7.21e-09,  7.21e-09,  7.21e-09,  7.21e-09,  7.21e-09,  8.15e-09,  8.15e-09,  8.15e-09,  8.15e-09,  8.15e-09,  8.15e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.64e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.66e-08,  1.67e-08,  1.71e-08,  1.87e-08,
  2.50e-09,  2.86e-09,  3.44e-09,  3.44e-09,  3.44e-09,  3.44e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.64e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.66e-08,  1.67e-08,  1.73e-08,  1.87e-08,
  2.50e-09,  3.44e-09,  3.44e-09,  3.44e-09,  3.44e-09,  3.44e-09,  4.44e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  5.38e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  6.32e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  7.27e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  8.21e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  9.16e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  1.01e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.20e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.39e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.58e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.67e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.66e-08,  1.67e-08,  1.75e-08,  1.87e-08,
  2.50e-09,  3.44e-09,  3.44e-09,  3.44e-09,  3.44e-09,  3.44e-09,  4.44e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  5.38e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  6.32e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  7.27e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  8.21e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  9.16e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  1.01e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.20e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.39e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.58e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.67e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.66e-08,  1.67e-08,  1.77e-08,  1.87e-08,
  2.50e-09,  3.44e-09,  3.44e-09,  3.44e-09,  3.44e-09,  3.44e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.64e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.66e-08,  1.67e-08,  1.74e-08,  1.87e-08,
  3.48e-09,  3.44e-09,  3.44e-09,  3.44e-09,  3.44e-09,  3.44e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.64e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.66e-08,  1.67e-08,  1.74e-08,  1.74e-08,
  3.49e-09,  4.43e-09,  4.43e-09,  4.43e-09,  4.43e-09,  4.43e-09,  5.37e-09,  5.37e-09,  5.37e-09,  5.37e-09,  5.37e-09,  5.37e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  7.26e-09,  7.26e-09,  7.26e-09,  7.26e-09,  7.26e-09,  7.26e-09,  8.20e-09,  8.20e-09,  8.20e-09,  8.20e-09,  8.20e-09,  8.20e-09,  9.15e-09,  9.15e-09,  9.15e-09,  9.15e-09,  9.15e-09,  9.15e-09,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.74e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.76e-08,  1.74e-08,  1.84e-08,  1.96e-08,
  3.49e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.74e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.76e-08,  1.75e-08,  1.84e-08,  1.84e-08,
  3.49e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.76e-08,  1.75e-08,  1.84e-08,  1.96e-08,
  3.49e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.76e-08,  1.76e-08,  1.84e-08,  1.84e-08,
  3.49e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.76e-08,  1.77e-08,  1.84e-08,  1.96e-08,
  4.48e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.74e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.76e-08,  1.77e-08,  1.84e-08,  1.84e-08,
  4.49e-09,  5.42e-09,  5.42e-09,  5.42e-09,  5.42e-09,  5.42e-09,  6.37e-09,  6.37e-09,  6.37e-09,  6.37e-09,  6.37e-09,  6.37e-09,  7.31e-09,  7.31e-09,  7.31e-09,  7.31e-09,  7.31e-09,  7.31e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  9.20e-09,  9.20e-09,  9.20e-09,  9.20e-09,  9.20e-09,  9.20e-09,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.84e-08,  1.94e-08,  2.06e-08,
  4.49e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.85e-08,  1.94e-08,  2.06e-08,
  4.49e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.85e-08,  1.94e-08,  2.06e-08,
  4.49e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.86e-08,  1.94e-08,  2.06e-08,
  4.49e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.86e-08,  1.94e-08,  2.06e-08,
  5.47e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.86e-08,  1.94e-08,  1.95e-08,
  5.48e-09,  6.42e-09,  6.42e-09,  6.42e-09,  6.42e-09,  6.42e-09,  7.36e-09,  7.36e-09,  7.36e-09,  7.36e-09,  7.36e-09,  7.36e-09,  8.31e-09,  8.31e-09,  8.31e-09,  8.31e-09,  8.31e-09,  8.31e-09,  9.25e-09,  9.25e-09,  9.25e-09,  9.25e-09,  9.25e-09,  9.25e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.94e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  1.94e-08,  2.04e-08,  2.16e-08,
  5.48e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.94e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  1.95e-08,  2.04e-08,  2.16e-08,
  5.48e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  1.95e-08,  2.04e-08,  2.16e-08,
  5.48e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  1.96e-08,  2.04e-08,  2.16e-08,
  5.48e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  1.96e-08,  2.04e-08,  2.16e-08,
  6.47e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.94e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  1.96e-08,  2.04e-08,  2.06e-08,
  6.48e-09,  7.41e-09,  7.41e-09,  7.41e-09,  7.41e-09,  7.41e-09,  8.36e-09,  8.36e-09,  8.36e-09,  8.36e-09,  8.36e-09,  8.36e-09,  9.30e-09,  9.30e-09,  9.30e-09,  9.30e-09,  9.30e-09,  9.30e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  2.04e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.04e-08,  2.13e-08,  2.17e-08,
  6.48e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  2.04e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.04e-08,  2.13e-08,  2.26e-08,
  6.48e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.05e-08,  2.13e-08,  2.26e-08,
  6.48e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.13e-08,  2.26e-08,
  6.48e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.13e-08,  2.26e-08,
  7.47e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  2.04e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.13e-08,  2.17e-08,
  7.47e-09,  8.41e-09,  8.41e-09,  8.41e-09,  8.41e-09,  8.41e-09,  9.35e-09,  9.35e-09,  9.35e-09,  9.35e-09,  9.35e-09,  9.35e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  7.47e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  7.47e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  7.47e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  7.47e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  8.46e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  8.47e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  8.47e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  8.47e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  8.47e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  8.47e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  9.46e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  9.47e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  9.47e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  9.47e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  9.47e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  9.47e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.05e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.05e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.05e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.05e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.05e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.05e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.71e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.81e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.90e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.99e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.74e-08,  1.74e-08,  1.74e-08,  1.74e-08,  1.74e-08,  1.80e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.90e-08,  1.93e-08,  1.93e-08,  1.93e-08,  1.93e-08,  1.93e-08,  1.99e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.80e-08,  1.81e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.83e-08,  1.89e-08,  1.90e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.93e-08,  1.99e-08,  2.00e-08,  2.00e-08,  2.01e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.80e-08,  1.81e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.83e-08,  1.89e-08,  1.90e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.93e-08,  1.99e-08,  2.00e-08,  2.00e-08,  2.01e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.80e-08,  1.81e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.83e-08,  1.89e-08,  1.90e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.93e-08,  1.99e-08,  2.00e-08,  2.00e-08,  2.01e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.80e-08,  1.81e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.83e-08,  1.89e-08,  1.90e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.93e-08,  1.99e-08,  2.00e-08,  2.00e-08,  2.01e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.80e-08,  1.81e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.83e-08,  1.89e-08,  1.90e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.93e-08,  1.99e-08,  2.00e-08,  2.00e-08,  2.01e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.78e-08,  1.79e-08,  1.85e-08,  1.86e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.46e-08
  };


  // delta_clb_to_clb_2 for "pdc.net"
  double delta_clb_to_clb_2_hard[92][92] = {
  0.00e+00,  2.10e-09,  2.86e-09,  3.44e-09,  3.44e-09,  3.44e-09,  3.44e-09,  4.44e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  5.38e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  6.32e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  7.27e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  8.21e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  9.16e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  1.01e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.20e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.39e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.58e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.60e-08,  1.61e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.63e-08,  1.68e-08,
  2.19e-09,  2.19e-09,  2.86e-09,  3.44e-09,  3.44e-09,  3.44e-09,  3.44e-09,  4.44e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  5.38e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  6.32e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  7.27e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  8.21e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  9.16e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  1.01e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.20e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.39e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.58e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.62e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.66e-08,  1.67e-08,  1.70e-08,
  2.50e-09,  2.50e-09,  2.86e-09,  3.44e-09,  3.44e-09,  3.44e-09,  3.44e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.66e-08,  1.67e-08,  1.71e-08,
  2.50e-09,  2.50e-09,  3.44e-09,  3.44e-09,  3.44e-09,  3.44e-09,  3.44e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.65e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.66e-08,  1.67e-08,  1.73e-08,
  2.50e-09,  2.50e-09,  3.44e-09,  3.44e-09,  3.44e-09,  3.44e-09,  3.44e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.67e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.66e-08,  1.67e-08,  1.74e-08,
  2.50e-09,  2.50e-09,  3.44e-09,  3.44e-09,  3.44e-09,  3.44e-09,  3.44e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.66e-08,  1.67e-08,  1.76e-08,
  3.49e-09,  3.49e-09,  3.44e-09,  3.44e-09,  3.44e-09,  3.44e-09,  3.44e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  4.38e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  5.33e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  6.27e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  7.22e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  8.16e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  9.10e-09,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.00e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.19e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.66e-08,  1.67e-08,  1.72e-08,
  3.49e-09,  3.49e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.65e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.76e-08,  1.75e-08,  1.75e-08,
  3.49e-09,  3.49e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.65e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.76e-08,  1.75e-08,  1.75e-08,
  3.49e-09,  3.49e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.76e-08,  1.75e-08,  1.75e-08,
  3.49e-09,  3.49e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.76e-08,  1.77e-08,  1.75e-08,
  3.49e-09,  3.49e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.76e-08,  1.77e-08,  1.82e-08,
  4.49e-09,  4.49e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  4.44e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  5.38e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  6.32e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  7.27e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  8.21e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  9.16e-09,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.01e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.10e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.29e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.48e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.76e-08,  1.77e-08,  1.82e-08,
  4.49e-09,  4.49e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.75e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.85e-08,  1.85e-08,
  4.49e-09,  4.49e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.75e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.85e-08,  1.85e-08,
  4.49e-09,  4.49e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.85e-08,  1.85e-08,
  4.49e-09,  4.49e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.86e-08,  1.85e-08,
  4.49e-09,  4.49e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.86e-08,  1.92e-08,
  5.48e-09,  5.48e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  5.43e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  6.38e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  7.32e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  8.26e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  9.21e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.20e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.39e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.58e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.86e-08,  1.92e-08,
  5.48e-09,  5.48e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.85e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  1.95e-08,  1.95e-08,
  5.48e-09,  5.48e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.85e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  1.95e-08,  1.95e-08,
  5.48e-09,  5.48e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  1.95e-08,  1.95e-08,
  5.48e-09,  5.48e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.95e-08,
  5.48e-09,  5.48e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  1.96e-08,  2.02e-08,
  6.48e-09,  6.48e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.87e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  1.96e-08,  2.02e-08,
  6.48e-09,  6.48e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.95e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.05e-08,  2.26e-08,
  6.48e-09,  6.48e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.95e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.05e-08,  2.26e-08,
  6.48e-09,  6.48e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.05e-08,  2.26e-08,
  6.48e-09,  6.48e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.26e-08,
  6.48e-09,  6.48e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.26e-08,
  7.47e-09,  7.47e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.26e-08,
  7.47e-09,  7.47e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  7.47e-09,  7.47e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  7.47e-09,  7.47e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  7.47e-09,  7.47e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  7.47e-09,  7.47e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  8.47e-09,  8.47e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  8.47e-09,  8.47e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  8.47e-09,  8.47e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  8.47e-09,  8.47e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  8.47e-09,  8.47e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  8.47e-09,  8.47e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  9.47e-09,  9.47e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  9.47e-09,  9.47e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  9.47e-09,  9.47e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  9.47e-09,  9.47e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  9.47e-09,  9.47e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  9.47e-09,  9.47e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.05e-08,  1.05e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.05e-08,  1.05e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.05e-08,  1.05e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.05e-08,  1.05e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.05e-08,  1.05e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.05e-08,  1.05e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.82e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  2.01e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.81e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.91e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  2.00e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.71e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.81e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.90e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.99e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.70e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.80e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.89e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.99e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.80e-08,  1.81e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.83e-08,  1.89e-08,  1.90e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.93e-08,  1.99e-08,  2.00e-08,  2.00e-08,  2.01e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.80e-08,  1.81e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.83e-08,  1.89e-08,  1.90e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.93e-08,  1.99e-08,  2.00e-08,  2.00e-08,  2.01e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.80e-08,  1.81e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.83e-08,  1.89e-08,  1.90e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.93e-08,  1.99e-08,  2.00e-08,  2.00e-08,  2.01e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.80e-08,  1.81e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.83e-08,  1.89e-08,  1.90e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.93e-08,  1.99e-08,  2.00e-08,  2.00e-08,  2.01e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.80e-08,  1.81e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.83e-08,  1.89e-08,  1.90e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.93e-08,  1.99e-08,  2.00e-08,  2.00e-08,  2.01e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  1.26e-08,  1.26e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.80e-08,  1.81e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.83e-08,  1.89e-08,  1.90e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.93e-08,  1.99e-08,  2.00e-08,  2.00e-08,  2.01e-08,  2.02e-08,  2.02e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,
  0.00e+00,  1.26e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.78e-08,  1.79e-08,  1.85e-08,  1.86e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08,  2.26e-08
  };


  // delta_inpad_to_clb
  double delta_inpad_to_clb_hard[93][93] = {
 -1.00e+00,  2.07e-09,  2.75e-09,  3.01e-09,  3.01e-09,  3.01e-09,  3.01e-09,  3.58e-09,  3.95e-09,  3.95e-09,  3.95e-09,  3.95e-09,  3.95e-09,  4.52e-09,  4.90e-09,  4.90e-09,  4.90e-09,  4.90e-09,  4.90e-09,  5.46e-09,  5.84e-09,  5.84e-09,  5.84e-09,  5.84e-09,  5.84e-09,  6.41e-09,  6.79e-09,  6.79e-09,  6.79e-09,  6.79e-09,  6.79e-09,  7.35e-09,  7.73e-09,  7.73e-09,  7.73e-09,  7.73e-09,  7.73e-09,  8.30e-09,  8.67e-09,  8.67e-09,  8.67e-09,  8.67e-09,  8.67e-09,  9.24e-09,  9.62e-09,  9.62e-09,  9.62e-09,  9.62e-09,  9.62e-09,  1.02e-08,  1.06e-08,  1.06e-08,  1.06e-08,  1.06e-08,  1.06e-08,  1.11e-08,  1.15e-08,  1.15e-08,  1.15e-08,  1.15e-08,  1.15e-08,  1.21e-08,  1.25e-08,  1.25e-08,  1.25e-08,  1.25e-08,  1.25e-08,  1.30e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.40e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.49e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.58e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.68e-08,  1.69e-08,
  2.07e-09,  2.15e-09,  2.24e-09,  2.33e-09,  2.42e-09,  2.52e-09,  3.10e-09,  3.18e-09,  3.27e-09,  3.36e-09,  3.45e-09,  3.55e-09,  4.13e-09,  4.21e-09,  4.30e-09,  4.39e-09,  4.48e-09,  4.58e-09,  5.16e-09,  5.24e-09,  5.33e-09,  5.42e-09,  5.51e-09,  5.61e-09,  6.19e-09,  6.27e-09,  6.36e-09,  6.45e-09,  6.54e-09,  6.64e-09,  7.22e-09,  7.30e-09,  7.39e-09,  7.48e-09,  7.57e-09,  7.67e-09,  8.25e-09,  8.33e-09,  8.42e-09,  8.51e-09,  8.60e-09,  8.70e-09,  9.28e-09,  9.36e-09,  9.45e-09,  9.54e-09,  9.63e-09,  9.73e-09,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.08e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.18e-08,  1.24e-08,  1.25e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.69e-08,
  2.76e-09,  2.76e-09,  2.86e-09,  3.23e-09,  3.32e-09,  3.40e-09,  3.51e-09,  4.08e-09,  4.17e-09,  4.26e-09,  4.35e-09,  4.44e-09,  4.54e-09,  5.11e-09,  5.20e-09,  5.29e-09,  5.38e-09,  5.47e-09,  5.57e-09,  6.14e-09,  6.23e-09,  6.32e-09,  6.41e-09,  6.50e-09,  6.60e-09,  7.17e-09,  7.26e-09,  7.35e-09,  7.44e-09,  7.53e-09,  7.63e-09,  8.20e-09,  8.29e-09,  8.38e-09,  8.47e-09,  8.56e-09,  8.66e-09,  9.23e-09,  9.32e-09,  9.41e-09,  9.50e-09,  9.59e-09,  9.69e-09,  1.03e-08,  1.04e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.16e-08,  1.18e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.28e-08,  1.34e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.59e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.85e-08,  1.70e-08,
  3.05e-09,  3.05e-09,  3.14e-09,  3.23e-09,  3.32e-09,  3.40e-09,  3.51e-09,  4.08e-09,  4.17e-09,  4.26e-09,  4.35e-09,  4.44e-09,  4.54e-09,  5.11e-09,  5.20e-09,  5.29e-09,  5.38e-09,  5.47e-09,  5.57e-09,  6.14e-09,  6.23e-09,  6.32e-09,  6.41e-09,  6.50e-09,  6.60e-09,  7.17e-09,  7.26e-09,  7.35e-09,  7.44e-09,  7.53e-09,  7.63e-09,  8.20e-09,  8.29e-09,  8.38e-09,  8.47e-09,  8.56e-09,  8.66e-09,  9.23e-09,  9.32e-09,  9.41e-09,  9.50e-09,  9.59e-09,  9.69e-09,  1.03e-08,  1.04e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.16e-08,  1.18e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.28e-08,  1.34e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.59e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.85e-08,  1.70e-08,
  3.05e-09,  3.05e-09,  3.14e-09,  3.23e-09,  3.32e-09,  3.40e-09,  3.51e-09,  4.08e-09,  4.17e-09,  4.26e-09,  4.35e-09,  4.44e-09,  4.54e-09,  5.11e-09,  5.20e-09,  5.29e-09,  5.38e-09,  5.47e-09,  5.57e-09,  6.14e-09,  6.23e-09,  6.32e-09,  6.41e-09,  6.50e-09,  6.60e-09,  7.17e-09,  7.26e-09,  7.35e-09,  7.44e-09,  7.53e-09,  7.63e-09,  8.20e-09,  8.29e-09,  8.38e-09,  8.47e-09,  8.56e-09,  8.66e-09,  9.23e-09,  9.32e-09,  9.41e-09,  9.50e-09,  9.59e-09,  9.69e-09,  1.03e-08,  1.04e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.16e-08,  1.18e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.28e-08,  1.34e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.59e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.85e-08,  1.71e-08,
  3.05e-09,  3.05e-09,  3.14e-09,  3.23e-09,  3.32e-09,  3.40e-09,  3.51e-09,  4.08e-09,  4.17e-09,  4.26e-09,  4.35e-09,  4.44e-09,  4.54e-09,  5.11e-09,  5.20e-09,  5.29e-09,  5.38e-09,  5.47e-09,  5.57e-09,  6.14e-09,  6.23e-09,  6.32e-09,  6.41e-09,  6.50e-09,  6.60e-09,  7.17e-09,  7.26e-09,  7.35e-09,  7.44e-09,  7.53e-09,  7.63e-09,  8.20e-09,  8.29e-09,  8.38e-09,  8.47e-09,  8.56e-09,  8.66e-09,  9.23e-09,  9.32e-09,  9.41e-09,  9.50e-09,  9.59e-09,  9.69e-09,  1.03e-08,  1.04e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.16e-08,  1.18e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.28e-08,  1.34e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.59e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.85e-08,  1.72e-08,
  3.05e-09,  3.05e-09,  3.14e-09,  3.23e-09,  3.32e-09,  3.40e-09,  3.51e-09,  4.08e-09,  4.17e-09,  4.26e-09,  4.35e-09,  4.44e-09,  4.54e-09,  5.11e-09,  5.20e-09,  5.29e-09,  5.38e-09,  5.47e-09,  5.57e-09,  6.14e-09,  6.23e-09,  6.32e-09,  6.41e-09,  6.50e-09,  6.60e-09,  7.17e-09,  7.26e-09,  7.35e-09,  7.44e-09,  7.53e-09,  7.63e-09,  8.20e-09,  8.29e-09,  8.38e-09,  8.47e-09,  8.56e-09,  8.66e-09,  9.23e-09,  9.32e-09,  9.41e-09,  9.50e-09,  9.59e-09,  9.69e-09,  1.03e-08,  1.04e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.16e-08,  1.18e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.28e-08,  1.34e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.59e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.85e-08,  1.74e-08,
  3.61e-09,  4.00e-09,  4.08e-09,  4.17e-09,  4.26e-09,  4.35e-09,  4.45e-09,  4.94e-09,  5.03e-09,  5.12e-09,  5.20e-09,  5.29e-09,  5.40e-09,  5.89e-09,  5.97e-09,  6.06e-09,  6.15e-09,  6.24e-09,  6.34e-09,  6.83e-09,  6.92e-09,  7.00e-09,  7.09e-09,  7.18e-09,  7.29e-09,  7.77e-09,  7.86e-09,  7.95e-09,  8.04e-09,  8.13e-09,  8.23e-09,  8.72e-09,  8.81e-09,  8.89e-09,  8.98e-09,  9.07e-09,  9.17e-09,  9.66e-09,  9.75e-09,  9.84e-09,  9.93e-09,  1.00e-08,  1.01e-08,  1.06e-08,  1.07e-08,  1.08e-08,  1.09e-08,  1.10e-08,  1.11e-08,  1.16e-08,  1.16e-08,  1.17e-08,  1.18e-08,  1.19e-08,  1.20e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.28e-08,  1.28e-08,  1.30e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.39e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.63e-08,  1.64e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.74e-08,  1.78e-08,  1.79e-08,
  4.05e-09,  4.05e-09,  4.14e-09,  4.22e-09,  4.31e-09,  4.40e-09,  4.50e-09,  5.08e-09,  5.17e-09,  5.25e-09,  5.34e-09,  5.43e-09,  5.53e-09,  6.11e-09,  6.20e-09,  6.28e-09,  6.37e-09,  6.46e-09,  6.56e-09,  7.14e-09,  7.23e-09,  7.31e-09,  7.40e-09,  7.49e-09,  7.60e-09,  8.17e-09,  8.26e-09,  8.34e-09,  8.43e-09,  8.52e-09,  8.63e-09,  9.20e-09,  9.29e-09,  9.37e-09,  9.46e-09,  9.55e-09,  9.66e-09,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.13e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.44e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.85e-08,  1.86e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.95e-08,  1.80e-08,
  4.05e-09,  4.05e-09,  4.14e-09,  4.22e-09,  4.31e-09,  4.40e-09,  4.50e-09,  5.08e-09,  5.17e-09,  5.25e-09,  5.34e-09,  5.43e-09,  5.53e-09,  6.11e-09,  6.20e-09,  6.28e-09,  6.37e-09,  6.46e-09,  6.56e-09,  7.14e-09,  7.23e-09,  7.31e-09,  7.40e-09,  7.49e-09,  7.60e-09,  8.17e-09,  8.26e-09,  8.34e-09,  8.43e-09,  8.52e-09,  8.63e-09,  9.20e-09,  9.29e-09,  9.37e-09,  9.46e-09,  9.55e-09,  9.66e-09,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.13e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.44e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.85e-08,  1.86e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.95e-08,  1.81e-08,
  4.05e-09,  4.05e-09,  4.14e-09,  4.22e-09,  4.31e-09,  4.40e-09,  4.50e-09,  5.08e-09,  5.17e-09,  5.25e-09,  5.34e-09,  5.43e-09,  5.53e-09,  6.11e-09,  6.20e-09,  6.28e-09,  6.37e-09,  6.46e-09,  6.56e-09,  7.14e-09,  7.23e-09,  7.31e-09,  7.40e-09,  7.49e-09,  7.60e-09,  8.17e-09,  8.26e-09,  8.34e-09,  8.43e-09,  8.52e-09,  8.63e-09,  9.20e-09,  9.29e-09,  9.37e-09,  9.46e-09,  9.55e-09,  9.66e-09,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.13e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.44e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.85e-08,  1.86e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.95e-08,  1.82e-08,
  4.05e-09,  4.05e-09,  4.14e-09,  4.22e-09,  4.31e-09,  4.40e-09,  4.50e-09,  5.08e-09,  5.17e-09,  5.25e-09,  5.34e-09,  5.43e-09,  5.53e-09,  6.11e-09,  6.20e-09,  6.28e-09,  6.37e-09,  6.46e-09,  6.56e-09,  7.14e-09,  7.23e-09,  7.31e-09,  7.40e-09,  7.49e-09,  7.60e-09,  8.17e-09,  8.26e-09,  8.34e-09,  8.43e-09,  8.52e-09,  8.63e-09,  9.20e-09,  9.29e-09,  9.37e-09,  9.46e-09,  9.55e-09,  9.66e-09,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.13e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.44e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.85e-08,  1.86e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.95e-08,  1.83e-08,
  4.05e-09,  4.05e-09,  4.14e-09,  4.22e-09,  4.31e-09,  4.40e-09,  4.50e-09,  5.08e-09,  5.17e-09,  5.25e-09,  5.34e-09,  5.43e-09,  5.53e-09,  6.11e-09,  6.20e-09,  6.28e-09,  6.37e-09,  6.46e-09,  6.56e-09,  7.14e-09,  7.23e-09,  7.31e-09,  7.40e-09,  7.49e-09,  7.60e-09,  8.17e-09,  8.26e-09,  8.34e-09,  8.43e-09,  8.52e-09,  8.63e-09,  9.20e-09,  9.29e-09,  9.37e-09,  9.46e-09,  9.55e-09,  9.66e-09,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.13e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.44e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.85e-08,  1.86e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.95e-08,  1.84e-08,
  4.61e-09,  4.99e-09,  5.08e-09,  5.17e-09,  5.26e-09,  5.34e-09,  5.45e-09,  5.94e-09,  6.02e-09,  6.11e-09,  6.20e-09,  6.29e-09,  6.39e-09,  6.88e-09,  6.97e-09,  7.06e-09,  7.14e-09,  7.23e-09,  7.34e-09,  7.83e-09,  7.91e-09,  8.00e-09,  8.09e-09,  8.18e-09,  8.28e-09,  8.77e-09,  8.86e-09,  8.94e-09,  9.03e-09,  9.12e-09,  9.23e-09,  9.71e-09,  9.80e-09,  9.89e-09,  9.98e-09,  1.01e-08,  1.02e-08,  1.07e-08,  1.07e-08,  1.08e-08,  1.09e-08,  1.10e-08,  1.11e-08,  1.16e-08,  1.17e-08,  1.18e-08,  1.19e-08,  1.20e-08,  1.21e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.28e-08,  1.29e-08,  1.30e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.38e-08,  1.39e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.49e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.82e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.84e-08,  1.88e-08,  1.90e-08,
  5.04e-09,  5.04e-09,  5.13e-09,  5.22e-09,  5.31e-09,  5.40e-09,  5.50e-09,  6.07e-09,  6.16e-09,  6.25e-09,  6.34e-09,  6.43e-09,  6.53e-09,  7.10e-09,  7.19e-09,  7.28e-09,  7.37e-09,  7.46e-09,  7.56e-09,  8.13e-09,  8.22e-09,  8.31e-09,  8.40e-09,  8.49e-09,  8.59e-09,  9.17e-09,  9.25e-09,  9.34e-09,  9.43e-09,  9.52e-09,  9.62e-09,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.05e-08,  1.07e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.23e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.05e-08,  1.91e-08,
  5.04e-09,  5.04e-09,  5.13e-09,  5.22e-09,  5.31e-09,  5.40e-09,  5.50e-09,  6.07e-09,  6.16e-09,  6.25e-09,  6.34e-09,  6.43e-09,  6.53e-09,  7.10e-09,  7.19e-09,  7.28e-09,  7.37e-09,  7.46e-09,  7.56e-09,  8.13e-09,  8.22e-09,  8.31e-09,  8.40e-09,  8.49e-09,  8.59e-09,  9.17e-09,  9.25e-09,  9.34e-09,  9.43e-09,  9.52e-09,  9.62e-09,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.05e-08,  1.07e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.23e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.05e-08,  1.92e-08,
  5.04e-09,  5.04e-09,  5.13e-09,  5.22e-09,  5.31e-09,  5.40e-09,  5.50e-09,  6.07e-09,  6.16e-09,  6.25e-09,  6.34e-09,  6.43e-09,  6.53e-09,  7.10e-09,  7.19e-09,  7.28e-09,  7.37e-09,  7.46e-09,  7.56e-09,  8.13e-09,  8.22e-09,  8.31e-09,  8.40e-09,  8.49e-09,  8.59e-09,  9.17e-09,  9.25e-09,  9.34e-09,  9.43e-09,  9.52e-09,  9.62e-09,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.05e-08,  1.07e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.23e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.05e-08,  1.93e-08,
  5.04e-09,  5.04e-09,  5.13e-09,  5.22e-09,  5.31e-09,  5.40e-09,  5.50e-09,  6.07e-09,  6.16e-09,  6.25e-09,  6.34e-09,  6.43e-09,  6.53e-09,  7.10e-09,  7.19e-09,  7.28e-09,  7.37e-09,  7.46e-09,  7.56e-09,  8.13e-09,  8.22e-09,  8.31e-09,  8.40e-09,  8.49e-09,  8.59e-09,  9.17e-09,  9.25e-09,  9.34e-09,  9.43e-09,  9.52e-09,  9.62e-09,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.05e-08,  1.07e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.23e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.05e-08,  1.94e-08,
  5.04e-09,  5.04e-09,  5.13e-09,  5.22e-09,  5.31e-09,  5.40e-09,  5.50e-09,  6.07e-09,  6.16e-09,  6.25e-09,  6.34e-09,  6.43e-09,  6.53e-09,  7.10e-09,  7.19e-09,  7.28e-09,  7.37e-09,  7.46e-09,  7.56e-09,  8.13e-09,  8.22e-09,  8.31e-09,  8.40e-09,  8.49e-09,  8.59e-09,  9.17e-09,  9.25e-09,  9.34e-09,  9.43e-09,  9.52e-09,  9.62e-09,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.05e-08,  1.07e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.23e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.05e-08,  1.95e-08,
  5.60e-09,  5.99e-09,  6.08e-09,  6.16e-09,  6.25e-09,  6.34e-09,  6.44e-09,  6.93e-09,  7.02e-09,  7.11e-09,  7.20e-09,  7.28e-09,  7.39e-09,  7.88e-09,  7.96e-09,  8.05e-09,  8.14e-09,  8.23e-09,  8.33e-09,  8.82e-09,  8.91e-09,  9.00e-09,  9.08e-09,  9.17e-09,  9.28e-09,  9.77e-09,  9.85e-09,  9.94e-09,  1.00e-08,  1.01e-08,  1.02e-08,  1.07e-08,  1.08e-08,  1.09e-08,  1.10e-08,  1.11e-08,  1.12e-08,  1.17e-08,  1.17e-08,  1.18e-08,  1.19e-08,  1.20e-08,  1.21e-08,  1.26e-08,  1.27e-08,  1.28e-08,  1.29e-08,  1.29e-08,  1.31e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.39e-08,  1.40e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.47e-08,  1.48e-08,  1.49e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.59e-08,  1.64e-08,  1.65e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.83e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.92e-08,  1.93e-08,  1.93e-08,  1.93e-08,  1.93e-08,  1.93e-08,  1.98e-08,  2.01e-08,
  6.04e-09,  6.04e-09,  6.13e-09,  6.22e-09,  6.30e-09,  6.39e-09,  6.50e-09,  7.07e-09,  7.16e-09,  7.25e-09,  7.33e-09,  7.42e-09,  7.53e-09,  8.10e-09,  8.19e-09,  8.28e-09,  8.36e-09,  8.45e-09,  8.56e-09,  9.13e-09,  9.22e-09,  9.31e-09,  9.39e-09,  9.48e-09,  9.59e-09,  1.02e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.74e-08,  1.75e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.15e-08,  2.02e-08,
  6.04e-09,  6.04e-09,  6.13e-09,  6.22e-09,  6.30e-09,  6.39e-09,  6.50e-09,  7.07e-09,  7.16e-09,  7.25e-09,  7.33e-09,  7.42e-09,  7.53e-09,  8.10e-09,  8.19e-09,  8.28e-09,  8.36e-09,  8.45e-09,  8.56e-09,  9.13e-09,  9.22e-09,  9.31e-09,  9.39e-09,  9.48e-09,  9.59e-09,  1.02e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.74e-08,  1.75e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.15e-08,  2.03e-08,
  6.04e-09,  6.04e-09,  6.13e-09,  6.22e-09,  6.30e-09,  6.39e-09,  6.50e-09,  7.07e-09,  7.16e-09,  7.25e-09,  7.33e-09,  7.42e-09,  7.53e-09,  8.10e-09,  8.19e-09,  8.28e-09,  8.36e-09,  8.45e-09,  8.56e-09,  9.13e-09,  9.22e-09,  9.31e-09,  9.39e-09,  9.48e-09,  9.59e-09,  1.02e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.74e-08,  1.75e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.15e-08,  2.04e-08,
  6.04e-09,  6.04e-09,  6.13e-09,  6.22e-09,  6.30e-09,  6.39e-09,  6.50e-09,  7.07e-09,  7.16e-09,  7.25e-09,  7.33e-09,  7.42e-09,  7.53e-09,  8.10e-09,  8.19e-09,  8.28e-09,  8.36e-09,  8.45e-09,  8.56e-09,  9.13e-09,  9.22e-09,  9.31e-09,  9.39e-09,  9.48e-09,  9.59e-09,  1.02e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.74e-08,  1.75e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.15e-08,  2.05e-08,
  6.04e-09,  6.04e-09,  6.13e-09,  6.22e-09,  6.30e-09,  6.39e-09,  6.50e-09,  7.07e-09,  7.16e-09,  7.25e-09,  7.33e-09,  7.42e-09,  7.53e-09,  8.10e-09,  8.19e-09,  8.28e-09,  8.36e-09,  8.45e-09,  8.56e-09,  9.13e-09,  9.22e-09,  9.31e-09,  9.39e-09,  9.48e-09,  9.59e-09,  1.02e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.74e-08,  1.75e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.15e-08,  2.06e-08,
  6.60e-09,  6.98e-09,  7.07e-09,  7.16e-09,  7.25e-09,  7.34e-09,  7.44e-09,  7.93e-09,  8.02e-09,  8.10e-09,  8.19e-09,  8.28e-09,  8.38e-09,  8.87e-09,  8.96e-09,  9.05e-09,  9.14e-09,  9.22e-09,  9.33e-09,  9.82e-09,  9.90e-09,  9.99e-09,  1.01e-08,  1.02e-08,  1.03e-08,  1.08e-08,  1.08e-08,  1.09e-08,  1.10e-08,  1.11e-08,  1.12e-08,  1.17e-08,  1.18e-08,  1.19e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.26e-08,  1.27e-08,  1.28e-08,  1.29e-08,  1.30e-08,  1.31e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.39e-08,  1.39e-08,  1.40e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.49e-08,  1.50e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.57e-08,  1.58e-08,  1.59e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.74e-08,  1.75e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.93e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  2.02e-08,  2.03e-08,  2.03e-08,  2.03e-08,  2.03e-08,  2.03e-08,  2.08e-08,  2.12e-08,
  7.04e-09,  7.04e-09,  7.12e-09,  7.21e-09,  7.30e-09,  7.39e-09,  7.49e-09,  8.07e-09,  8.15e-09,  8.24e-09,  8.33e-09,  8.42e-09,  8.52e-09,  9.10e-09,  9.18e-09,  9.27e-09,  9.36e-09,  9.45e-09,  9.55e-09,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.12e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.25e-08,  1.26e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.15e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.25e-08,  2.13e-08,
  7.04e-09,  7.04e-09,  7.12e-09,  7.21e-09,  7.30e-09,  7.39e-09,  7.49e-09,  8.07e-09,  8.15e-09,  8.24e-09,  8.33e-09,  8.42e-09,  8.52e-09,  9.10e-09,  9.18e-09,  9.27e-09,  9.36e-09,  9.45e-09,  9.55e-09,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.12e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.25e-08,  1.26e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.15e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.25e-08,  2.14e-08,
  7.04e-09,  7.04e-09,  7.12e-09,  7.21e-09,  7.30e-09,  7.39e-09,  7.49e-09,  8.07e-09,  8.15e-09,  8.24e-09,  8.33e-09,  8.42e-09,  8.52e-09,  9.10e-09,  9.18e-09,  9.27e-09,  9.36e-09,  9.45e-09,  9.55e-09,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.12e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.25e-08,  1.26e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.15e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.25e-08,  2.15e-08,
  7.04e-09,  7.04e-09,  7.12e-09,  7.21e-09,  7.30e-09,  7.39e-09,  7.49e-09,  8.07e-09,  8.15e-09,  8.24e-09,  8.33e-09,  8.42e-09,  8.52e-09,  9.10e-09,  9.18e-09,  9.27e-09,  9.36e-09,  9.45e-09,  9.55e-09,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.12e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.25e-08,  1.26e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.15e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.25e-08,  2.16e-08,
  7.04e-09,  7.04e-09,  7.12e-09,  7.21e-09,  7.30e-09,  7.39e-09,  7.49e-09,  8.07e-09,  8.15e-09,  8.24e-09,  8.33e-09,  8.42e-09,  8.52e-09,  9.10e-09,  9.18e-09,  9.27e-09,  9.36e-09,  9.45e-09,  9.55e-09,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.12e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.25e-08,  1.26e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.15e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.25e-08,  2.17e-08,
  7.59e-09,  7.98e-09,  8.07e-09,  8.16e-09,  8.24e-09,  8.33e-09,  8.44e-09,  8.92e-09,  9.01e-09,  9.10e-09,  9.19e-09,  9.28e-09,  9.38e-09,  9.87e-09,  9.96e-09,  1.00e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.08e-08,  1.09e-08,  1.10e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.18e-08,  1.18e-08,  1.19e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.27e-08,  1.28e-08,  1.29e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.39e-08,  1.40e-08,  1.41e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.49e-08,  1.49e-08,  1.50e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.59e-08,  1.60e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.03e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.12e-08,  2.13e-08,  2.13e-08,  2.13e-08,  2.13e-08,  2.13e-08,  2.18e-08,  2.23e-08,
  8.03e-09,  8.03e-09,  8.12e-09,  8.21e-09,  8.30e-09,  8.38e-09,  8.49e-09,  9.06e-09,  9.15e-09,  9.24e-09,  9.33e-09,  9.41e-09,  9.52e-09,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.25e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.29e-08,  2.35e-08,  2.24e-08,
  8.03e-09,  8.03e-09,  8.12e-09,  8.21e-09,  8.30e-09,  8.38e-09,  8.49e-09,  9.06e-09,  9.15e-09,  9.24e-09,  9.33e-09,  9.41e-09,  9.52e-09,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.25e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.29e-08,  2.35e-08,  2.25e-08,
  8.03e-09,  8.03e-09,  8.12e-09,  8.21e-09,  8.30e-09,  8.38e-09,  8.49e-09,  9.06e-09,  9.15e-09,  9.24e-09,  9.33e-09,  9.41e-09,  9.52e-09,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.25e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.29e-08,  2.35e-08,  2.26e-08,
  8.03e-09,  8.03e-09,  8.12e-09,  8.21e-09,  8.30e-09,  8.38e-09,  8.49e-09,  9.06e-09,  9.15e-09,  9.24e-09,  9.33e-09,  9.41e-09,  9.52e-09,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.25e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.29e-08,  2.35e-08,  2.46e-08,
  8.03e-09,  8.03e-09,  8.12e-09,  8.21e-09,  8.30e-09,  8.38e-09,  8.49e-09,  9.06e-09,  9.15e-09,  9.24e-09,  9.33e-09,  9.41e-09,  9.52e-09,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.25e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.29e-08,  2.35e-08,  2.46e-08,
  8.59e-09,  8.98e-09,  9.06e-09,  9.15e-09,  9.24e-09,  9.33e-09,  9.43e-09,  9.92e-09,  1.00e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.09e-08,  1.10e-08,  1.10e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.18e-08,  1.19e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.28e-08,  1.28e-08,  1.29e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.37e-08,  1.38e-08,  1.39e-08,  1.40e-08,  1.40e-08,  1.42e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.49e-08,  1.50e-08,  1.51e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.58e-08,  1.59e-08,  1.60e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.70e-08,  1.75e-08,  1.76e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.94e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.12e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.22e-08,  2.23e-08,  2.23e-08,  2.23e-08,  2.23e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  9.03e-09,  9.03e-09,  9.11e-09,  9.20e-09,  9.29e-09,  9.38e-09,  9.48e-09,  1.01e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.14e-08,  1.15e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.17e-08,  2.18e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.29e-08,  2.35e-08,  2.35e-08,  2.36e-08,  2.37e-08,  2.38e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  9.03e-09,  9.03e-09,  9.11e-09,  9.20e-09,  9.29e-09,  9.38e-09,  9.48e-09,  1.01e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.14e-08,  1.15e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.17e-08,  2.18e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.29e-08,  2.35e-08,  2.35e-08,  2.36e-08,  2.37e-08,  2.38e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  9.03e-09,  9.03e-09,  9.11e-09,  9.20e-09,  9.29e-09,  9.38e-09,  9.48e-09,  1.01e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.14e-08,  1.15e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.17e-08,  2.18e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.29e-08,  2.35e-08,  2.35e-08,  2.36e-08,  2.37e-08,  2.38e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  9.03e-09,  9.03e-09,  9.11e-09,  9.20e-09,  9.29e-09,  9.38e-09,  9.48e-09,  1.01e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.14e-08,  1.15e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.17e-08,  2.18e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.29e-08,  2.35e-08,  2.35e-08,  2.36e-08,  2.37e-08,  2.38e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  9.03e-09,  9.03e-09,  9.11e-09,  9.20e-09,  9.29e-09,  9.38e-09,  9.48e-09,  1.01e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.14e-08,  1.15e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.17e-08,  2.18e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.29e-08,  2.35e-08,  2.35e-08,  2.36e-08,  2.37e-08,  2.38e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  9.58e-09,  9.97e-09,  1.01e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.09e-08,  1.10e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.19e-08,  1.19e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.28e-08,  1.29e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.37e-08,  1.38e-08,  1.39e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.47e-08,  1.48e-08,  1.49e-08,  1.50e-08,  1.50e-08,  1.51e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.59e-08,  1.60e-08,  1.61e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.68e-08,  1.69e-08,  1.70e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.80e-08,  1.85e-08,  1.86e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.04e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.22e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.00e-08,  1.00e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.52e-08,  1.53e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.98e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.34e-08,  2.35e-08,  2.36e-08,  2.37e-08,  2.38e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.00e-08,  1.00e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.52e-08,  1.53e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.98e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.34e-08,  2.35e-08,  2.36e-08,  2.37e-08,  2.38e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.00e-08,  1.00e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.52e-08,  1.53e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.98e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.34e-08,  2.35e-08,  2.36e-08,  2.37e-08,  2.38e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.00e-08,  1.00e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.52e-08,  1.53e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.98e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.34e-08,  2.35e-08,  2.36e-08,  2.37e-08,  2.38e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.00e-08,  1.00e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.52e-08,  1.53e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.98e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.34e-08,  2.35e-08,  2.36e-08,  2.37e-08,  2.38e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.06e-08,  1.10e-08,  1.11e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.19e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.29e-08,  1.29e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.38e-08,  1.39e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.47e-08,  1.48e-08,  1.49e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.57e-08,  1.58e-08,  1.59e-08,  1.60e-08,  1.60e-08,  1.61e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.70e-08,  1.71e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.78e-08,  1.79e-08,  1.80e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.90e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.14e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.10e-08,  1.10e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.62e-08,  1.63e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.93e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.24e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.34e-08,  2.35e-08,  2.36e-08,  2.36e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.10e-08,  1.10e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.62e-08,  1.63e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.93e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.24e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.34e-08,  2.35e-08,  2.36e-08,  2.36e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.10e-08,  1.10e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.62e-08,  1.63e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.93e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.24e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.34e-08,  2.35e-08,  2.36e-08,  2.36e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.10e-08,  1.10e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.62e-08,  1.63e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.93e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.24e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.34e-08,  2.35e-08,  2.36e-08,  2.36e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.10e-08,  1.10e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.62e-08,  1.63e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.93e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.24e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.34e-08,  2.35e-08,  2.36e-08,  2.36e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.16e-08,  1.20e-08,  1.21e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.29e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.39e-08,  1.39e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.48e-08,  1.49e-08,  1.50e-08,  1.51e-08,  1.51e-08,  1.53e-08,  1.57e-08,  1.58e-08,  1.59e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.69e-08,  1.70e-08,  1.71e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.80e-08,  1.81e-08,  1.86e-08,  1.87e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.90e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.00e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.20e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.87e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  2.03e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.33e-08,  2.34e-08,  2.35e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.20e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.87e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  2.03e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.33e-08,  2.34e-08,  2.35e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.20e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.87e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  2.03e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.33e-08,  2.34e-08,  2.35e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.20e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.87e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  2.03e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.33e-08,  2.34e-08,  2.35e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.20e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.87e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  2.03e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.28e-08,  2.33e-08,  2.34e-08,  2.35e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.26e-08,  1.30e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.39e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.48e-08,  1.49e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.58e-08,  1.59e-08,  1.60e-08,  1.61e-08,  1.61e-08,  1.62e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.79e-08,  1.80e-08,  1.81e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.90e-08,  1.91e-08,  1.96e-08,  1.97e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.00e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.10e-08,  2.15e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.30e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.51e-08,  1.52e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.82e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.33e-08,  2.34e-08,  2.35e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.30e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.51e-08,  1.52e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.82e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.33e-08,  2.34e-08,  2.35e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.30e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.51e-08,  1.52e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.82e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.33e-08,  2.34e-08,  2.35e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.30e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.51e-08,  1.52e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.82e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.33e-08,  2.34e-08,  2.35e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.30e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.51e-08,  1.52e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.82e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.33e-08,  2.34e-08,  2.35e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.36e-08,  1.40e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.49e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.58e-08,  1.59e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.68e-08,  1.69e-08,  1.70e-08,  1.71e-08,  1.71e-08,  1.72e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.80e-08,  1.81e-08,  1.82e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.89e-08,  1.90e-08,  1.91e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.00e-08,  2.01e-08,  2.06e-08,  2.07e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.10e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.20e-08,  2.25e-08,  2.25e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.40e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.61e-08,  1.62e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.76e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.92e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.12e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.22e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.33e-08,  2.34e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.40e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.61e-08,  1.62e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.76e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.92e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.12e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.22e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.33e-08,  2.34e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.40e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.61e-08,  1.62e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.76e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.92e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.12e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.22e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.33e-08,  2.34e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.40e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.61e-08,  1.62e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.76e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.92e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.12e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.22e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.33e-08,  2.34e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.40e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.61e-08,  1.62e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.76e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.92e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.12e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.22e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.33e-08,  2.34e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.46e-08,  1.50e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.59e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.62e-08,  1.64e-08,  1.68e-08,  1.69e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.78e-08,  1.79e-08,  1.80e-08,  1.80e-08,  1.81e-08,  1.82e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.90e-08,  1.91e-08,  1.92e-08,  1.97e-08,  1.98e-08,  1.98e-08,  1.99e-08,  2.00e-08,  2.01e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.10e-08,  2.11e-08,  2.16e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.20e-08,  2.25e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.50e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.71e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.12e-08,  2.13e-08,  2.14e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.22e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.32e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.50e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.71e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.12e-08,  2.13e-08,  2.14e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.22e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.32e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.50e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.71e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.12e-08,  2.13e-08,  2.14e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.22e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.32e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.50e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.71e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.12e-08,  2.13e-08,  2.14e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.22e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.32e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.50e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.71e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.12e-08,  2.13e-08,  2.14e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.22e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.27e-08,  2.32e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.56e-08,  1.59e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.69e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.72e-08,  1.73e-08,  1.78e-08,  1.79e-08,  1.80e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.88e-08,  1.89e-08,  1.90e-08,  1.90e-08,  1.91e-08,  1.92e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.00e-08,  2.01e-08,  2.02e-08,  2.07e-08,  2.08e-08,  2.08e-08,  2.09e-08,  2.10e-08,  2.11e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.20e-08,  2.21e-08,  2.26e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.60e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.81e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.94e-08,  1.95e-08,  2.01e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.11e-08,  2.12e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.22e-08,  2.23e-08,  2.24e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.32e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.60e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.81e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.94e-08,  1.95e-08,  2.01e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.11e-08,  2.12e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.22e-08,  2.23e-08,  2.24e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.32e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.60e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.81e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.94e-08,  1.95e-08,  2.01e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.11e-08,  2.12e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.22e-08,  2.23e-08,  2.24e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.32e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.60e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.81e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.94e-08,  1.95e-08,  2.01e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.11e-08,  2.12e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.22e-08,  2.23e-08,  2.24e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.32e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.60e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.81e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.94e-08,  1.95e-08,  2.01e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.11e-08,  2.12e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.22e-08,  2.23e-08,  2.24e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.32e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.66e-08,  1.69e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.79e-08,  1.80e-08,  1.81e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.88e-08,  1.89e-08,  1.90e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.98e-08,  1.99e-08,  1.99e-08,  2.00e-08,  2.01e-08,  2.02e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.10e-08,  2.11e-08,  2.12e-08,  2.17e-08,  2.18e-08,  2.18e-08,  2.19e-08,  2.20e-08,  2.21e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.70e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.74e-08,  1.80e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  2.01e-08,  2.02e-08,  2.03e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.11e-08,  2.12e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.21e-08,  2.22e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.70e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.74e-08,  1.80e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  2.01e-08,  2.02e-08,  2.03e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.11e-08,  2.12e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.21e-08,  2.22e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.70e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.74e-08,  1.80e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  2.01e-08,  2.02e-08,  2.03e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.11e-08,  2.12e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.21e-08,  2.22e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.70e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.74e-08,  1.80e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  2.01e-08,  2.02e-08,  2.03e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.11e-08,  2.12e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.21e-08,  2.22e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.70e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.74e-08,  1.80e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  2.01e-08,  2.02e-08,  2.03e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.11e-08,  2.12e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.21e-08,  2.22e-08,  2.23e-08,  2.24e-08,  2.25e-08,  2.26e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.76e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.80e-08,  1.81e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.88e-08,  1.89e-08,  1.90e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.00e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,
  1.76e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.80e-08,  1.81e-08,  1.87e-08,  1.88e-08,  1.88e-08,  1.89e-08,  1.90e-08,  1.91e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.00e-08,  2.01e-08,  2.02e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.10e-08,  2.11e-08,  2.12e-08,  2.18e-08,  2.18e-08,  2.19e-08,  2.20e-08,  2.21e-08,  2.22e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08, -1.00e+00

  };



  // delta_clb_to_outpad
  double delta_clb_to_outpad_hard[93][93] = {
 -1.00e+00,  2.07e-09,  3.35e-09,  3.33e-09,  4.02e-09,  4.01e-09,  4.70e-09,  3.58e-09,  5.37e-09,  5.35e-09,  6.04e-09,  6.03e-09,  6.72e-09,  4.52e-09,  7.39e-09,  7.37e-09,  8.06e-09,  8.05e-09,  8.74e-09,  5.46e-09,  9.41e-09,  9.39e-09,  1.01e-08,  1.01e-08,  1.08e-08,  6.41e-09,  1.14e-08,  1.14e-08,  1.21e-08,  1.21e-08,  1.28e-08,  7.35e-09,  1.35e-08,  1.34e-08,  1.41e-08,  1.41e-08,  1.48e-08,  8.30e-09,  1.55e-08,  1.55e-08,  1.61e-08,  1.61e-08,  1.68e-08,  9.24e-09,  1.75e-08,  1.75e-08,  1.82e-08,  1.81e-08,  1.88e-08,  1.02e-08,  1.95e-08,  1.95e-08,  2.02e-08,  2.02e-08,  2.09e-08,  1.11e-08,  2.15e-08,  2.15e-08,  2.22e-08,  2.22e-08,  2.29e-08,  1.21e-08,  2.36e-08,  2.35e-08,  2.42e-08,  2.42e-08,  2.49e-08,  1.30e-08,  2.56e-08,  2.56e-08,  2.62e-08,  2.62e-08,  2.69e-08,  1.40e-08,  2.76e-08,  2.76e-08,  2.83e-08,  2.83e-08,  2.89e-08,  1.49e-08,  2.96e-08,  2.96e-08,  3.03e-08,  3.56e-08,  3.56e-08,  1.58e-08,  3.56e-08,  3.56e-08,  3.56e-08,  3.56e-08,  3.56e-08,  1.68e-08,  3.56e-08,
  2.67e-09,  2.67e-09,  2.86e-09,  3.23e-09,  3.32e-09,  3.40e-09,  3.51e-09,  4.08e-09,  4.17e-09,  4.26e-09,  4.35e-09,  4.44e-09,  4.54e-09,  5.11e-09,  5.20e-09,  5.29e-09,  5.38e-09,  5.47e-09,  5.57e-09,  6.14e-09,  6.23e-09,  6.32e-09,  6.41e-09,  6.50e-09,  6.60e-09,  7.17e-09,  7.26e-09,  7.35e-09,  7.44e-09,  7.53e-09,  7.63e-09,  8.20e-09,  8.29e-09,  8.38e-09,  8.47e-09,  8.56e-09,  8.66e-09,  9.23e-09,  9.32e-09,  9.41e-09,  9.50e-09,  9.59e-09,  9.69e-09,  1.03e-08,  1.04e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.16e-08,  1.18e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.28e-08,  1.34e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.59e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.85e-08,  3.56e-08,
  2.88e-09,  2.76e-09,  2.86e-09,  3.23e-09,  3.32e-09,  3.40e-09,  3.51e-09,  4.08e-09,  4.17e-09,  4.26e-09,  4.35e-09,  4.44e-09,  4.54e-09,  5.11e-09,  5.20e-09,  5.29e-09,  5.38e-09,  5.47e-09,  5.57e-09,  6.14e-09,  6.23e-09,  6.32e-09,  6.41e-09,  6.50e-09,  6.60e-09,  7.17e-09,  7.26e-09,  7.35e-09,  7.44e-09,  7.53e-09,  7.63e-09,  8.20e-09,  8.29e-09,  8.38e-09,  8.47e-09,  8.56e-09,  8.66e-09,  9.23e-09,  9.32e-09,  9.41e-09,  9.50e-09,  9.59e-09,  9.69e-09,  1.03e-08,  1.04e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.16e-08,  1.18e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.28e-08,  1.34e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.59e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.85e-08,  3.56e-08,
  3.38e-09,  3.45e-09,  3.14e-09,  3.23e-09,  3.32e-09,  3.40e-09,  3.51e-09,  4.08e-09,  4.17e-09,  4.26e-09,  4.35e-09,  4.44e-09,  4.54e-09,  5.11e-09,  5.20e-09,  5.29e-09,  5.38e-09,  5.47e-09,  5.57e-09,  6.14e-09,  6.23e-09,  6.32e-09,  6.41e-09,  6.50e-09,  6.60e-09,  7.17e-09,  7.26e-09,  7.35e-09,  7.44e-09,  7.53e-09,  7.63e-09,  8.20e-09,  8.29e-09,  8.38e-09,  8.47e-09,  8.56e-09,  8.66e-09,  9.23e-09,  9.32e-09,  9.41e-09,  9.50e-09,  9.59e-09,  9.69e-09,  1.03e-08,  1.04e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.16e-08,  1.18e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.28e-08,  1.34e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.59e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.85e-08,  3.56e-08,
  3.59e-09,  3.05e-09,  3.14e-09,  3.23e-09,  3.32e-09,  3.40e-09,  3.51e-09,  4.08e-09,  4.17e-09,  4.26e-09,  4.35e-09,  4.44e-09,  4.54e-09,  5.11e-09,  5.20e-09,  5.29e-09,  5.38e-09,  5.47e-09,  5.57e-09,  6.14e-09,  6.23e-09,  6.32e-09,  6.41e-09,  6.50e-09,  6.60e-09,  7.17e-09,  7.26e-09,  7.35e-09,  7.44e-09,  7.53e-09,  7.63e-09,  8.20e-09,  8.29e-09,  8.38e-09,  8.47e-09,  8.56e-09,  8.66e-09,  9.23e-09,  9.32e-09,  9.41e-09,  9.50e-09,  9.59e-09,  9.69e-09,  1.03e-08,  1.04e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.16e-08,  1.18e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.28e-08,  1.34e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.59e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.85e-08,  3.56e-08,
  3.59e-09,  3.05e-09,  3.14e-09,  3.23e-09,  3.32e-09,  3.40e-09,  3.51e-09,  4.08e-09,  4.17e-09,  4.26e-09,  4.35e-09,  4.44e-09,  4.54e-09,  5.11e-09,  5.20e-09,  5.29e-09,  5.38e-09,  5.47e-09,  5.57e-09,  6.14e-09,  6.23e-09,  6.32e-09,  6.41e-09,  6.50e-09,  6.60e-09,  7.17e-09,  7.26e-09,  7.35e-09,  7.44e-09,  7.53e-09,  7.63e-09,  8.20e-09,  8.29e-09,  8.38e-09,  8.47e-09,  8.56e-09,  8.66e-09,  9.23e-09,  9.32e-09,  9.41e-09,  9.50e-09,  9.59e-09,  9.69e-09,  1.03e-08,  1.04e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.16e-08,  1.18e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.28e-08,  1.34e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.59e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.85e-08,  3.56e-08,
  3.59e-09,  3.05e-09,  3.14e-09,  3.23e-09,  3.32e-09,  3.40e-09,  3.51e-09,  4.08e-09,  4.17e-09,  4.26e-09,  4.35e-09,  4.44e-09,  4.54e-09,  5.11e-09,  5.20e-09,  5.29e-09,  5.38e-09,  5.47e-09,  5.57e-09,  6.14e-09,  6.23e-09,  6.32e-09,  6.41e-09,  6.50e-09,  6.60e-09,  7.17e-09,  7.26e-09,  7.35e-09,  7.44e-09,  7.53e-09,  7.63e-09,  8.20e-09,  8.29e-09,  8.38e-09,  8.47e-09,  8.56e-09,  8.66e-09,  9.23e-09,  9.32e-09,  9.41e-09,  9.50e-09,  9.59e-09,  9.69e-09,  1.03e-08,  1.04e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.16e-08,  1.18e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.28e-08,  1.34e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.59e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.85e-08,  3.56e-08,
  4.68e-09,  4.05e-09,  4.14e-09,  4.22e-09,  4.31e-09,  4.40e-09,  4.50e-09,  5.08e-09,  5.17e-09,  5.25e-09,  5.34e-09,  5.43e-09,  5.53e-09,  6.11e-09,  6.20e-09,  6.28e-09,  6.37e-09,  6.46e-09,  6.56e-09,  7.14e-09,  7.23e-09,  7.31e-09,  7.40e-09,  7.49e-09,  7.60e-09,  8.17e-09,  8.26e-09,  8.34e-09,  8.43e-09,  8.52e-09,  8.63e-09,  9.20e-09,  9.29e-09,  9.37e-09,  9.46e-09,  9.55e-09,  9.66e-09,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.13e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.44e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.85e-08,  1.86e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.95e-08,  3.56e-08,
  4.68e-09,  4.05e-09,  4.14e-09,  4.22e-09,  4.31e-09,  4.40e-09,  4.50e-09,  5.08e-09,  5.17e-09,  5.25e-09,  5.34e-09,  5.43e-09,  5.53e-09,  6.11e-09,  6.20e-09,  6.28e-09,  6.37e-09,  6.46e-09,  6.56e-09,  7.14e-09,  7.23e-09,  7.31e-09,  7.40e-09,  7.49e-09,  7.60e-09,  8.17e-09,  8.26e-09,  8.34e-09,  8.43e-09,  8.52e-09,  8.63e-09,  9.20e-09,  9.29e-09,  9.37e-09,  9.46e-09,  9.55e-09,  9.66e-09,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.13e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.44e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.85e-08,  1.86e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.95e-08,  3.56e-08,
  4.68e-09,  4.05e-09,  4.14e-09,  4.22e-09,  4.31e-09,  4.40e-09,  4.50e-09,  5.08e-09,  5.17e-09,  5.25e-09,  5.34e-09,  5.43e-09,  5.53e-09,  6.11e-09,  6.20e-09,  6.28e-09,  6.37e-09,  6.46e-09,  6.56e-09,  7.14e-09,  7.23e-09,  7.31e-09,  7.40e-09,  7.49e-09,  7.60e-09,  8.17e-09,  8.26e-09,  8.34e-09,  8.43e-09,  8.52e-09,  8.63e-09,  9.20e-09,  9.29e-09,  9.37e-09,  9.46e-09,  9.55e-09,  9.66e-09,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.13e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.44e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.85e-08,  1.86e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.95e-08,  3.56e-08,
  4.68e-09,  4.05e-09,  4.14e-09,  4.22e-09,  4.31e-09,  4.40e-09,  4.50e-09,  5.08e-09,  5.17e-09,  5.25e-09,  5.34e-09,  5.43e-09,  5.53e-09,  6.11e-09,  6.20e-09,  6.28e-09,  6.37e-09,  6.46e-09,  6.56e-09,  7.14e-09,  7.23e-09,  7.31e-09,  7.40e-09,  7.49e-09,  7.60e-09,  8.17e-09,  8.26e-09,  8.34e-09,  8.43e-09,  8.52e-09,  8.63e-09,  9.20e-09,  9.29e-09,  9.37e-09,  9.46e-09,  9.55e-09,  9.66e-09,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.13e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.44e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.85e-08,  1.86e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.95e-08,  3.56e-08,
  4.68e-09,  4.05e-09,  4.14e-09,  4.22e-09,  4.31e-09,  4.40e-09,  4.50e-09,  5.08e-09,  5.17e-09,  5.25e-09,  5.34e-09,  5.43e-09,  5.53e-09,  6.11e-09,  6.20e-09,  6.28e-09,  6.37e-09,  6.46e-09,  6.56e-09,  7.14e-09,  7.23e-09,  7.31e-09,  7.40e-09,  7.49e-09,  7.60e-09,  8.17e-09,  8.26e-09,  8.34e-09,  8.43e-09,  8.52e-09,  8.63e-09,  9.20e-09,  9.29e-09,  9.37e-09,  9.46e-09,  9.55e-09,  9.66e-09,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.13e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.44e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.85e-08,  1.86e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.95e-08,  3.56e-08,
  4.68e-09,  4.05e-09,  4.14e-09,  4.22e-09,  4.31e-09,  4.40e-09,  4.50e-09,  5.08e-09,  5.17e-09,  5.25e-09,  5.34e-09,  5.43e-09,  5.53e-09,  6.11e-09,  6.20e-09,  6.28e-09,  6.37e-09,  6.46e-09,  6.56e-09,  7.14e-09,  7.23e-09,  7.31e-09,  7.40e-09,  7.49e-09,  7.60e-09,  8.17e-09,  8.26e-09,  8.34e-09,  8.43e-09,  8.52e-09,  8.63e-09,  9.20e-09,  9.29e-09,  9.37e-09,  9.46e-09,  9.55e-09,  9.66e-09,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.13e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.38e-08,  1.44e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.69e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.85e-08,  1.86e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.95e-08,  3.56e-08,
  5.76e-09,  5.04e-09,  5.13e-09,  5.22e-09,  5.31e-09,  5.40e-09,  5.50e-09,  6.07e-09,  6.16e-09,  6.25e-09,  6.34e-09,  6.43e-09,  6.53e-09,  7.10e-09,  7.19e-09,  7.28e-09,  7.37e-09,  7.46e-09,  7.56e-09,  8.13e-09,  8.22e-09,  8.31e-09,  8.40e-09,  8.49e-09,  8.59e-09,  9.17e-09,  9.25e-09,  9.34e-09,  9.43e-09,  9.52e-09,  9.62e-09,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.05e-08,  1.07e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.23e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.05e-08,  3.56e-08,
  5.76e-09,  5.04e-09,  5.13e-09,  5.22e-09,  5.31e-09,  5.40e-09,  5.50e-09,  6.07e-09,  6.16e-09,  6.25e-09,  6.34e-09,  6.43e-09,  6.53e-09,  7.10e-09,  7.19e-09,  7.28e-09,  7.37e-09,  7.46e-09,  7.56e-09,  8.13e-09,  8.22e-09,  8.31e-09,  8.40e-09,  8.49e-09,  8.59e-09,  9.17e-09,  9.25e-09,  9.34e-09,  9.43e-09,  9.52e-09,  9.62e-09,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.05e-08,  1.07e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.23e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.05e-08,  3.56e-08,
  5.76e-09,  5.04e-09,  5.13e-09,  5.22e-09,  5.31e-09,  5.40e-09,  5.50e-09,  6.07e-09,  6.16e-09,  6.25e-09,  6.34e-09,  6.43e-09,  6.53e-09,  7.10e-09,  7.19e-09,  7.28e-09,  7.37e-09,  7.46e-09,  7.56e-09,  8.13e-09,  8.22e-09,  8.31e-09,  8.40e-09,  8.49e-09,  8.59e-09,  9.17e-09,  9.25e-09,  9.34e-09,  9.43e-09,  9.52e-09,  9.62e-09,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.05e-08,  1.07e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.23e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.05e-08,  3.56e-08,
  5.76e-09,  5.04e-09,  5.13e-09,  5.22e-09,  5.31e-09,  5.40e-09,  5.50e-09,  6.07e-09,  6.16e-09,  6.25e-09,  6.34e-09,  6.43e-09,  6.53e-09,  7.10e-09,  7.19e-09,  7.28e-09,  7.37e-09,  7.46e-09,  7.56e-09,  8.13e-09,  8.22e-09,  8.31e-09,  8.40e-09,  8.49e-09,  8.59e-09,  9.17e-09,  9.25e-09,  9.34e-09,  9.43e-09,  9.52e-09,  9.62e-09,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.05e-08,  1.07e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.23e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.05e-08,  3.56e-08,
  5.76e-09,  5.04e-09,  5.13e-09,  5.22e-09,  5.31e-09,  5.40e-09,  5.50e-09,  6.07e-09,  6.16e-09,  6.25e-09,  6.34e-09,  6.43e-09,  6.53e-09,  7.10e-09,  7.19e-09,  7.28e-09,  7.37e-09,  7.46e-09,  7.56e-09,  8.13e-09,  8.22e-09,  8.31e-09,  8.40e-09,  8.49e-09,  8.59e-09,  9.17e-09,  9.25e-09,  9.34e-09,  9.43e-09,  9.52e-09,  9.62e-09,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.05e-08,  1.07e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.23e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.05e-08,  3.56e-08,
  5.76e-09,  5.04e-09,  5.13e-09,  5.22e-09,  5.31e-09,  5.40e-09,  5.50e-09,  6.07e-09,  6.16e-09,  6.25e-09,  6.34e-09,  6.43e-09,  6.53e-09,  7.10e-09,  7.19e-09,  7.28e-09,  7.37e-09,  7.46e-09,  7.56e-09,  8.13e-09,  8.22e-09,  8.31e-09,  8.40e-09,  8.49e-09,  8.59e-09,  9.17e-09,  9.25e-09,  9.34e-09,  9.43e-09,  9.52e-09,  9.62e-09,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.05e-08,  1.07e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.23e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.48e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.05e-08,  3.56e-08,
  6.84e-09,  6.04e-09,  6.13e-09,  6.22e-09,  6.30e-09,  6.39e-09,  6.50e-09,  7.07e-09,  7.16e-09,  7.25e-09,  7.33e-09,  7.42e-09,  7.53e-09,  8.10e-09,  8.19e-09,  8.28e-09,  8.36e-09,  8.45e-09,  8.56e-09,  9.13e-09,  9.22e-09,  9.31e-09,  9.39e-09,  9.48e-09,  9.59e-09,  1.02e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.74e-08,  1.75e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.15e-08,  3.56e-08,
  6.84e-09,  6.04e-09,  6.13e-09,  6.22e-09,  6.30e-09,  6.39e-09,  6.50e-09,  7.07e-09,  7.16e-09,  7.25e-09,  7.33e-09,  7.42e-09,  7.53e-09,  8.10e-09,  8.19e-09,  8.28e-09,  8.36e-09,  8.45e-09,  8.56e-09,  9.13e-09,  9.22e-09,  9.31e-09,  9.39e-09,  9.48e-09,  9.59e-09,  1.02e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.74e-08,  1.75e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.15e-08,  3.56e-08,
  6.84e-09,  6.04e-09,  6.13e-09,  6.22e-09,  6.30e-09,  6.39e-09,  6.50e-09,  7.07e-09,  7.16e-09,  7.25e-09,  7.33e-09,  7.42e-09,  7.53e-09,  8.10e-09,  8.19e-09,  8.28e-09,  8.36e-09,  8.45e-09,  8.56e-09,  9.13e-09,  9.22e-09,  9.31e-09,  9.39e-09,  9.48e-09,  9.59e-09,  1.02e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.74e-08,  1.75e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.15e-08,  3.56e-08,
  6.84e-09,  6.04e-09,  6.13e-09,  6.22e-09,  6.30e-09,  6.39e-09,  6.50e-09,  7.07e-09,  7.16e-09,  7.25e-09,  7.33e-09,  7.42e-09,  7.53e-09,  8.10e-09,  8.19e-09,  8.28e-09,  8.36e-09,  8.45e-09,  8.56e-09,  9.13e-09,  9.22e-09,  9.31e-09,  9.39e-09,  9.48e-09,  9.59e-09,  1.02e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.74e-08,  1.75e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.15e-08,  3.56e-08,
  6.84e-09,  6.04e-09,  6.13e-09,  6.22e-09,  6.30e-09,  6.39e-09,  6.50e-09,  7.07e-09,  7.16e-09,  7.25e-09,  7.33e-09,  7.42e-09,  7.53e-09,  8.10e-09,  8.19e-09,  8.28e-09,  8.36e-09,  8.45e-09,  8.56e-09,  9.13e-09,  9.22e-09,  9.31e-09,  9.39e-09,  9.48e-09,  9.59e-09,  1.02e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.74e-08,  1.75e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.15e-08,  3.56e-08,
  6.84e-09,  6.04e-09,  6.13e-09,  6.22e-09,  6.30e-09,  6.39e-09,  6.50e-09,  7.07e-09,  7.16e-09,  7.25e-09,  7.33e-09,  7.42e-09,  7.53e-09,  8.10e-09,  8.19e-09,  8.28e-09,  8.36e-09,  8.45e-09,  8.56e-09,  9.13e-09,  9.22e-09,  9.31e-09,  9.39e-09,  9.48e-09,  9.59e-09,  1.02e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.27e-08,  1.33e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.58e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.74e-08,  1.75e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.89e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.15e-08,  3.56e-08,
  7.92e-09,  7.04e-09,  7.12e-09,  7.21e-09,  7.30e-09,  7.39e-09,  7.49e-09,  8.07e-09,  8.15e-09,  8.24e-09,  8.33e-09,  8.42e-09,  8.52e-09,  9.10e-09,  9.18e-09,  9.27e-09,  9.36e-09,  9.45e-09,  9.55e-09,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.12e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.25e-08,  1.26e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.15e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.46e-08,  3.56e-08,
  7.92e-09,  7.04e-09,  7.12e-09,  7.21e-09,  7.30e-09,  7.39e-09,  7.49e-09,  8.07e-09,  8.15e-09,  8.24e-09,  8.33e-09,  8.42e-09,  8.52e-09,  9.10e-09,  9.18e-09,  9.27e-09,  9.36e-09,  9.45e-09,  9.55e-09,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.12e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.25e-08,  1.26e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.15e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.46e-08,  3.56e-08,
  7.92e-09,  7.04e-09,  7.12e-09,  7.21e-09,  7.30e-09,  7.39e-09,  7.49e-09,  8.07e-09,  8.15e-09,  8.24e-09,  8.33e-09,  8.42e-09,  8.52e-09,  9.10e-09,  9.18e-09,  9.27e-09,  9.36e-09,  9.45e-09,  9.55e-09,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.12e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.25e-08,  1.26e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.15e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.46e-08,  3.56e-08,
  7.92e-09,  7.04e-09,  7.12e-09,  7.21e-09,  7.30e-09,  7.39e-09,  7.49e-09,  8.07e-09,  8.15e-09,  8.24e-09,  8.33e-09,  8.42e-09,  8.52e-09,  9.10e-09,  9.18e-09,  9.27e-09,  9.36e-09,  9.45e-09,  9.55e-09,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.12e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.25e-08,  1.26e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.15e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.46e-08,  3.56e-08,
  7.92e-09,  7.04e-09,  7.12e-09,  7.21e-09,  7.30e-09,  7.39e-09,  7.49e-09,  8.07e-09,  8.15e-09,  8.24e-09,  8.33e-09,  8.42e-09,  8.52e-09,  9.10e-09,  9.18e-09,  9.27e-09,  9.36e-09,  9.45e-09,  9.55e-09,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.12e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.25e-08,  1.26e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.15e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.46e-08,  3.56e-08,
  7.92e-09,  7.04e-09,  7.12e-09,  7.21e-09,  7.30e-09,  7.39e-09,  7.49e-09,  8.07e-09,  8.15e-09,  8.24e-09,  8.33e-09,  8.42e-09,  8.52e-09,  9.10e-09,  9.18e-09,  9.27e-09,  9.36e-09,  9.45e-09,  9.55e-09,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.12e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.25e-08,  1.26e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.37e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.68e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.15e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.46e-08,  3.56e-08,
  9.00e-09,  8.03e-09,  8.12e-09,  8.21e-09,  8.30e-09,  8.38e-09,  8.49e-09,  9.06e-09,  9.15e-09,  9.24e-09,  9.33e-09,  9.41e-09,  9.52e-09,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  9.00e-09,  8.03e-09,  8.12e-09,  8.21e-09,  8.30e-09,  8.38e-09,  8.49e-09,  9.06e-09,  9.15e-09,  9.24e-09,  9.33e-09,  9.41e-09,  9.52e-09,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  9.00e-09,  8.03e-09,  8.12e-09,  8.21e-09,  8.30e-09,  8.38e-09,  8.49e-09,  9.06e-09,  9.15e-09,  9.24e-09,  9.33e-09,  9.41e-09,  9.52e-09,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  9.00e-09,  8.03e-09,  8.12e-09,  8.21e-09,  8.30e-09,  8.38e-09,  8.49e-09,  9.06e-09,  9.15e-09,  9.24e-09,  9.33e-09,  9.41e-09,  9.52e-09,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  9.00e-09,  8.03e-09,  8.12e-09,  8.21e-09,  8.30e-09,  8.38e-09,  8.49e-09,  9.06e-09,  9.15e-09,  9.24e-09,  9.33e-09,  9.41e-09,  9.52e-09,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  9.00e-09,  8.03e-09,  8.12e-09,  8.21e-09,  8.30e-09,  8.38e-09,  8.49e-09,  9.06e-09,  9.15e-09,  9.24e-09,  9.33e-09,  9.41e-09,  9.52e-09,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.22e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.47e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.63e-08,  1.64e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.19e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.01e-08,  9.03e-09,  9.11e-09,  9.20e-09,  9.29e-09,  9.38e-09,  9.48e-09,  1.01e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.14e-08,  1.15e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.17e-08,  2.18e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.01e-08,  9.03e-09,  9.11e-09,  9.20e-09,  9.29e-09,  9.38e-09,  9.48e-09,  1.01e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.14e-08,  1.15e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.17e-08,  2.18e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.01e-08,  9.03e-09,  9.11e-09,  9.20e-09,  9.29e-09,  9.38e-09,  9.48e-09,  1.01e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.14e-08,  1.15e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.17e-08,  2.18e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.01e-08,  9.03e-09,  9.11e-09,  9.20e-09,  9.29e-09,  9.38e-09,  9.48e-09,  1.01e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.14e-08,  1.15e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.17e-08,  2.18e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.01e-08,  9.03e-09,  9.11e-09,  9.20e-09,  9.29e-09,  9.38e-09,  9.48e-09,  1.01e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.14e-08,  1.15e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.17e-08,  2.18e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.01e-08,  9.03e-09,  9.11e-09,  9.20e-09,  9.29e-09,  9.38e-09,  9.48e-09,  1.01e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.14e-08,  1.15e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.26e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.57e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.88e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  1.98e-08,  2.04e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.17e-08,  2.18e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.12e-08,  1.00e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.52e-08,  1.53e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.98e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.12e-08,  1.00e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.52e-08,  1.53e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.98e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.12e-08,  1.00e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.52e-08,  1.53e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.98e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.12e-08,  1.00e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.52e-08,  1.53e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.98e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.12e-08,  1.00e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.52e-08,  1.53e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.98e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.12e-08,  1.00e-08,  1.01e-08,  1.02e-08,  1.03e-08,  1.04e-08,  1.05e-08,  1.11e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.36e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.52e-08,  1.53e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.67e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.83e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.96e-08,  1.98e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.08e-08,  2.14e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.22e-08,  1.10e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.62e-08,  1.63e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.93e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.22e-08,  1.10e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.62e-08,  1.63e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.93e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.22e-08,  1.10e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.62e-08,  1.63e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.93e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.22e-08,  1.10e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.62e-08,  1.63e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.93e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.22e-08,  1.10e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.62e-08,  1.63e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.93e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.22e-08,  1.10e-08,  1.11e-08,  1.12e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.46e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.62e-08,  1.63e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.77e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.87e-08,  1.93e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.18e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.33e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.87e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  2.03e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.33e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.87e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  2.03e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.33e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.87e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  2.03e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.33e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.87e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  2.03e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.33e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.87e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  2.03e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.33e-08,  1.20e-08,  1.21e-08,  1.22e-08,  1.23e-08,  1.24e-08,  1.25e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.56e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.85e-08,  1.87e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  1.97e-08,  2.03e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.13e-08,  2.14e-08,  2.15e-08,  2.15e-08,  2.16e-08,  2.17e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.44e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.51e-08,  1.52e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.82e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.44e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.51e-08,  1.52e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.82e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.44e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.51e-08,  1.52e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.82e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.44e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.51e-08,  1.52e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.82e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.44e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.51e-08,  1.52e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.82e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.44e-08,  1.30e-08,  1.31e-08,  1.32e-08,  1.33e-08,  1.34e-08,  1.35e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.51e-08,  1.52e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.66e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.76e-08,  1.82e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.07e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.55e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.61e-08,  1.62e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.76e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.92e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.55e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.61e-08,  1.62e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.76e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.92e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.55e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.61e-08,  1.62e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.76e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.92e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.55e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.61e-08,  1.62e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.76e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.92e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.55e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.61e-08,  1.62e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.76e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.92e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.55e-08,  1.40e-08,  1.41e-08,  1.42e-08,  1.43e-08,  1.44e-08,  1.45e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.61e-08,  1.62e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.74e-08,  1.76e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.86e-08,  1.92e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.05e-08,  2.06e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.66e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.71e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.66e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.71e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.66e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.71e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.66e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.71e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.66e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.71e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.66e-08,  1.50e-08,  1.51e-08,  1.52e-08,  1.53e-08,  1.54e-08,  1.55e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.71e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  1.96e-08,  2.02e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.77e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.81e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.94e-08,  1.95e-08,  2.01e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.77e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.81e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.94e-08,  1.95e-08,  2.01e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.77e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.81e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.94e-08,  1.95e-08,  2.01e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.77e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.81e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.94e-08,  1.95e-08,  2.01e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.77e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.81e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.94e-08,  1.95e-08,  2.01e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.77e-08,  1.60e-08,  1.61e-08,  1.62e-08,  1.63e-08,  1.64e-08,  1.65e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.74e-08,  1.75e-08,  1.81e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.94e-08,  1.95e-08,  2.01e-08,  2.02e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.06e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.87e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.74e-08,  1.80e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  2.01e-08,  2.02e-08,  2.03e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.87e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.74e-08,  1.80e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  2.01e-08,  2.02e-08,  2.03e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.87e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.74e-08,  1.80e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  2.01e-08,  2.02e-08,  2.03e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.87e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.74e-08,  1.80e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  2.01e-08,  2.02e-08,  2.03e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.87e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.74e-08,  1.80e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  2.01e-08,  2.02e-08,  2.03e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.87e-08,  1.70e-08,  1.71e-08,  1.72e-08,  1.73e-08,  1.73e-08,  1.74e-08,  1.80e-08,  1.81e-08,  1.82e-08,  1.83e-08,  1.84e-08,  1.85e-08,  1.91e-08,  1.91e-08,  1.92e-08,  1.93e-08,  1.94e-08,  1.95e-08,  2.01e-08,  2.02e-08,  2.03e-08,  2.03e-08,  2.04e-08,  2.05e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.94e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.80e-08,  1.81e-08,  1.87e-08,  1.88e-08,  1.88e-08,  1.89e-08,  1.90e-08,  1.91e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.00e-08,  2.01e-08,  2.02e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.10e-08,  2.11e-08,  2.12e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  3.56e-08,
  1.94e-08,  1.76e-08,  1.77e-08,  1.78e-08,  1.79e-08,  1.80e-08,  1.81e-08,  1.87e-08,  1.88e-08,  1.88e-08,  1.89e-08,  1.90e-08,  1.91e-08,  1.97e-08,  1.98e-08,  1.99e-08,  2.00e-08,  2.01e-08,  2.02e-08,  2.07e-08,  2.08e-08,  2.09e-08,  2.10e-08,  2.11e-08,  2.12e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08,  2.46e-08, -1.00e+00
  };


  // delta_inpad_to_outpad
  double delta_inpad_to_outpad_hard[94][94] = {
  0.00e+00,  2.15e-09,  2.24e-09,  2.33e-09,  2.42e-09,  2.52e-09,  3.10e-09,  3.18e-09,  3.27e-09,  3.36e-09,  3.45e-09,  3.55e-09,  4.13e-09,  4.21e-09,  4.30e-09,  4.39e-09,  4.48e-09,  4.58e-09,  5.16e-09,  5.24e-09,  5.33e-09,  5.42e-09,  5.51e-09,  5.61e-09,  6.19e-09,  6.27e-09,  6.36e-09,  6.45e-09,  6.54e-09,  6.64e-09,  7.22e-09,  7.30e-09,  7.39e-09,  7.48e-09,  7.57e-09,  7.67e-09,  8.25e-09,  8.33e-09,  8.42e-09,  8.51e-09,  8.60e-09,  8.70e-09,  9.28e-09,  9.36e-09,  9.45e-09,  9.54e-09,  9.63e-09,  9.73e-09,  1.03e-08,  1.04e-08,  1.05e-08,  1.06e-08,  1.07e-08,  1.08e-08,  1.13e-08,  1.14e-08,  1.15e-08,  1.16e-08,  1.17e-08,  1.18e-08,  1.24e-08,  1.25e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08,  1.35e-08, -1.00e+00,  3.66e-08,
  2.17e-09,  2.66e-09,  2.86e-09,  3.35e-09,  3.54e-09,  3.54e-09,  3.54e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.86e-08,  1.86e-08,  3.66e-08,
  2.27e-09,  2.86e-09,  2.86e-09,  3.54e-09,  3.54e-09,  3.54e-09,  3.54e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.86e-08,  1.86e-08,  3.66e-08,
  2.36e-09,  3.54e-09,  3.54e-09,  3.54e-09,  3.54e-09,  3.54e-09,  3.54e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.86e-08,  1.86e-08,  3.66e-08,
  2.46e-09,  3.54e-09,  3.54e-09,  3.54e-09,  3.54e-09,  3.54e-09,  3.54e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.86e-08,  1.86e-08,  3.66e-08,
  2.57e-09,  3.54e-09,  3.54e-09,  3.54e-09,  3.54e-09,  3.54e-09,  3.54e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.86e-08,  1.86e-08,  3.66e-08,
  3.16e-09,  3.54e-09,  3.54e-09,  3.54e-09,  3.54e-09,  3.54e-09,  3.54e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.86e-08,  1.86e-08,  3.66e-08,
  3.25e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  5.48e-09,  5.48e-09,  5.48e-09,  5.48e-09,  5.48e-09,  5.48e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  6.43e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  7.37e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  8.32e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  9.26e-09,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.02e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.11e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.30e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.68e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.83e-08,  1.83e-08,  3.66e-08,
  3.35e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.96e-08,  1.96e-08,  3.66e-08,
  3.45e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.96e-08,  1.96e-08,  3.66e-08,
  3.54e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.96e-08,  1.96e-08,  3.66e-08,
  3.66e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.96e-08,  1.96e-08,  3.66e-08,
  4.24e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  4.57e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.96e-08,  1.96e-08,  3.66e-08,
  4.33e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  6.48e-09,  6.48e-09,  6.48e-09,  6.48e-09,  6.48e-09,  6.48e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  7.42e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  8.37e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  9.31e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.12e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.21e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.40e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.78e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.93e-08,  1.93e-08,  3.66e-08,
  4.43e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.07e-08,  2.07e-08,  3.66e-08,
  4.53e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.07e-08,  2.07e-08,  3.66e-08,
  4.62e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.07e-08,  2.07e-08,  3.66e-08,
  4.74e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.07e-08,  2.07e-08,  3.66e-08,
  5.32e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  5.60e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.07e-08,  2.07e-08,  3.66e-08,
  5.42e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  7.47e-09,  7.47e-09,  7.47e-09,  7.47e-09,  7.47e-09,  7.47e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  8.42e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  9.36e-09,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.03e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.31e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.50e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.88e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  1.97e-08,  2.03e-08,  2.03e-08,  3.66e-08,
  5.51e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.17e-08,  2.17e-08,  3.66e-08,
  5.61e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.17e-08,  2.17e-08,  3.66e-08,
  5.71e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.17e-08,  2.17e-08,  3.66e-08,
  5.82e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.17e-08,  2.17e-08,  3.66e-08,
  6.40e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  6.63e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.17e-08,  2.17e-08,  3.66e-08,
  6.50e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.47e-09,  8.47e-09,  8.47e-09,  8.47e-09,  8.47e-09,  8.47e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  9.41e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.13e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.22e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.41e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.60e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.79e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  1.98e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.13e-08,  2.13e-08,  3.66e-08,
  6.59e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.27e-08,  2.27e-08,  3.66e-08,
  6.69e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.27e-08,  2.27e-08,  3.66e-08,
  6.79e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.27e-08,  2.27e-08,  3.66e-08,
  6.90e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.27e-08,  2.27e-08,  3.66e-08,
  7.48e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  7.66e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.27e-08,  2.27e-08,  3.66e-08,
  7.58e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.47e-09,  9.47e-09,  9.47e-09,  9.47e-09,  9.47e-09,  9.47e-09,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.04e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.23e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.32e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.51e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.70e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.89e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.17e-08,  2.17e-08,  2.17e-08,  2.17e-08,  2.17e-08,  2.17e-08,  2.23e-08,  2.23e-08,  3.66e-08,
  7.68e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.38e-08,  2.38e-08,  3.66e-08,
  7.77e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.38e-08,  2.38e-08,  3.66e-08,
  7.87e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.38e-08,  2.38e-08,  3.66e-08,
  7.98e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.38e-08,  2.38e-08,  3.66e-08,
  8.57e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  8.69e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.38e-08,  2.38e-08,  3.66e-08,
  8.66e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.05e-08,  1.05e-08,  1.05e-08,  1.05e-08,  1.05e-08,  1.05e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.14e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.42e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.61e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  1.99e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.08e-08,  2.18e-08,  2.18e-08,  2.18e-08,  2.18e-08,  2.18e-08,  2.18e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  8.76e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  8.85e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  8.95e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  9.06e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  9.65e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  9.72e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  9.74e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.15e-08,  1.15e-08,  1.15e-08,  1.15e-08,  1.15e-08,  1.15e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.24e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.33e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.52e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.71e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.18e-08,  2.18e-08,  2.18e-08,  2.18e-08,  2.18e-08,  2.18e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  9.84e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  9.94e-09,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.00e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.01e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.07e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.08e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.25e-08,  1.25e-08,  1.25e-08,  1.25e-08,  1.25e-08,  1.25e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.43e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.62e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.81e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.10e-08,  2.10e-08,  2.10e-08,  2.10e-08,  2.10e-08,  2.10e-08,  2.19e-08,  2.19e-08,  2.19e-08,  2.19e-08,  2.19e-08,  2.19e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.09e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.10e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.11e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.12e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.18e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.19e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.34e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.53e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.72e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  1.91e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.10e-08,  2.10e-08,  2.10e-08,  2.10e-08,  2.10e-08,  2.10e-08,  2.19e-08,  2.19e-08,  2.19e-08,  2.19e-08,  2.19e-08,  2.19e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.20e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.21e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.22e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.23e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.29e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.28e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.30e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.44e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.63e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.82e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.01e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.20e-08,  2.20e-08,  2.20e-08,  2.20e-08,  2.20e-08,  2.20e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.31e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.32e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.33e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.38e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.54e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.73e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  1.92e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.20e-08,  2.20e-08,  2.20e-08,  2.20e-08,  2.20e-08,  2.20e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.49e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.64e-08,  1.74e-08,  1.74e-08,  1.74e-08,  1.74e-08,  1.74e-08,  1.74e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.83e-08,  1.93e-08,  1.93e-08,  1.93e-08,  1.93e-08,  1.93e-08,  1.93e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.02e-08,  2.12e-08,  2.12e-08,  2.12e-08,  2.12e-08,  2.12e-08,  2.12e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.59e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.74e-08,  1.74e-08,  1.74e-08,  1.74e-08,  1.74e-08,  1.74e-08,  1.84e-08,  1.84e-08,  1.84e-08,  1.84e-08,  1.84e-08,  1.84e-08,  1.93e-08,  1.93e-08,  1.93e-08,  1.93e-08,  1.93e-08,  1.93e-08,  2.03e-08,  2.03e-08,  2.03e-08,  2.03e-08,  2.03e-08,  2.03e-08,  2.12e-08,  2.12e-08,  2.12e-08,  2.12e-08,  2.12e-08,  2.12e-08,  2.22e-08,  2.22e-08,  2.22e-08,  2.22e-08,  2.22e-08,  2.22e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.69e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.84e-08,  1.84e-08,  1.84e-08,  1.84e-08,  1.84e-08,  1.84e-08,  1.94e-08,  1.94e-08,  1.94e-08,  1.94e-08,  1.94e-08,  1.94e-08,  2.03e-08,  2.03e-08,  2.03e-08,  2.03e-08,  2.03e-08,  2.03e-08,  2.13e-08,  2.13e-08,  2.13e-08,  2.13e-08,  2.13e-08,  2.13e-08,  2.22e-08,  2.22e-08,  2.22e-08,  2.22e-08,  2.22e-08,  2.22e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.80e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.11e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.21e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.31e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
  1.45e-08,  1.86e-08,  1.86e-08,  1.86e-08,  1.86e-08,  1.86e-08,  1.86e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  1.90e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.00e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.09e-08,  2.19e-08,  2.19e-08,  2.19e-08,  2.19e-08,  2.19e-08,  2.19e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  3.66e-08,
 -1.00e+00,  1.86e-08,  1.86e-08,  1.86e-08,  1.86e-08,  1.86e-08,  1.86e-08,  1.96e-08,  1.96e-08,  1.96e-08,  1.96e-08,  1.96e-08,  1.96e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.07e-08,  2.17e-08,  2.17e-08,  2.17e-08,  2.17e-08,  2.17e-08,  2.17e-08,  2.27e-08,  2.27e-08,  2.27e-08,  2.27e-08,  2.27e-08,  2.27e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08,  2.55e-08, -1.00e+00,
  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08,  3.66e-08, -1.00e+00, -1.00e+00
  };

